import datetime
import struct
import ujson

from typing import BinaryIO, Dict, List, Optional, Self, Tuple

from edge.model.types import Location, Geoposition, Dir
from edge.service.about import IAboutService
from edge.service.tz import ITzService
from edge.types import FormatTemplate


class BlobMeta:
    def __init__(self):
        pass

    def __str__(self) -> str:
        raise NotImplementedError

    def serialize(self) -> Dict:
        raise NotImplementedError


class JsonBlobMeta(BlobMeta):
    def __init__(self, jnode: Dict):
        BlobMeta.__init__(self)
        self._jnode = jnode

    def __str__(self) -> str:
        return ujson.dumps(self._jnode)

    def serialize(self) -> Dict:
        return self._jnode


class StructBlobMeta(BlobMeta):
    """
        struct meta_v0 {
        uint8_t    ver_;
        uint8_t    blob_name_size;
        uint16_t   blob_sign_size;
        uint16_t   blob_cert_size;
        uint8_t    reserved[26];
        uint8_t    blob_name_data[96];
        uint8_t    blob_sign_data[1024];
        uint8_t    blob_cert_data[8192];
    }

    struct meta_v1 {
        uint8_t    ver_;
        uint8_t    blob_name_size;
        uint16_t   blob_sign_size;
        uint16_t   blob_cert_size;
        uint8_t    reserved[26];
        uint8_t    blob_name_data[96];
        uint8_t    blob_sign_data[8192];
        uint8_t    blob_cert_data[8192];
    }
    """

    def __init__(self, sign_data: str, sign_size: int, cert: Optional[str] = ''):
        BlobMeta.__init__(self)
        self.struct = struct.pack('<BBHH26x96s8192s8192s', 1, 0, sign_size,
                                  len(cert), '', sign_data, cert)

    def __str__(self):
        raise NotImplementedError

    def serialize(self):
        raise NotImplementedError


class Blob:
    def __init__(self, profile: str, file_ext: str):
        self._name: Optional[str] = None
        self._profile = profile
        self._file_ext = file_ext
        self._hash = None

    @property
    def name(self) -> str:
        return self._name or ''

    @name.setter
    def name(self, value: str):
        if len(value):
            self._name = value
        else:
            raise RuntimeError

    @property
    def profile(self) -> str:
        return self._profile

    @property
    def min_ts(self) -> int:
        """
        :returns:
            value in usec
        """
        raise NotImplementedError

    @property
    def max_ts(self) -> int:
        """
        :return:
            value in usec
        """
        raise NotImplementedError

    @property
    def file_path(self) -> Optional[str]:
        raise NotImplementedError

    @property
    def file_ext(self) -> str:
        return self._file_ext

    @property
    def file_data(self) -> str:
        raise NotImplementedError

    @property
    def file_meta(self) -> Optional[str]:
        return None

    @property
    def hash(self) -> str:
        return self._hash

    @hash.setter
    def hash(self, value: Tuple[str, str]) -> None:
        assert len(value) == 2
        self._hash = value

    def serialize(self, tz_service: Optional[ITzService] = None) -> Dict:
        raise NotImplementedError


class MemBlob(Blob):
    def __init__(self, profile: str, ts: int, data: bytes,
                 meta: Optional[BlobMeta] = None,
                 file_ext: str = 'jpeg'):

        super().__init__(profile, file_ext=file_ext)

        self._ts = ts
        self._data = data
        self._meta = meta

    @property
    def min_ts(self) -> int:
        return self._ts

    @property
    def max_ts(self) -> int:
        return self._ts

    @property
    def file_path(self) -> Optional[str]:
        return None

    @property
    def file_data(self) -> bytes:
        return self._data

    @property
    def meta(self) -> Optional[BlobMeta]:
        return self._meta

    @meta.setter
    def meta(self, value: Optional[BlobMeta]):
        self._meta = value

    def serialize(self, tz_service: Optional[ITzService] = None) -> Dict:
        # jnode = {
        #     "name": "0_150402739915.jpeg",
        #     "info": {
        #         "size": [1920, 1200, 100500],
        #         "trans": {
        #             "o": [0.0, 0.0],
        #             "i": [1.0, 0.0],
        #             "j": [0.0, 1.0]
        #         }
        #     },
        #     "hash": [
        #         "md5",
        #         "2ce2e89e20dd5ad3b611f12e56e6fa79f35be149d8b226c23666ced9c73acce2"
        #     ]
        # }

        jnode = {
            'name': self.name,
        }

        if tz_service:
            jnode['time'] = tz_service.get_datetime_isostr(self.min_ts / 1e6)

        if self._meta:
            jnode['info'] = self._meta.serialize()

        if self.hash:
            jnode['hash'] = self.hash

        return jnode


class FileBlob(Blob):
    def __init__(self, min_ts: int, max_ts: int, data: BinaryIO,
                 meta: Optional[BlobMeta] = None,
                 profile: str = 'measurement'):

        super().__init__(profile, file_ext='mp4')

        self._min_ts = min_ts
        self._max_ts = max_ts

        self._file = data
        self._meta = meta

    def __del__(self):
        self._file.close()

    @property
    def min_ts(self) -> int:
        return self._min_ts

    @property
    def max_ts(self) -> int:
        return self._max_ts

    @property
    def file_path(self) -> str:
        return self._file.name

    @property
    def file_data(self) -> bytes:
        return self._file.read()

    def serialize(self, tz_service: Optional[ITzService] = None) -> Dict:
        jnode = {
            'name': self.name,
        }

        if tz_service:
            jnode['time'] = [
                tz_service.get_datetime_isostr(self.min_ts / 1e6),
                tz_service.get_datetime_isostr(self.max_ts / 1e6)
            ]

        if self._meta:
            jnode['info'] = self._meta.serialize()

        if self.hash:
            jnode['hash'] = self.hash

        return jnode


class ImFooterColArgs:
    def __init__(self, width: int, align: int, template: str):
        self.width = width
        self.align = align
        self.template = template

    def eval(self, mapping: Dict[str, str]) -> Self:
        value = FormatTemplate(self.template).substitute(mapping)
        return ImFooterColArgs(self.width, self.align, value)


class ImFooterRowArgs:
    def __init__(self, enabled: bool, cols: List[ImFooterColArgs]):
        self.enabled = enabled
        self.cols = cols

    def eval(self, mapping: Dict[str, str]) -> Self:
        cols = [col.eval(mapping) for col in self.cols]
        return ImFooterRowArgs(self.enabled, cols)


class ImFooterArgs:
    def __init__(self, height: int, font_size: int, rows: List[ImFooterRowArgs]):
        """
        :param height:
        :param font_size:
            Font size, pixels
        :param rows:
        """
        self.height = height
        self.font_size = font_size
        self.rows = rows

    def eval(self, footer_vars) -> Self:
        """
        :type footer_vars: ImFooterVars
        """
        var_names = set()
        for row in self.rows:
            for col in row.cols:
                var_names.update(FormatTemplate.get_pholders(col.template))

        mapping = {}
        for var_name in var_names:
            if var_name == 'datetime':
                mapping[var_name] = footer_vars.datetime

            elif var_name == 'datetime-ms':
                mapping[var_name] = footer_vars.datetime_ms

            elif var_name == 'datetime-control':
                mapping[var_name] = footer_vars.control_datetime

            elif var_name == 'datetime-control-ms':
                mapping[var_name] = footer_vars.control_datetime_ms

            elif var_name == 'time-control-ms':
                mapping[var_name] = footer_vars.control_time_ms

            elif var_name == 'post-name':
                mapping[var_name] = footer_vars.location

            elif var_name == 'geolocation':
                mapping[var_name] = footer_vars.geolocation

            elif var_name == 'plate':
                mapping[var_name] = footer_vars.plate_text

            elif var_name == 'speed':
                mapping[var_name] = footer_vars.spot_speed

            elif var_name == 'serial':
                mapping[var_name] = footer_vars.serial

            elif var_name == 'lane':
                mapping[var_name] = footer_vars.lane  # TODO mb skip first symbol ?

            elif var_name == 'dir':
                mapping[var_name] = footer_vars.dir

            elif var_name == 'video-speed':
                mapping[var_name] = footer_vars.video_speed

            elif var_name == 'video-speed2':
                mapping[var_name] = footer_vars.ra_cam_speed

            elif var_name == 'section-speed':
                mapping[var_name] = footer_vars.section_speed

            elif var_name == 'section-overspeed':
                mapping[var_name] = footer_vars.section_overspeed

            elif var_name == 'radar-speed':
                mapping[var_name] = footer_vars.radar_speed

            elif var_name == 'main-speed':
                mapping[var_name] = footer_vars.main_speed

            elif var_name == 'wim-speed':
                mapping[var_name] = footer_vars.wim_speed

            elif var_name == 'control-speed':
                mapping[var_name] = footer_vars.control_speed

            elif var_name == 'wim-overspeed':
                mapping[var_name] = footer_vars.wim_overspeed

            elif var_name == 'main-overspeed':
                mapping[var_name] = footer_vars.main_overspeed

            elif var_name == 'radar-speed2':
                ra_speed2_str = ''
                ra_speed_str = footer_vars.radar_speed
                if ra_speed_str not in ['', '-']:
                    ra_speed_int = int(ra_speed_str)
                    if ra_speed_int > 240:
                        ra_speed2_str = '>240'
                    elif ra_speed_int < 20:
                        ra_speed2_str = '<20'
                    else:
                        ra_speed2_str = ra_speed_str

                mapping[var_name] = ra_speed2_str

            elif var_name == 'max-speed':
                mapping[var_name] = footer_vars.max_speed

            elif var_name == 'speed-limit':
                mapping[var_name] = footer_vars.max_speed

            elif var_name == 'radar-overspeed':
                mapping[var_name] = footer_vars.radar_overspeed

            elif var_name == 'vendor':
                mapping[var_name] = footer_vars.vendor

            elif var_name == 'model':
                mapping[var_name] = footer_vars.model

            elif var_name == 'halt-time':
                mapping[var_name] = footer_vars.halt_time

            else:
                raise NotImplementedError

        rows = [r.eval(mapping) for r in self.rows]
        return ImFooterArgs(self.height, self.font_size, rows)


class ImFooterVars:
    def __init__(self,
                 tz_service: ITzService,
                 about_service: IAboutService,
                 ts: int,
                 control_ts: Optional[int] = None,
                 defs: Optional[Dict] = None,
                 location: Optional[Location] = None,
                 geolocation: Optional[Geoposition] = None,
                 plate_text: Optional[str] = None,
                 spot_speed: Optional[float] = None,
                 direction: Optional[Dir] = Dir(Dir.UNKNOWN),
                 video_speed: Optional[float] = None,
                 ra_cam_speed: Optional[float] = None,
                 radar_speed: Optional[float] = None,
                 max_speed: Optional[float] = None,
                 radar_overspeed: Optional[float] = None,
                 section_speed: Optional[float] = None,
                 section_overspeed: Optional[float] = None,
                 wim_speed: Optional[float] = None,
                 main_speed: Optional[float] = None,
                 control_speed: Optional[float] = None,
                 wim_overspeed: Optional[float] = None,
                 main_overspeed: Optional[float] = None,
                 halt_time: Optional[int] = None):

        self.geo = geolocation
        dt_local = tz_service.get_datetime(ts / 1e6)
        self.dt_utc = tz_service.get_datetime_utc(ts / 1e6)
        self._about_service = about_service
        # <22.01.2018 17:00:33>
        self.datetime = dt_local.strftime('%d.%m.%Y %H:%M:%S')
        # srtip 3 last digits of microsecond to get string like this <22.01.2018 17:00:33.544>
        self.datetime_ms = dt_local.strftime('%d.%m.%Y %H:%M:%S.%f')[:-3]

        if control_ts is None:
            self.control_datetime = ''
            self.control_datetime_ms = ''
            self.control_time_ms = ''
        else:
            control_dt = tz_service.get_datetime(control_ts / 1e6)
            # <22.01.2018 17:00:33>
            self.control_datetime = control_dt.strftime('%d.%m.%Y %H:%M:%S')
            # srtip 3 last digits of microsecond to get string like this <22.01.2018 17:00:33.544>
            self.control_datetime_ms = control_dt.strftime('%d.%m.%Y %H:%M:%S.%f')[:-3]
            self.control_time_ms = control_dt.strftime('%H:%M:%S.%f')[:-3]

        defs = defs or dict()
        self.location = location.place if (location is not None and location.place != '') else defs.get('post-name', '')
        self.lane = location.lane if (location is not None and location.lane != '') else defs.get('lane', '')
        self.geolocation = str(geolocation) if geolocation else defs.get('geolocation', '')
        self.plate_text = plate_text or defs.get('plate', '')
        self.spot_speed = str(spot_speed or defs.get('speed', ''))
        self.radar_speed = str(radar_speed or defs.get('radar-speed', ''))
        self.video_speed = str(video_speed or defs.get('video-speed', ''))
        self.ra_cam_speed = str(abs(ra_cam_speed)) if ra_cam_speed else defs.get('video-speed2', '')
        self.max_speed = str(max_speed or defs.get('max-speed', ''))
        self.radar_overspeed = str(radar_overspeed or defs.get('radar-overspeed', ''))
        self.section_speed = str(section_speed or defs.get('section-speed', ''))
        self.section_overspeed = str(section_overspeed or defs.get('section-overspeed', ''))

        self.wim_speed = str(wim_speed or defs.get('wim-speed', ''))
        self.main_speed = str(main_speed or defs.get('main-speed', ''))
        self.control_speed = str(control_speed or defs.get('control-speed', ''))

        self.wim_overspeed = str(wim_overspeed or defs.get('wim-overspeed', ''))
        self.main_overspeed = str(main_overspeed or defs.get('main-overspeed', ''))

        self.dir = defs.get(str(direction)) or str(direction)
        self.halt_time = halt_time

    @property
    def serial(self) -> str:
        return self._about_service.get_hw_serial()

    @property
    def vendor(self) -> str:
        return self._about_service.get_hw_vendor()

    @property
    def model(self) -> str:
        return self._about_service.get_hw_model()


class PlateImArgs:
    def __init__(self):
        self.width, self.height = 134, 30  # ~ 5:1


class ThumbImArgs:
    def __init__(self):
        # self.width, self.height = 220, 165  # 4:3
        self.width, self.height = 224, 224  # 1:1 for MMR


class VehicleImAnnoArgs:
    def __init__(self, plate: bool = False):
        self.plate = plate


class VehicleImArgs:
    def __init__(self, width: int, footer: ImFooterArgs, anno: VehicleImAnnoArgs):
        self.wh_ratio = 4.0 / 3.0

        self.width: int | float = width
        self.height: int | float = width / self.wh_ratio  # 4:3

        self.footer = footer
        self.anno = anno


class SceneImAnnoArgs:
    def __init__(self, track: bool = False, vehicle: bool = False, plate: bool = False):
        self.track = track
        self.vehicle = vehicle
        self.plate = plate


class SceneImArgs:
    def __init__(self, width: int, footer: ImFooterArgs, anno: SceneImAnnoArgs):
        self.width = width
        self.footer = footer
        self.anno = anno


class LiveImArgs:
    def __init__(self, footer):
        """
        :type footer: ImFooterArgs
        """
        self.footer = footer


class ContrastScheduleConf:
    MODE_ALWAYS = 'always'
    MODE_NEVER = 'never'
    MODE_NIGHT = 'twilight'
    MODE_TIME = 'time'

    def __init__(self, mode: str,
                 time_from: Optional[datetime.time] = None,
                 time_to: Optional[datetime.time] = None):

        self.mode = mode
        self.time_from = time_from
        self.time_to = time_to


class ImEngineArgs:
    def __init__(self, scene: SceneImArgs,
                 scene_footer: Dict,
                 vehicle: Optional[VehicleImArgs] = None,
                 vehicle_footer: Optional[Dict] = None,
                 contrast_args: Optional[ContrastScheduleConf] = None):

        self.scene = scene
        self.vehicle = vehicle
        self.thumb = ThumbImArgs()
        self.plate = PlateImArgs()

        self.scene_footer = scene_footer
        self.body_footer = vehicle_footer
        self.contrast_args = contrast_args
