from typing import List, Set, Tuple, Union
from shapely.geometry import polygon, point

LanePolygonPoint = Tuple[int, int]
LanePolygon = List[LanePolygonPoint]

LaneConfInt = Tuple[int, LanePolygon]
LaneConfStr = Tuple[str, LanePolygon]

LanesConfInt = List[LaneConfInt]
LanesConfStr = List[LaneConfStr]


class LaneDetector:
    def __init__(self,
                 lane_ident_to_polygon_map: Union[LaneConfInt, LanesConfStr],
                 road_img_width: int,
                 road_img_height: int):
        self._lane_ident_to_polygon_map = []

        for lane_id, polymask in lane_ident_to_polygon_map:
            self._lane_ident_to_polygon_map.append((lane_id, polygon.Polygon(polymask)))

        if len(self._lane_ident_to_polygon_map) > 0 and (road_img_height <= 0 or road_img_width <= 0):
            raise ValueError

        self._road_img_width = float(road_img_width)
        self._road_img_height = float(road_img_height)

    def get_lane(self, x: float, y: float) -> Union[Set[int], Set[str]]:
        """
        :param x:
            x coordinate of point [0.0, 1.0]
        :param y:
            y coordinate of point [0.0, 1.0]
        """
        return self.get_lane3(int(x * self._road_img_width), int(y * self._road_img_height))

    def get_lane2(self, left: float, top: float, right: float, bottom: float) -> Union[Set[int], Set[str]]:
        return self.get_lane((left + right) / 2, (top + bottom) / 2)

    def get_lane3(self, x: Union[int, float], y: Union[int, float]) -> Union[Set[int], Set[str]]:
        """
        :param x:
            x coordinate of point [0, road_img_width] px
        :param y:
            y coordinate of point [0, road_img_height] px
        """
        lane_set = set()
        if not len(self._lane_ident_to_polygon_map):
            return lane_set

        plate_center = point.Point(int(x), int(y))
        for ident, poly in self._lane_ident_to_polygon_map:
            if poly.contains(plate_center):
                lane_set.add(ident)

        return lane_set
