from typing import Dict, List, Optional, Self, Tuple

from edge_bbox.types import RectI
from edge.model.track import LrTrackItem
from edge.model.types import TimeSpan


class ImPlace:
    def __init__(self, r: int, x: int, y: int):
        self._r = r
        self._min_x = x - r
        self._max_x = x + r
        self._min_y = y - r
        self._max_y = y + r

    def __contains__(self, point_xy: Tuple[int, int]) -> bool:
        same_x = self._min_x < point_xy[0] < self._max_x
        same_y = self._min_y < point_xy[1] < self._max_y
        return same_x and same_y

    @property
    def xy(self) -> Tuple[int, int]:
        return self._min_x + self._r, self._min_y + self._r

    @property
    def r(self) -> int:
        return self._r

    @classmethod
    def create(cls, jnode: List[int]) -> Self:
        if len(jnode) != 3:
            raise ValueError

        return cls(*jnode)

    def serialize(self) -> List[int]:
        return [self._r, self._min_x + self._r, self._min_y + self._r]


class LrTrendItem:
    def __init__(self, place: ImPlace, span: TimeSpan, event: bool = False):
        self.place = place
        self.range = span
        self.event = event

    @classmethod
    def unpack(cls, jnode: Dict) -> Self:
        place = ImPlace.create(jnode['place'])
        span = TimeSpan.create(jnode['range'])
        return cls(place, span, bool(jnode.get('event')))

    def serialize(self) -> Dict:
        jnode = {
            'place': self.place.serialize(),
            'range': [self.range.min_ts, self.range.max_ts]
        }

        if self.event:
            jnode['event'] = True

        return jnode

    def set_event(self) -> None:
        self.event = True

    def clone(self) -> Self:
        return LrTrendItem(self.place, TimeSpan(self.range.min_ts, self.range.max_ts), self.event)


class LrTrend(list):
    def __init__(self, plate_wh: Tuple[int, int] = (0, 0), items: List[LrTrendItem] = ()):
        self.plate_w, self.plate_h = plate_wh
        super().__init__(items)

    @classmethod
    def deserialize(cls, jnode: Dict) -> Self:
        plate_wh = jnode['plate-wh']
        items = [LrTrendItem.unpack(x) for x in jnode['timeline']]
        return cls(plate_wh, items)

    def pack(self) -> Optional[Dict]:
        if len(self):
            return {
                'plate-wh': [self.plate_w, self.plate_h],
                'timeline': [item.serialize() for item in self]
            }

    @property
    def min_ts(self) -> int:
        return self[0].range.min_ts if len(self) else 0

    @property
    def max_ts(self) -> int:
        return self[-1].range.max_ts if len(self) else 0

    def set_event(self, event_ts: int) -> None:
        for item in self:  # type: LrTrendItem
            if event_ts in item.range:
                item.set_event()
                break

    def prolong(self, lr_items: List[LrTrackItem], range_tout: int, lost_tout: int) -> None:
        entry = None
        for lr_item in lr_items:  # type: LrTrackItem
            lr_bbox = RectI(*lr_item.plate_quad.get_abs_bbox())
            if entry is None:
                # cleanup
                if len(self) and lr_item.ts > (self[-1].range.max_ts + lost_tout):
                    del self[:]  # self.clear()

                if len(self):
                    entry = self[-1]

            if entry:
                # upscale plate-wh
                if lr_bbox.iw > self.plate_w and lr_bbox.ih > self.plate_h:
                    self.plate_w = lr_bbox.iw
                    self.plate_h = lr_bbox.ih

                if lr_bbox.center in entry.place:
                    if lr_item.ts < entry.range.max_ts + range_tout:
                        entry.range.max_ts = lr_item.ts
                        continue

            entry = LrTrendItem(ImPlace(lr_bbox.ih, *lr_bbox.center), TimeSpan(lr_item.ts, lr_item.ts))
            self.append(entry)

    def retrend(self, lost_tout: int = 0) -> Self:
        trend_item: LrTrendItem = self[0].clone()
        trend_items: List[LrTrendItem] = [trend_item]

        for item in self:  # type: LrTrendItem
            if item.place.xy in trend_item.place:
                if item.range.min_ts < trend_items[-1].range.max_ts + lost_tout:
                    trend_item.range.max_ts = item.range.max_ts
                    if item.event:
                        trend_item.set_event()

                    continue

            trend_item = item.clone()
            trend_items.append(trend_item)

        return LrTrend((self.plate_w, self.plate_h), trend_items)
