from typing import Dict, List, Self, Optional

from edge.model.geoposition import Geoposition, encode_geoposition


class Dir:
    NOWAY = 10
    UNKNOWN = 1
    APPROACHING = 11
    RECEDING = 12

    def __init__(self, direction: int):
        self._int_to_str_map = {
            Dir.UNKNOWN: 'unknown',
            Dir.APPROACHING: 'approaching',
            Dir.RECEDING: 'receding',
            Dir.NOWAY: 'any'
        }
        self.dir = direction

    @classmethod
    def deserialize(cls, direction: str) -> Self:
        str_to_int_map = {
            'unknown': Dir.UNKNOWN,
            'approaching': Dir.APPROACHING,
            'receding': Dir.RECEDING,
            'any': Dir.NOWAY
        }
        int_dir = str_to_int_map.get(direction)

        if int_dir is None:
            raise ValueError

        return cls(int_dir)

    def __eq__(self, other: Self) -> bool:
        return self.dir == other.dir

    def __ne__(self, other: Self) -> bool:
        return self.dir != other.dir

    def __str__(self) -> str:
        return self._int_to_str_map[self.dir]


class Location:
    def __init__(self, place: str = '', lane: str = ''):
        self.place = place
        self.lane = lane

    @classmethod
    def deserialize(cls, jnode: Dict) -> Self:
        place = jnode.get('place', '')
        return cls(place)

    def serialize(self, parent_jnode: Dict, jnode_key: str = 'location') -> Dict:
        jnode = {
        }

        if len(self.place):
            jnode['place'] = self.place

        if len(self.lane):
            jnode['lane'] = self.lane

        if len(self.place) or len(self.lane):
            parent_jnode[jnode_key] = jnode

        return parent_jnode


class Origin:
    def __init__(self, uuid: str, time: str, name: str, location: Location,
                 geolocation: Optional[Geoposition] = None):
        self._location = location
        self._geolocation = geolocation

        self._jnode = {
            'uuid': uuid,
            'time': time,
            'name': name
        }

    @property
    def uuid(self) -> str:
        return self._jnode['uuid']

    @property
    def time(self) -> str:
        return self._jnode['time']

    @property
    def location(self) -> Location:
        return self._location

    @property
    def geolocation(self) -> Optional[Geoposition]:
        return self._geolocation

    def serialize(self) -> Dict:
        self._location.serialize(self._jnode)

        if not self._jnode.get('geolocation') and self._geolocation:
            self._jnode['geolocation'] = encode_geoposition(self._geolocation)

        return self._jnode


class TimeSpan:
    def __init__(self, min_ts: int, max_ts: int):
        self.min_ts = min_ts
        self.max_ts = max_ts

    def __contains__(self, ts: int) -> bool:
        return self.min_ts <= ts <= self.max_ts

    @classmethod
    def create(cls, jnode: List[int]) -> Self:
        """
        :raise ValueError:
        """
        if len(jnode) != 2:
            raise ValueError

        return cls(*jnode)

    @property
    def duration(self) -> int:
        return self.max_ts - self.min_ts

    def check_overlap(self, other: Self) -> bool:
        return not ((other.min_ts > self.max_ts) or (other.max_ts < self.min_ts))

    def intersect(self, other: Self) -> Optional[Self]:
        if self.check_overlap(other):
            min_ts = max([self.min_ts, other.min_ts])
            max_ts = min([self.max_ts, other.max_ts])
            return TimeSpan(min_ts, max_ts)
