import os
import math
from PIL import Image
import iso8601

from typing import Dict, Optional
from datetime import date

from edge import eventlog
from edge.types import ImageRef
from edge.com import fs

from edge_rawutil import rawutil


class ImageFileQueue:
    def __init__(self, twisted_reactor, target_codec: Optional[str], root_dir: str,
                 min_file_count: int, max_file_count: int, log_level: str = 'info'):

        self._root_dir = root_dir
        self._min_file_count = min_file_count
        self._max_file_count = max_file_count

        self._target_codec = target_codec

        self._rawutil = rawutil.RawUtil.create(rawutil.MAX_WIDTH, rawutil.MAX_HEIGHT, log_level)
        twisted_reactor.addSystemEventTrigger('before', 'shutdown', self._rawutil.destroy)

        fs.silent_mkdir_r(self._root_dir)
        file_names = os.listdir(self._root_dir)

        file_names.sort(reverse=False, key=self._unpack_ts)

        self._last_file_suffix = self._unpack_ts(file_names.pop()) if file_names else 0
        self._expected_file_count = len(file_names)

    @staticmethod
    def _unpack_ts(name: str) -> int:
        return int(name.split('-')[-1])

    @classmethod
    def _cleanup(cls, root_dir: str, min_file_count: int) -> int:
        file_names = os.listdir(root_dir)
        file_names.sort(reverse=False, key=cls._unpack_ts)

        actual_file_count = len(file_names)
        for file_name in file_names:
            if actual_file_count <= min_file_count:
                break

            fs.silent_rm_r(fs.join_path(root_dir, str(file_name)))
            actual_file_count -= 1

        return actual_file_count

    def _get_next_filepath(self, next_name: int) -> str:
        self._last_file_suffix += 1
        next_path = fs.join_path(self._root_dir, f'{self._last_file_suffix}-{next_name}')
        return next_path

    def add_image_ref(self, image_ref: ImageRef) -> ImageRef:
        next_filepath = self._get_next_filepath(image_ref.ts)

        if self._target_codec is None:
            fs.edit_dir(self._root_dir, os.link, image_ref.url, next_filepath)

        elif image_ref.codec == 'JPEG' and self._target_codec == 'JPEG':
            fs.edit_dir(self._root_dir, os.link, image_ref.url, next_filepath)

        else:
            rgba_pil_image = self._rawutil.handle_image(
                image_ref.url, image_ref.codec, image_ref.width, image_ref.height)
            if not rgba_pil_image:
                raise RuntimeError('convert raw-image to rgba')

            if self._target_codec == 'JPEG':
                fs.edit_dir(self._root_dir, fs.write_jpeg_file, rgba_pil_image, next_filepath)

            elif self._target_codec == 'RGBA':
                fs.edit_dir(self._root_dir, fs.write_rgba_file, rgba_pil_image, next_filepath)

        self._expected_file_count += 1
        if self._expected_file_count >= self._max_file_count:
            self._expected_file_count = self._cleanup(self._root_dir, self._min_file_count)

        return ImageRef(ts=image_ref.ts, url=next_filepath, codec=self._target_codec, width=image_ref.width,
                        height=image_ref.height)

    def add_image(self, ts: Optional[int], rgba_pil_image: Image.Image) -> ImageRef:
        next_filepath = self._get_next_filepath(ts)

        if self._target_codec == 'JPEG':
            fs.edit_dir(self._root_dir, fs.write_jpeg_file, rgba_pil_image, next_filepath)

        elif self._target_codec == 'RGBA':
            fs.edit_dir(self._root_dir, fs.write_rgba_file, rgba_pil_image, next_filepath)

        self._expected_file_count += 1
        if self._expected_file_count >= self._max_file_count:
            self._expected_file_count = self._cleanup(self._root_dir, self._min_file_count)

        w, h = rgba_pil_image.size
        return ImageRef(ts=ts, url=next_filepath, codec=self._target_codec, width=w, height=h)

    def add_blob_image(self, blob_bytes: str, blob_ts: Optional[int] = None,
                       blob_codec: Optional[str] = None) -> ImageRef:

        next_filepath = self._get_next_filepath(blob_ts)

        # write file to the queue
        def write_bin_file(content_bytes):
            with open(next_filepath, 'wb') as bin_file:
                bin_file.write(content_bytes)

        fs.edit_dir(self._root_dir, write_bin_file, blob_bytes)

        # cleanup the queue
        self._expected_file_count += 1
        if self._expected_file_count >= self._max_file_count:
            self._expected_file_count = self._cleanup(self._root_dir, self._min_file_count)

        return ImageRef(ts=blob_ts, url=next_filepath, codec=blob_codec)

    def remove_image(self, image_ref) -> None:
        # assert image_ref.codec == 'JPEG' #todo : move to where called

        # noinspection PyBroadException
        try:
            os.unlink(image_ref.url)
            self._expected_file_count -= 1
        except Exception:
            eventlog.log_x()


class JsonFileQueue:
    def __init__(self, args):
        """
        :type args: QueueArgs
        """
        self._root_dir = args.endpoint
        self._min_file_count = args.min_file_count
        self._max_file_count = args.max_file_count

        fs.silent_mkdir_r(self._root_dir)
        file_name_list = list(map(str, os.listdir(self._root_dir)))
        file_name_list.sort(key=lambda name: int(name.split('-')[0]))

        self._last_file_suffix = int(file_name_list.pop().split('-')[0]) if file_name_list else 0
        self._expected_file_count = len(file_name_list)

    @staticmethod
    def _cleanup(root_dir: str, min_file_count: int) -> int:
        file_name_list = list(map(str, os.listdir(root_dir)))
        file_name_list.sort(key=lambda name: int(name.split('-')[0]))

        actual_file_count = len(file_name_list)
        for file_name in file_name_list:
            if actual_file_count <= min_file_count:
                break

            fs.silent_rm_r(fs.join_path(root_dir, str(file_name)))
            actual_file_count -= 1

        return actual_file_count

    def _get_next_filepath(self, next_filename: Optional[str] = None):
        self._last_file_suffix += 1
        if next_filename:
            next_filepath = fs.join_path(self._root_dir, f'{self._last_file_suffix}-{next_filename}')
        else:
            next_filepath = fs.join_path(self._root_dir, str(self._last_file_suffix))

        return next_filepath

    def add_jnode(self, jnode: Dict, next_filename: Optional[str] = None):
        next_filepath = self._get_next_filepath(next_filename)
        fs.edit_dir(self._root_dir, fs.write_json_file, jnode, next_filepath)

        self._expected_file_count += 1
        if self._expected_file_count >= self._max_file_count:
            self._expected_file_count = self._cleanup(self._root_dir, self._min_file_count)


class ImageFileTree:
    def __init__(self, twisted_reactor, root_dir: str, min_df: float, log_level: str = 'info'):
        self._root_dir = root_dir

        self._min_df = min_df
        self._files_in_dir = 250.0

        self._rawutil = rawutil.RawUtil.create(rawutil.MAX_WIDTH, rawutil.MAX_HEIGHT, log_level)
        twisted_reactor.addSystemEventTrigger('before', 'shutdown', self._rawutil.destroy)

        fs.silent_mkdir_r(self._root_dir)
        self._date_seq, self._file_seq = self._load_state(self._root_dir)

    @staticmethod
    def _load_state(root_path: str):
        """
        :type root_path: str
        :rtype: (date | None, long)
        """
        date_dirs = os.listdir(root_path)
        if not date_dirs:
            eventlog.log_i(f'done: root <{root_path}> is empty')
            return None, 0

        # define /:root/:max-date
        max_date_iso = max(date_dirs, key=lambda d: iso8601.parse_date(d).date())
        max_date_path = os.path.join(root_path, max_date_iso)

        # define /:root/:max-date/:max-seq
        seq_dirs = os.listdir(max_date_path)
        max_sec_dir = max(seq_dirs, key=lambda d: int(d))
        max_sec_path = os.path.join(max_date_path, max_sec_dir)

        # define /:root/:max-date/:max-seq/:max-file-seq
        im_files = os.listdir(max_sec_path)
        max_file_seq = max(im_files, key=lambda f: int(f.split('-')[0]))

        answer = iso8601.parse_date(max_date_iso).date(), int(max_file_seq.split('-')[0])
        eventlog.log_i(f'done: max-date:{max_date_iso}, max-seq:{max_file_seq}')
        return answer

    @staticmethod
    def _rotate(root_path: str) -> bool:
        date_dirs = os.listdir(root_path)
        if not date_dirs:
            eventlog.log_e(f'fail: root <{root_path}> is empty')
            return False

        min_date_iso = min(date_dirs, key=lambda d: iso8601.parse_date(d).date())
        min_date_path = os.path.join(root_path, min_date_iso)

        seq_dirs = os.listdir(min_date_path)
        if not seq_dirs:
            eventlog.log_e(f'fail: date-root <{min_date_path}> is empty')
            return False

        min_seq_dir = min(seq_dirs, key=lambda x: int(x))
        min_seq_path = fs.join_path(min_date_path, min_seq_dir)

        fs.silent_rm_r(min_seq_path)
        eventlog.log_i(f'done: rm <{min_seq_path}>')
        return True

    def _cleanup(self) -> bool:
        df = fs.get_df(self._root_dir)
        while True:
            if df > self._min_df:
                return True

            eventlog.log_i('exec: df:{:.4}, min-df:{:.4}'.format(df, self._min_df))
            success = self._rotate(self._root_dir)
            df = fs.get_df(self._root_dir)

            if success:
                eventlog.log_i('done: df:{:.4}, min-df:{:.4}'.format(df, self._min_df))
                continue

            eventlog.log_e('fail: df:{:.4}, min-df:{:.4}'.format(df, self._min_df))
            return False

    def _get_next_filepath(self, blob_ts: int, blob_name: str) -> str:
        blob_date = date.fromtimestamp(blob_ts / 1000000.0)
        if blob_date != self._date_seq:
            self._date_seq = blob_date
            self._file_seq = 0

        self._file_seq += 1

        seq_file = '{:05}-{}.jpeg'.format(self._file_seq, blob_name)
        seq_dir = '{:05}'.format(int(math.ceil(self._file_seq / self._files_in_dir) * self._files_in_dir))
        return fs.join_path(self._root_dir, self._date_seq.isoformat(), seq_dir, seq_file)

    def add_blob_image(self, blob_name: Optional[str], blob_bytes: str,
                       blob_ts: Optional[int] = None,
                       blob_format: Optional[str] = None) -> Optional[ImageRef]:

        if self._cleanup():
            next_filepath = self._get_next_filepath(blob_ts, blob_name)

            def write_bin_file(content_bytes):
                with open(next_filepath, 'wb') as bin_file:
                    bin_file.write(content_bytes)

            fs.edit_dir(os.path.dirname(next_filepath), write_bin_file, blob_bytes)
            return ImageRef(ts=blob_ts, url=next_filepath, codec=blob_format)

        eventlog.log_e('fail: disk is full')
