import datetime
import iso8601
import pytz

from typing import Tuple, Union

from edge.service.types import TzServiceArgs


def datetime_to_isostr(dt: datetime.datetime, microseconds: bool = False) -> str:
    tz_suffix = dt.strftime('%z')
    if tz_suffix == '+0000':
        tz_suffix = 'Z'

    precision = 6 if microseconds else 3
    precision += 1
    return dt.strftime('%Y-%m-%dT%H:%M:%S') + dt.strftime('.%f')[:precision] + tz_suffix


def isostr_to_datetime(isodate: str) -> datetime.datetime:
    return iso8601.parse_date(isodate)


def isostr_to_timestamp_usec(isodate: str) -> int:
    dt = iso8601.parse_date(isodate)
    epoch_dt = datetime.datetime.fromtimestamp(0, dt.tzinfo)
    ts = 1e6 * (dt - epoch_dt).total_seconds()
    return int(ts)


class ITzService:
    def __init__(self):
        pass

    def datetime_to_isostr(self, dt: datetime.datetime, microseconds: bool = False) -> str:
        raise NotImplementedError

    def get_datetime_isostr(self, ts: Union[int, float]) -> str:
        raise NotImplementedError

    def get_datetime(self, ts: Union[int, float]) -> datetime.datetime:
        raise NotImplementedError

    def get_datetime_utc(self, ts: int | float) -> datetime.datetime:
        raise NotImplementedError

    def get_datetime_now_isostr(self) -> str:
        raise NotImplementedError

    def get_datetime_now_isostr_seconds(self) -> Tuple[str, int]:
        raise NotImplementedError


class _TzService(ITzService):
    def __init__(self, args: TzServiceArgs):
        ITzService.__init__(self)
        self._tz = args.tz_info

    def datetime_to_isostr(self, dt: datetime.datetime, microseconds: bool = False) -> str:
        dt2 = dt.astimezone(self._tz)
        return datetime_to_isostr(dt2, microseconds)

    def get_datetime(self, ts: Union[int, float]) -> datetime.datetime:
        """
        :param ts:
            time in seconds since the epoch
        """
        return datetime.datetime.fromtimestamp(float(ts), tz=self._tz)

    def get_datetime_utc(self, ts: int | float) -> datetime.datetime:
        """
        :param ts:
            time in seconds since the epoch
        """
        return datetime.datetime.fromtimestamp(float(ts), tz=pytz.utc)

    def get_datetime_isostr(self, ts: Union[int, float]) -> str:
        """
        :param ts:
            time in seconds since the epoch
        :returns:
            ISO8601
        """
        # 2018-01-23T15:49:47.124+0400
        dt = self.get_datetime(ts)
        return datetime_to_isostr(dt)

    def get_datetime_now_isostr(self) -> str:
        """
        :returns:
            ISO8601
        """
        dt = datetime.datetime.now(self._tz)
        return datetime_to_isostr(dt)

    def get_datetime_now_isostr_seconds(self) -> Tuple[str, int]:
        """
        :returns:
            (ISO8601, time in seconds since the epoch)
        """
        dt = datetime.datetime.now(self._tz)
        return datetime_to_isostr(dt), int((dt - datetime.datetime(1970, 1, 1, tzinfo=self._tz)).total_seconds())


def init_service(args: TzServiceArgs) -> ITzService:
    return _TzService(args)


# noinspection PyUnusedLocal
def fini_service(service: ITzService):
    pass
