from typing import List, Optional, Self, Set, Union

from edge.com import fs
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.creator import BaseCreator, UiModToConfMap
from edge.webserver.sysconf.uiconf.constroller import CtrlUiconf


# noinspection PyUnusedLocal
def tr_get_creators(entry_uimod: Union[ph.UiMod, ph.UiMod2],
                    all_used_uimods: Set[Union[ph.UiMod, ph.UiMod2]],
                    uimod_to_newconf_map: UiModToConfMap,
                    forced: bool = False) -> List[BaseCreator]:

    if entry_uimod.type != ph.UI_TYPE_FUSION:
        return []

    try:
        (ra_uimod, cam_uimod, rstreaming_uimod, hstreaming_uimod, transit_uimod, fusion_uimod, rules_uimod,
         lpr_uimod, location_uimod) = (
            next(filter(lambda x: x.type == ph.UI_TYPE_RADAR, all_used_uimods), None),
            next(filter(lambda x: x.type in [ph.UI_TYPE_MV_CAMERA, ph.UI_TYPE_RTSP_CAMERA], all_used_uimods)),
            next(filter(lambda x: x.type == ph.UI_TYPE_RTSP_STREAMING, all_used_uimods), None),
            next(filter(lambda x: x.type == ph.UI_TYPE_HTTP_STREAMING, all_used_uimods)),
            next(filter(lambda x: x.type == ph.UI_TYPE_TRANSIT, all_used_uimods)),
            next(filter(lambda x: x.type == ph.UI_TYPE_FUSION, all_used_uimods)),
            next(filter(lambda x: x.type == ph.UI_TYPE_TRAFFIC_RULES, all_used_uimods)),
            next(filter(lambda x: x.type == ph.UI_TYPE_LPR, all_used_uimods)),
            next(filter(lambda x: x.type == ph.UI_TYPE_LOCATION, all_used_uimods), None))

        activate = any([
            uimod_to_newconf_map.get(cam_uimod),
            uimod_to_newconf_map.get(hstreaming_uimod),
            uimod_to_newconf_map.get(transit_uimod),
            uimod_to_newconf_map.get(rstreaming_uimod),
            uimod_to_newconf_map.get(location_uimod)
        ])

        if forced or activate:
            creator = TrCreator(
                ra_uimod, cam_uimod, lpr_uimod, rstreaming_uimod, hstreaming_uimod,
                rules_uimod, transit_uimod, fusion_uimod, location_uimod)

            return [creator]

    except StopIteration:
        pass

    return []


class TrCreator(BaseCreator):
    def __init__(self, ra_uimod: ph.UiMod,
                 cam_uimod: ph.UiMod,
                 lpr_uimod: ph.UiMod,
                 rstreaming_uimod: Optional[ph.UiMod],
                 hstreaming_uimod: ph.UiMod,
                 rules_uimod: ph.UiMod,
                 transit_uimod: ph.UiMod,
                 fusion_uimod: ph.UiMod,
                 location_uimod: Optional[ph.UiMod]):

        BaseCreator.__init__(self)
        self._ra_uimod = ra_uimod
        self._cam_uimod = cam_uimod
        self._lpr_uimod = lpr_uimod
        self._hstreaming_uimod = hstreaming_uimod
        self._rstreaming_uimod = rstreaming_uimod
        self._rules_uimod = rules_uimod
        self._transit_uimod = transit_uimod
        self._fusion_uimod = fusion_uimod
        self._location_uimod = location_uimod

        self._tr_dgmod = dg.DgMod.create(dg.EDGE_TYPE_TRACKER, self._fusion_uimod.pid, self._fusion_uimod.id)
        self._main_misc = None

    @property
    def dgmod(self) -> dg.DgMod:
        return self._tr_dgmod

    def prepare(self, uimod_to_newconf_map: UiModToConfMap) -> None:
        cam_uiconf = uimod_to_newconf_map.get(self._cam_uimod) or self._cam_uimod.load_conf()
        hstreaming_uiconf = uimod_to_newconf_map.get(self._hstreaming_uimod) or self._hstreaming_uimod.load_conf()
        hstreaming_pub = hstreaming_uiconf['rw']
        transit_uiconf = uimod_to_newconf_map.get(self._transit_uimod) or self._transit_uimod.load_conf()

        tr_dgmod = self._tr_dgmod
        spool_dir = ph.get_spool_dir(self._fusion_uimod.pid)
        if self._cam_uimod.type == ph.UI_TYPE_MV_CAMERA:
            cam_pri = cam_uiconf['ro']
            if cam_pri.get('raw-format', 'MONO8T') == 'MONO8T':
                sourse_dir = fs.join_path(spool_dir, f'mg{self._cam_uimod.id}-tr{self._fusion_uimod.id}-scene')
            else:
                sourse_dir = fs.join_path(spool_dir, f'rc{self._cam_uimod.id}-tr{self._fusion_uimod.id}-scene')

        elif self._cam_uimod.type == ph.UI_TYPE_RTSP_CAMERA:
            sourse_dir = fs.join_path(spool_dir, f'rg{self._cam_uimod.id}-tr{self._fusion_uimod.id}-scene')

        else:
            raise NotImplementedError

        plate_detected_mode = transit_uiconf['ro'].get('plate-detected-mode', False)

        # create main.conf
        main_conf = {
            'systest': CtrlUiconf.get_systest(),
            'plate-detected-mode': plate_detected_mode,
            'Monitoring': {
                'StatFile': fs.join_path(spool_dir, 'zabbix', 'tr0'),
                'StatTout': 5
            },
            'FileApi': {
                'AddTrackReact': {
                    'Dir': fs.join_path(spool_dir, f'lr{self._lpr_uimod.id}-tr{self._fusion_uimod.id}-track')
                },
                'AddSceneReact': {
                    'MgDir': sourse_dir,
                    'LrDir': fs.join_path(spool_dir, f'lr{self._lpr_uimod.id}-tr{self._fusion_uimod.id}-scene'),
                    'Tout': 750
                },
                'TrackEvent': {
                    'endpoints': [
                        fs.join_path(spool_dir, f'tr{self._fusion_uimod.id}-sc{self._rules_uimod.id}-track')
                    ],
                    'debounce-tout': 0 if plate_detected_mode else 3  # TODO: add tout when refactor debouncer
                },
                'SceneEvent': []
            },
            'store': {
                'max-file-count': 20,  # max-file-count >> debounce-tout * lr_track_fps (or we'll lose some events)
                'endpoint': fs.join_path(spool_dir, f'tr{tr_dgmod.id}-store')
            },
            'services': {
                'camera': self._cam_uimod.conf_path
            }
        }

        if self._location_uimod:
            main_conf['services']['location'] = self._location_uimod.conf_path
            location_uiconf = uimod_to_newconf_map.get(self._location_uimod) or self._location_uimod.load_conf()
            main_conf['ra-vehicle-ltrb'] = location_uiconf['ro'].get('ra-vehicle-ltrb', [1, 2, 1, 0])

        # enable http streaming
        if hstreaming_pub['enabled']:
            # noinspection PyUnresolvedReferences
            main_conf['FileApi']['SceneEvent'].append({
                'SkipRate': hstreaming_pub['live']['videostream']['skip-rate'],
                'Dir': fs.join_path(spool_dir, f'tr{self._fusion_uimod.id}-or{self._hstreaming_uimod.id}-scene')
            })

        # enable rtsp-streaming
        if self._rstreaming_uimod:
            rstreaming_uiconf = uimod_to_newconf_map.get(self._rstreaming_uimod) or self._rstreaming_uimod.load_conf()
            rstreaming_pub = rstreaming_uiconf.get('rw')
            if rstreaming_pub['enabled']:
                # noinspection PyUnresolvedReferences
                main_conf['FileApi']['SceneEvent'].append({
                    'SkipRate': rstreaming_uiconf['rw']['live']['videostream']['skip-rate'],
                    'Dir': fs.join_path(spool_dir, f'tr{self._fusion_uimod.id}-or{self._rstreaming_uimod.id}-scene')
                })

        # add ra-dir for speed pipelines
        if self._ra_uimod:
            main_conf['FileApi']['AddSceneReact']['RaDir'] = fs.join_path(
                spool_dir, f'ra{self._ra_uimod.id}-tr{self._fusion_uimod.id}-scene')

        # create misc.conf
        misc_conf = {
            'min_valid_radar_track_length': 5,
            'lpr_track_caching_time': 4,
            'radar_track_caching_time': 3
        }

        self._main_misc = main_conf, misc_conf

    def complete(self) -> None:
        main_conf, misc_conf = self._main_misc  # type: dict, dict

        edge_tr = self._tr_dgmod
        edge_tr.install()
        edge_tr.save_main_conf(main_conf)
        edge_tr.save_misc_conf(misc_conf)
        edge_tr.save_setupstore_conf(active=True)

    def rollback(self) -> None:
        pass

    def notify(self) -> None:
        dg.LifeCycle.restart_dgmod(self.dgmod.type, self.dgmod.id)


class TrDeleter:
    def __init__(self, dgmod: dg.DgMod):
        self._dgmod = dgmod

    @classmethod
    def create(cls, entry_uimod: ph.UiMod) -> Optional[Self]:
        if entry_uimod.type == ph.UI_TYPE_FUSION:
            dgmod = dg.DgMod.create(dg.EDGE_TYPE_TRACKER, entry_uimod.pid, entry_uimod.id)
            return cls(dgmod)

    def run(self) -> None:
        self._dgmod.uninstall()
