from typing import Dict, List, Self, Tuple, Union

from edge.types import ImageRef


class JsonImRef(ImageRef):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def deserialize(cls, jnode: Dict) -> Self:
        blob_ts = jnode['Ts']
        blob_url = jnode.get('Url')
        if blob_url is None or len(blob_url) == 0:
            blob_url = None

        blob_format = jnode['Codec']
        blob_w, blob_h = jnode['Wh']

        return cls(blob_ts, blob_url, blob_format, int(blob_w), int(blob_h))


class RaMargin:
    def __init__(self,
                 left: Union[float, int],
                 top: Union[float, int],
                 right: Union[float, int],
                 bottom: Union[float, int]):

        self.left = float(left)
        self.right = float(right)
        self.top = float(top)
        self.bottom = float(bottom)

    @classmethod
    def unpack(cls, jnode: List[float]) -> Self:
        """
        :param jnode:
            TrScene#/Ra-ltrb
        """
        if len(jnode) != 4:
            raise ValueError

        return cls(*jnode)


class LrQuad:
    def __init__(self, x0: int, y0: int, x1: int, y1: int, x2: int, y2: int, x3: int, y3: int):
        self.x0 = x0
        self.y0 = y0

        self.x1 = x1
        self.y1 = y1

        self.x2 = x2
        self.y2 = y2

        self.x3 = x3
        self.y3 = y3

    def map(self,
            source_im_w: int,
            source_im_h: int,
            target_im_w: int,
            target_im_h: int,
            margin_tb: int = 0,
            margin_lr: int = 0
            ) -> Tuple[int, int, int, int, int, int, int, int]:

        x0 = int(float(self.x0) / source_im_w * target_im_w - margin_lr)
        y0 = int(float(self.y0) / source_im_h * target_im_h - margin_tb)

        x1 = int(float(self.x1) / source_im_w * target_im_w + margin_lr)
        y1 = int(float(self.y1) / source_im_h * target_im_h - margin_tb)

        x2 = int(float(self.x2) / source_im_w * target_im_w + margin_lr)
        y2 = int(float(self.y2) / source_im_h * target_im_h + margin_tb)

        x3 = int(float(self.x3) / source_im_w * target_im_w - margin_lr)
        y3 = int(float(self.y3) / source_im_h * target_im_h + margin_tb)

        return x0, y0, x1, y1, x2, y2, x3, y3

    def get_rel_bbox(self, source_im_w: int, source_im_h: int) -> Tuple[float, float, float, float]:
        fl = float(min(self.x0, self.x3)) / source_im_w
        fr = float(max(self.x1, self.x2)) / source_im_w
        ft = float(min(self.y0, self.y1)) / source_im_h
        fb = float(max(self.y2, self.y3)) / source_im_h

        return fl, ft, fr, fb

    def map_bbox_wh(self, source_im_w: int, source_im_h: int, target_im_w: int, target_im_h: int) -> Tuple[int, int]:
        bbox_il = min(self.x0, self.x3)
        bbox_ir = max(self.x1, self.x2)
        bbox_it = min(self.y0, self.y1)
        bbox_ib = max(self.y2, self.y3)

        target_bbox_fw = float(bbox_ir - bbox_il) / source_im_w * target_im_w
        target_bbox_fh = float(bbox_ib - bbox_it) / source_im_h * target_im_h
        return int(target_bbox_fw), int(target_bbox_fh)

    def map_bbox_ref(self, source_im_w: int, source_im_h: int, target_im_w: int, target_im_h: int) -> Tuple[int, int]:
        target_bbox_il = float(min(self.x0, self.x3)) / source_im_w * target_im_w
        target_bbox_ir = float(max(self.x1, self.x2)) / source_im_w * target_im_w
        target_bbox_fb = float(max(self.y2, self.y3)) / source_im_h * target_im_h

        return int(0.5 * (target_bbox_il + target_bbox_ir)), int(target_bbox_fb)

    def map_bbox_ref2(self, source_im_w: int, source_im_h: int, target_im_w: int, target_im_h: int) -> Tuple[int, int]:
        target_bbox_il = float(min(self.x0, self.x3)) / source_im_w * target_im_w
        target_bbox_ir = float(max(self.x1, self.x2)) / source_im_w * target_im_w
        target_bbox_ft = float(max(self.y0, self.y1)) / source_im_h * target_im_h

        return int(0.5 * (target_bbox_il + target_bbox_ir)), int(target_bbox_ft)

    @classmethod
    def unpack(cls, jnode: List[float]) -> Self:
        if len(jnode) != 8:
            raise RuntimeError

        return cls(*jnode)


class FaceBbox:
    def __init__(self, x0: int, y0: int, x1: int, y1: int):
        self.x0 = x0
        self.y0 = y0
        self.x1 = x1
        self.y1 = y1

    @classmethod
    def unpack(cls, jnode: List[float]) -> Self:
        if len(jnode) != 4:
            raise RuntimeError

        return cls(*jnode)


class LrSceneItem:
    def __init__(self, lr_oid: int, lr_dir: int, lr_text: str, lr_quad: LrQuad, lr_trace: List):
        self.lr_oid = lr_oid
        self.lr_dir = lr_dir
        self.lr_text = lr_text
        self.lr_quad = lr_quad
        self.lr_trace = lr_trace

    def map_bbox_ref(self, source_im_w: int, source_im_h: int, target_im_w: int, target_im_h: int) -> Tuple[int, int]:
        return self.lr_quad.map_bbox_ref(source_im_w, source_im_h, target_im_w, target_im_h)

    def map_bbox_ref2(self, source_im_w: int, source_im_h: int, target_im_w: int, target_im_h: int) -> Tuple[int, int]:
        return self.lr_quad.map_bbox_ref2(source_im_w, source_im_h, target_im_w, target_im_h)

    @classmethod
    def unpack(cls, jnode: Dict) -> Self:
        """
        :param jnode:
            TrScene#/Lr/:index
            TrScene#/LrRa/:index/Lr
        """
        lr_plate_jnode = jnode['Plate']

        lr_oid = jnode['Id']
        lr_dir = jnode['Dir']
        lr_text = lr_plate_jnode['P']
        lr_quad = LrQuad.unpack(lr_plate_jnode['Q'])
        lr_trace = jnode['Trace']

        return cls(lr_oid, lr_dir, lr_text, lr_quad, lr_trace)


class FaceSceneItem:
    def __init__(self, face_id: int, face_conf: float, face_bbox: FaceBbox):
        self.face_id = face_id
        self.face_conf = face_conf
        self.face_bbox = face_bbox

    @classmethod
    def unpack(cls, jnode: Dict) -> Self:
        """
        :param jnode:
            TrScene#/Lr/:index
            TrScene#/LrRa/:index/Lr
        """
        face_id = jnode['Id']
        face_conf = jnode['Conf']
        face_bbox = FaceBbox.unpack(jnode['R'])

        return cls(face_id, face_conf, face_bbox)


class RaSceneItem:
    def __init__(self, rel_im_x: float, rel_im_y: float, ra_speed: int, ra_len: int, ra_dir: int, ra_ltrb: RaMargin):
        self.ra_rel_im_x = rel_im_x
        self.ra_rel_im_y = rel_im_y

        self.ra_speed = ra_speed
        self.ra_len = ra_len
        self.ra_dir = ra_dir
        self.ra_ltrb = ra_ltrb

    def map_bbox_ref(self, target_im_w: int, target_im_h: int) -> Tuple[int, int]:
        abs_x = self.ra_rel_im_x * target_im_w
        abs_y = self.ra_rel_im_y * target_im_h
        return int(abs_x), int(abs_y)

    @classmethod
    def unpack(cls, jnode: Dict) -> Self:
        """
        :param jnode:
            TrScene#/Ra/:index
            TrScene#/LrRa/:index/Ra
        """
        rel_im_x, rel_im_y = jnode['xy']
        ra_speed = int(jnode['speed'])
        ra_len = int(jnode['len'])
        ra_dir = int(jnode['dir'])
        ra_ltrb = RaMargin.unpack(jnode['r'])

        return cls(float(rel_im_x), float(rel_im_y), ra_speed, ra_len, ra_dir, ra_ltrb)


class TrSceneItem:
    def __init__(self, lr_item: LrSceneItem, ra_item: RaSceneItem):
        self.lr_item = lr_item
        self.ra_item = ra_item

    @classmethod
    def unpack(cls, jnode: Dict) -> Self:
        """
        :param jnode:
            TrScene#/LrRa/:index
        """
        lr_item = LrSceneItem.unpack(jnode['Lr'])
        ra_item = RaSceneItem.unpack(jnode['Ra'])
        return cls(lr_item, ra_item)


class TrScene:
    def __init__(self,
                 im_ref: ImageRef,
                 tr_items: List[TrSceneItem],
                 lr_items: List[LrSceneItem],
                 ra_items: List[RaSceneItem],
                 face_items: List[FaceSceneItem]):

        self.im_ref = im_ref

        self.tr_items = tr_items
        self.lr_items = lr_items
        self.ra_items = ra_items
        self.face_items = face_items

    @classmethod
    def unpack(cls, jnode: Dict) -> Self:
        """
        :param jnode:
            TrScene#/
        """
        im_ref = JsonImRef.deserialize(jnode['Image'])

        tr_items = [TrSceneItem.unpack(item_jnode) for item_jnode in jnode.get('LrRa', [])]
        lr_items = [LrSceneItem.unpack(item_jnode) for item_jnode in jnode.get('Lr', [])]
        ra_items = [RaSceneItem.unpack(item_jnode) for item_jnode in jnode.get('Ra', [])]
        face_items = [FaceSceneItem.unpack(item_jnode) for item_jnode in jnode.get('Face', [])]

        return cls(im_ref, tr_items, lr_items, ra_items, face_items)
