from typing import Dict, Self

from edge import eventlog
from edge.image.jnode.types import JsonImEngineArgs
from edge.types import JnodeError

from speedcontrol.detectors.bus_lane.types import DetectorArgs, RuleArgs, LimitsArgs
from speedcontrol.detectors.com.jnode.types import JsonBaseRuleArgs, JsonVideoArgs
from speedcontrol.detectors.types import IDetector
from speedcontrol.model.jnode.types import JsonSceneArgs
from speedcontrol.model.types import Identity


class JsonLimitsArgs(LimitsArgs):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def deserialize(cls, limits_jnode: Dict) -> Self:
        """
        :raise JsonError:
        """
        # deserialize #/classes
        types_jnode = limits_jnode.get('classes', [])

        if type(types_jnode) != list:
            eventlog.log_e_param('#/classes')
            raise JnodeError

        elif len(types_jnode) == 0:
            eventlog.log_e_param('#/classes')
            raise JnodeError

        for v_type in types_jnode:
            if not type(v_type) in [str]:
                eventlog.log_e_param('#/classes/%s' % v_type)
                raise JnodeError

        return cls(types_jnode)


class JsonRuleArgs(RuleArgs):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def deserialize(cls, rule_jnode: Dict) -> Self:
        """
        :raise JsonError:
        """
        if type(rule_jnode) != dict:
            eventlog.log_e_param('#/')
            raise JnodeError

        # deserialize #/name
        # deserialize #/enabled
        # deserialize #/schedule
        # deserialize #/ticket
        rule = JsonBaseRuleArgs.deserialize(rule_jnode)

        # deserialize #/limits
        limits_jnode = rule_jnode.get('limits')
        if type(rule_jnode) != dict:
            eventlog.log_e_param('#/limits')
            raise JnodeError

        limits = JsonLimitsArgs.deserialize(limits_jnode)

        return cls(rule.name, rule.enabled, rule.schedule, rule.ticket, limits)


class JsonDetectorArgs(DetectorArgs):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def deserialize(cls, jnode: Dict, index: int, def_footer_mapping: Dict) -> Self:
        """
        :raise JsonError:
        """
        # deserialize #/name
        name = jnode.get('name')
        if not type(name) in [str] or len(name) > 32:
            eventlog.log_e_param('#/name')
            raise JnodeError

        # deserialize #/store
        store = jnode.get('store')
        if not (type(store) in [str] and 1 <= len(store) <= 2):
            eventlog.log_e_param('#/store')
            raise JnodeError

        # deserialize #/rules
        rules_jnode = jnode.get('rules')
        if type(rules_jnode) != list:
            eventlog.log_e_param('#/rules')
            raise JnodeError

        # deserialize #/scene
        scene_jnode = jnode.get('scene')
        if type(scene_jnode) != dict:
            eventlog.log_e_param('#/scene')
            raise JnodeError
        scene = JsonSceneArgs.deserialize(scene_jnode)

        rules = list(map(JsonRuleArgs.deserialize, rules_jnode))
        identity = Identity(IDetector.TYPE_BUS_LANE, index)

        # deserialize #/images
        images_jnode = jnode.get('images')
        if type(images_jnode) != dict:
            eventlog.log_e_param('#/images')
            raise JnodeError

        im_engine = JsonImEngineArgs.deserialize(images_jnode['violation'])

        # deserialize #/videos/measuments
        video_jnode = jnode.get('videos', {}).get('violation')
        if video_jnode is not None:
            if video_jnode.get('enabled', False):
                video_args = JsonVideoArgs.deserialize(video_jnode)
            else:
                video_args = None
        else:
            video_args = None

        min_ts_dist_usec = jnode.get('min-image-dist', 1e3) * 1e3

        return cls(identity, name, store, rules, def_footer_mapping, scene, im_engine,
                   video_args, min_ts_dist_usec)
