from typing import Dict, List, Optional, Union

from edge.image.types import ImEngineArgs
from edge.model.types import Origin
from edge.service.tz import ITzService

from speedcontrol.detectors.com.types import ScheduleArgs, VideoArgs
from speedcontrol.model.types import Identity, Evidence, SceneArgs, BestViolationWaypoint, Ticket


class LimitsArgs:
    def __init__(self, vehicle_types: List[str]):
        self.vehicle_types = vehicle_types


class RuleArgs:
    def __init__(self, name: str, enabled: bool, schedule: Optional[ScheduleArgs], ticket: Ticket, limits: LimitsArgs):
        self.name = name
        self.enabled = enabled
        self.schedule = schedule
        self.ticket = ticket
        self.limits = limits


class DetectorArgs:
    def __init__(self,
                 identity: Identity,
                 name: str,
                 store: str,
                 rules: List[RuleArgs],
                 def_footer_mapping: Dict,
                 scene: SceneArgs,
                 im_engine: ImEngineArgs,
                 video_args: Optional[VideoArgs],
                 min_ts_dist_usec: Union[float, int] = 1e6):

        self.identity = identity
        self.name = name
        self.store = store
        self.rules = rules
        self.def_footer_mapping = def_footer_mapping
        self.scene = scene
        self.im_engine = im_engine
        self.video = video_args
        self.min_ts_dist_usec = min_ts_dist_usec


class BusLaneEvidence(Evidence):
    def __init__(self, origin: Origin, ts: int, duration: int, args: RuleArgs, detector_name: str):
        super().__init__(Evidence.NAME_BUS_LANE, origin, ts)
        self.detector_name = detector_name
        self.args = args
        self.duration = duration

    def serialize(self, tz_service: ITzService) -> Dict:
        jnode = Evidence.serialize(self, tz_service)

        violation_waypoints_jnode = list()
        best_waypoints_jnode = None
        for waypoint in self._waypoints:
            if waypoint.name == waypoint.NAME_VIOLATION:
                violation_waypoints_jnode.append(waypoint.serialize(tz_service))
            elif waypoint.name == waypoint.NAME_VIOLATION_BEST:
                best_waypoints_jnode = waypoint.serialize(tz_service)

        jnode['evidence'] = {
            'detector': {
                'name': self.detector_name,
                'rule': {
                    'name': self.args.name,
                    'schedule': self.args.schedule.serialize(),
                    'ticket': {
                        'name': self.args.ticket.name
                    },
                    'limits': {
                        'types': self.args.limits.vehicle_types
                    }
                }
            },
            'measurements': {
                'duration': self.duration,
                'waypoints': {
                    'violation': violation_waypoints_jnode
                }
            }
        }

        if best_waypoints_jnode is not None:
            jnode['evidence']['measurements']['waypoints']['best'] = best_waypoints_jnode

        if len(self._videos):
            videos_jnode = [x.serialize(tz_service) for x in self._videos]
            jnode['evidence']['measurements']['videos'] = videos_jnode

        return jnode

    @property
    def best_waypoint(self) -> Optional[BestViolationWaypoint]:
        bw = [x for x in self.waypoints if x.name == BestViolationWaypoint.NAME_VIOLATION_BEST]
        return bw[0] if len(bw) else None
