from typing import Dict, Self

from edge import eventlog
from edge.image.jnode.types import JsonImEngineArgs
from edge.types import JnodeError

from speedcontrol.detectors.com.jnode.types import JsonBaseRuleArgs
from speedcontrol.detectors.parking.types import DetectorArgs, RuleArgs, RuleLimits
from speedcontrol.detectors.types import IDetector
from speedcontrol.model.jnode.types import JsonSceneArgs
from speedcontrol.model.types import Identity


class JsonRuleLimits(RuleLimits):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def deserialize(cls, limits_jnode: Dict) -> Self:
        """
        :raise JnodeError:
        """
        # deserialize #/max-standing-time
        max_standing_time = limits_jnode.get('max-standing-time')
        if not (int == type(max_standing_time) and max_standing_time > 0):
            eventlog.log_e_param('#/max-standing-time')
            raise JnodeError

        return cls(max_standing_time)


class JsonRuleArgs(RuleArgs):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def deserialize(cls, rule_jnode: Dict) -> Self:
        """
        :raise JnodeError:
        """
        if type(rule_jnode) != dict:
            eventlog.log_e_param('#/')
            raise JnodeError

        # deserialize #/name
        # deserialize #/enabled
        # deserialize #/schedule
        # deserialize #/ticket
        rule = JsonBaseRuleArgs.deserialize(rule_jnode)

        # deserialize #/limits
        limits_jnode = rule_jnode.get('limits')
        if type(rule_jnode) != dict:
            eventlog.log_e_param('#/limits')
            raise JnodeError

        limits = JsonRuleLimits.deserialize(limits_jnode)

        return cls(rule.name, rule.enabled, rule.schedule, rule.ticket, limits)


class JsonDetectorArgs(DetectorArgs):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def deserialize(cls, jnode: Dict, index: int, def_footer_mapping: Dict) -> Self:
        """
        :raise JnodeError:
        """
        # deserialize #/name
        name = jnode.get('name')
        if not type(name) in [str] or len(name) > 32:
            eventlog.log_e_param('#/name')
            raise JnodeError

        # deserialize #/store
        store = jnode.get('store')
        if not (type(store) in [str] and 1 <= len(store) <= 2):
            eventlog.log_e_param('#/store')
            raise JnodeError

        # deserialize #/scene
        scene_jnode = jnode.get('scene')
        if type(scene_jnode) != dict:
            eventlog.log_e_param('#/scene')
            raise JnodeError
        scene = JsonSceneArgs.deserialize(scene_jnode)

        # deserialize #/rules
        rules_jnode = jnode.get('rules')
        if type(rules_jnode) != list:
            eventlog.log_e_param('#/rules')
            raise JnodeError

        rules = [x for x in map(JsonRuleArgs.deserialize, rules_jnode) if x.enabled]
        identity = Identity(IDetector.TYPE_PARKING, index)

        # deserialize #/images
        images_jnode = jnode.get('images')
        if type(images_jnode) != dict:
            eventlog.log_e_param('#/images')
            raise JnodeError

        init_im_engine = JsonImEngineArgs.deserialize(images_jnode['violation'])
        fini_im_engine = init_im_engine

        # deserialize range tout
        range_tout = jnode.get('range-tout')
        if type(range_tout) not in [int]:
            eventlog.log_e_param('#/range_tout')
            raise JnodeError

        # deserialize lost tout
        lost_tout = jnode.get('lost-tout')
        if type(lost_tout) not in [int]:
            eventlog.log_e_param('#/lost_tout')
            raise JnodeError

        return cls(
            identity, name, store, scene, rules,
            init_im_engine, fini_im_engine, def_footer_mapping, range_tout, lost_tout)
