import functools
import math

from typing import Callable, Dict, List, Optional, Tuple

from edge import eventlog
from edge.image.image import ImEnv, ImEngine, ImSizeFitter
from edge.image.types import FileBlob, ImFooterVars
from edge.lane_detector import LaneDetector
from edge.model.types import TimeSpan
from edge.service.about import IAboutService
from edge.service.tz import ITzService

from edge_bbox.types import mk_plate_bbox2, RectF

from speedcontrol.detectors.base.best import check_quad_dist_from_img_borders
from speedcontrol.detectors.com.scene import RlSceneMask
from speedcontrol.detectors.com.types import VideoArgs
from speedcontrol.detectors.red_light.types import DetectorArgs, RedLightEvidence, RuleArgs
from speedcontrol.detectors.types import IDetector
from speedcontrol.model.models import Track, TrTrackItem
from speedcontrol.model.types import BaseEvent, Event, Waypoint, RedLightViolationWaypoint, BestWaypoint, Vehicle, \
    Evidence
from speedcontrol.services.tl import ITlService
from speedcontrol.services.vs import IVsService, VsImageInfo, VsError, VsSessionAbsentError


def points_far_enought(rect1: RectF, rect2: RectF) -> bool:
    plate_h = rect1.fb - rect1.ft
    max_distance = plate_h * 0.7  # % of plate height

    rect1_c_x, rect1_c_y = rect1.center
    rect2_c_x, rect2_c_y = rect2.center

    if math.fabs(rect1_c_x - rect2_c_x) < max_distance and math.fabs(rect1_c_y - rect2_c_y) < max_distance:
        # points are in same rect
        return False
    else:
        # points are far enought
        return True


class _SceneImageInfo(VsImageInfo):
    def __init__(self, tr_item: TrTrackItem):
        super().__init__(tr_item.lr_item.ts)
        self.tr_item = tr_item


class _RuleServices:
    def __init__(self,
                 tz_service: ITzService,
                 vs_service: IVsService,
                 about_service: IAboutService,
                 tl_service: ITlService,
                 lane_service: LaneDetector):

        self.tz = tz_service
        self.vs = vs_service
        self.about = about_service
        self.tl = tl_service
        self.lane = lane_service


class RuleStats:
    def __init__(self,
                 total: int = 0,
                 valid_light: int = 0,
                 violation_total: int = 0,
                 pre_items_invalid: int = 0,
                 on_items_invalid: int = 0,
                 post_items_invalid: int = 0,
                 make_evidence_total: int = 0,
                 load_images: int = 0,
                 make_images: int = 0):

        self.total = total
        self.valid_light = valid_light

        self.violation_total = violation_total
        self.pre_items_invalid = pre_items_invalid
        self.on_items_invalid = on_items_invalid
        self.post_items_invalid = post_items_invalid
        self.load_images = load_images

        self.make_evidence_total = make_evidence_total
        self.make_evidence_make_images = make_images


class Rule:
    def __init__(self, twisted_reactor,
                 services: _RuleServices,
                 rule_args: RuleArgs,
                 detector_name: str,
                 stats_name: str,
                 init_im_env: ImEnv,
                 postbar_init_im_env: ImEnv,
                 postbar_fini_im_env: ImEnv,
                 video_conf: Optional[VideoArgs]):

        self._stats_name = stats_name
        self._reactor = twisted_reactor
        self._detector_name = detector_name

        self._services = services
        self._args = rule_args
        self._pre_im_env = init_im_env
        self._post_im_env1 = postbar_init_im_env
        self._post_im_env2 = postbar_fini_im_env

        self._stats = RuleStats()

        self._video_conf = video_conf
        if video_conf:
            self._im_fitter = ImSizeFitter((video_conf.max_w, video_conf.max_h))

    def check(self, track: Track,
              base: BaseEvent,
              pre_stopbar: List[TrTrackItem],
              on_stopbar: List[TrTrackItem],
              post_stopbar: List[TrTrackItem]
              ) -> Optional[Tuple[Evidence, int]]:

        self._stats.total += 1

        tl_result = self._services.tl.check(track.lr_track.min_ts, track.lr_track.max_ts)

        if not len(tl_result):
            self._stats.valid_light += 1
            return None

        for violation in tl_result:
            self._stats.violation_total += 1

            # check pre-stopbar conditions
            pre_ts_range = violation.intersect(
                TimeSpan(pre_stopbar[0].ts, pre_stopbar[-1].ts))
            if pre_ts_range is None:
                self._stats.pre_items_invalid += 1
                continue

            pre_tr_items = [tr_item for tr_item in pre_stopbar if tr_item.ts in pre_ts_range]
            if not pre_tr_items:
                self._stats.pre_items_invalid += 1
                continue

            # check on-stopbar conditions
            on_ts_range = violation.intersect(
                TimeSpan(on_stopbar[0].ts, on_stopbar[-1].ts))
            if on_ts_range is None:
                self._stats.on_items_invalid += 1
                continue

            on_tr_items = [tr_item for tr_item in on_stopbar if tr_item.ts in on_ts_range]
            if not on_tr_items:
                self._stats.on_items_invalid += 1
                continue

            # check post-stopbar conditions
            post_ts_range = violation.intersect(
                TimeSpan(post_stopbar[0].ts, post_stopbar[-1].ts))
            if post_ts_range is None:
                self._stats.post_items_invalid += 1
                continue

            post_tr_items = [tr_item for tr_item in post_stopbar if tr_item.ts in post_ts_range]
            if not post_tr_items:
                self._stats.post_items_invalid += 1
                continue

            # prepare im-environments
            pre_im_infos = list(map(_SceneImageInfo, pre_tr_items))
            post_im_infos = list(map(_SceneImageInfo, post_tr_items))
            self._stats.make_evidence_total += 1

            pre_im_info = self._services.vs.export_image_rgba2(
                self._pre_im_env.scene_rgba,
                self._pre_im_env.footer_h,
                pre_im_infos[:-6:-1])  # type: [_SceneImageInfo]
            if pre_im_info is None:
                eventlog.log_d('fail: load pre-stopbar image')
                self._stats.load_images += 1
                continue

            post_im_info1 = self._services.vs.export_image_rgba2(
                self._post_im_env1.scene_rgba,
                self._post_im_env1.footer_h,
                post_im_infos[0:5])  # type: [_SceneImageInfo]
            if post_im_info1 is None:
                eventlog.log_d('fail: load post-stopbar image-1')
                self._stats.load_images += 1
                continue

            post_im_info2 = self._services.vs.export_image_rgba2(
                self._post_im_env2.scene_rgba,
                self._post_im_env2.footer_h,
                post_im_infos[:-6:-1])  # type: [_SceneImageInfo]
            if post_im_info2 is None:
                eventlog.log_d('fail: load post-stopbar image-2')
                self._stats.load_images += 1
                continue

            # prepare evidence
            evidence = self._mk_evidence(
                base, track,
                pre_im_info.tr_item,
                post_im_info1.tr_item,
                post_im_info2.tr_item)

            if evidence:
                ref_ts = post_tr_items[0].lr_item.ts // 1000
                return evidence, ref_ts

    def add_video(self, ref_ts: int, callback: Callable[[Optional[FileBlob]], None]) -> None:
        if self._video_conf:
            scaled_w, scaled_h = self._im_fitter.scaledown(
                self._pre_im_env.scene_rgba.wh)

            # noinspection PyBroadException
            try:
                return self._services.vs.export_video(
                    ref_ts - self._video_conf.pre_record,
                    ref_ts + self._video_conf.post_record,
                    scaled_w, scaled_h, self._video_conf.quality,
                    callback)

            except VsSessionAbsentError:
                eventlog.log_w('fail: vs (session-absent)')

            except VsError:
                eventlog.log_e('fail: vs')

            except Exception:
                eventlog.log_x()

        callback(None)

    def _mk_evidence(self,
                     base_event: BaseEvent,
                     track: Track,
                     pre_tr_item: TrTrackItem,
                     post_tr_item1: TrTrackItem,
                     post_tr_item2: TrTrackItem
                     ) -> Optional[RedLightEvidence]:

        plate_text = base_event.get_plate_text()

        # get best vehicle
        best_waypoint = base_event.measurements[0].waypoints[2]  # type: BestWaypoint
        best_vehicle: Vehicle = best_waypoint.vehicle

        def mk_waypoint(services: _RuleServices, im_env: ImEnv, tr_item: TrTrackItem, waypoint_name: str) -> Waypoint:
            lr_scene_img = track.lr_track.best_image
            plate_rect = RectF(*tr_item.lr_item.plate_quad.get_rel_bbox(lr_scene_img.width, lr_scene_img.height))
            plate_bbox = mk_plate_bbox2(plate_rect, *im_env.scene_rgba.wh)
            origin = services.about.get_origin(im_env.scene_rgba.ts)
            wp_direction = track.tr_track.get_item_direction(tr_item)

            lane_set = self._services.lane.get_lane3(*tr_item.get_plate_quad().center)
            if len(lane_set) == 1:
                origin.location.lane = lane_set.pop()

            green_amber_time, red_time, red_amber_time, halt_time = services.tl.get_times(im_env.scene_rgba.ts)
            footer_vars = ImFooterVars(services.tz, self._services.about, im_env.scene_rgba.ts,
                                       defs=im_env.footer_def_vars,
                                       location=origin.location,
                                       geolocation=origin.geolocation,
                                       plate_text=plate_text,
                                       spot_speed=tr_item.ra_item.speed if tr_item.ra_item else None,
                                       direction=wp_direction,
                                       radar_speed=best_vehicle.spot_speed,
                                       video_speed=best_vehicle.camera_speed,
                                       ra_cam_speed=best_vehicle.ra_cam_speed,
                                       wim_speed=best_vehicle.wim_speed,
                                       main_speed=best_vehicle.main_speed,
                                       control_speed=best_vehicle.control_speed,
                                       halt_time=halt_time)

            im_blobs = im_env.engine.mk_blobs(
                im_env.scene_rgba, plate_bbox, footer_vars, track.lr_track, tr_item.lr_item.plate_quad,
                vehicle=False, scene=True, plate=True)

            if len(im_blobs) != 2:
                eventlog.log_d('fail: mk image-blobs')
                raise RuntimeError

            waypoint = RedLightViolationWaypoint(
                waypoint_name, im_env.scene_rgba.ts, origin.location, [tr_item.lr_item.plate_quad],
                green_amber_time, red_time, red_amber_time, halt_time)

            waypoint.add_images(im_blobs)
            return waypoint

        try:
            pre_waypoint = mk_waypoint(
                self._services, self._pre_im_env, pre_tr_item,
                Waypoint.NAME_PRE_STOPBAR)

            post_waypoint1 = mk_waypoint(
                self._services, self._post_im_env1, post_tr_item1,
                Waypoint.NAME_POST_STOPBAR)

            post_waypoint2 = mk_waypoint(
                self._services, self._post_im_env2, post_tr_item2,
                Waypoint.NAME_POST_STOPBAR)

        except RuntimeError:
            self._stats.make_evidence_make_images += 1
            return None

        event_ts = self._post_im_env1.scene_rgba.ts

        evidence = RedLightEvidence(
            self._services.about.get_origin(event_ts),
            event_ts, self._args, self._detector_name)
        evidence.add_waypoints([pre_waypoint, post_waypoint1, post_waypoint2])
        return evidence

    def collect_stats(self, json_handler: Callable[[Dict], None]) -> None:
        violation_total_fail = self._stats.pre_items_invalid + self._stats.on_items_invalid + \
                               self._stats.post_items_invalid + self._stats.load_images
        jnode = {
            "type": "red-light~rule",
            "name": self._stats_name,
            "stat": {
                "assert-track": {
                    "count": [self._stats.total,
                              self._stats.total - self._stats.valid_light,
                              self._stats.valid_light
                              ],
                    "green-light": self._stats.valid_light
                },
                "assert-violation": {
                    "count": [self._stats.violation_total,
                              self._stats.violation_total - violation_total_fail,
                              violation_total_fail
                              ],
                    "pre-zone-unmatch": self._stats.pre_items_invalid,
                    "on-zone-unmatch": self._stats.on_items_invalid,
                    "post-zone-unmatch": self._stats.post_items_invalid,
                    "load-images": self._stats.load_images,
                },
                "make-evidence": {
                    "count": [self._stats.make_evidence_total,
                              self._stats.make_evidence_total - self._stats.make_evidence_make_images,
                              self._stats.make_evidence_make_images
                              ],
                    "make-images": self._stats.make_evidence_make_images
                }
            }
        }
        json_handler(jnode)


class DetectorStats:
    def __init__(self,
                 total: int = 0,
                 zone_invalid: int = 0,
                 lr_invalid: int = 0,
                 emit_total: int = 0,
                 emit_fail: int = 0):

        self.total = total
        self.zone_invalid = zone_invalid

        self.lr_invalid = lr_invalid

        self.emit_total = emit_total
        self.emit_fail = emit_fail


class _Detector(IDetector):
    def __init__(self, twisted_reactor,
                 tz_service: ITzService,
                 vs_service: IVsService,
                 about_service: IAboutService,
                 tl_service: ITlService,
                 lane_service: LaneDetector,
                 args: DetectorArgs):

        IDetector.__init__(self)
        self._stats = DetectorStats()
        self._identity = args.identity

        self._sc_detector = RlSceneMask(args.scene)

        im_engine = ImEngine(args.im_engine)
        prebar_im_env = ImEnv(
            im_engine, args.im_engine.scene.footer.height, args.def_footer_mapping)

        postbar_init_im_env = ImEnv(
            im_engine, args.im_engine.scene.footer.height, args.def_footer_mapping)

        postbar_fini_im_env = ImEnv(
            im_engine, args.im_engine.scene.footer.height, args.def_footer_mapping)

        self._rules = []

        rule_services = _RuleServices(tz_service, vs_service, about_service, tl_service, lane_service)
        rule_index = 0
        for rule in args.rules:
            rule_name = 'red-light.{}~rule.{}'.format(self._identity.id, rule_index)
            rule_index += 1
            rule = Rule(
                twisted_reactor,
                rule_services, rule, args.name, rule_name, prebar_im_env, postbar_init_im_env,
                postbar_fini_im_env, args.video)
            self._rules.append(rule)

        self._handler = None
        self._receding_direction = args.receding_direction

    def set_handler(self, event_handler: Callable[[Dict], None]) -> None:
        if self._handler:
            raise NotImplementedError

        self._handler = event_handler

    def handle_evidence(self, base: BaseEvent, evidence: RedLightEvidence, video_blob: Optional[FileBlob]) -> None:
        if video_blob:
            evidence.add_videos([video_blob])

        self._stats.emit_total += 1

        if not self._handler(Event(self._identity, base, evidence)):
            self._stats.emit_fail += 1

    def analyze(self, track: Track, base: BaseEvent) -> None:
        def split_track(
                sc_detector: RlSceneMask
        ) -> Optional[Tuple[List[TrTrackItem], List[TrTrackItem], List[TrTrackItem]]]:
            lr_scene_img = track.lr_track.best_image
            tr_items = [item for item in track.tr_track if item.lr_item and check_quad_dist_from_img_borders(
                    item.lr_item.plate_quad, lr_scene_img.width, lr_scene_img.height)]

            pre_stopbar, on_stopbar, post_stop_bar = [], [], []
            for tr_item in tr_items:
                check_answer = sc_detector.check_quad(tr_item.lr_item.plate_quad, self._receding_direction)

                if check_answer == sc_detector.STOPBAR_PRE:
                    pre_stopbar.append(tr_item)

                elif check_answer == sc_detector.STOPBAR_ON:
                    if len(pre_stopbar):
                        on_stopbar.append(tr_item)

                    else:
                        # we didn't have pre-stopbar items yet
                        # car direction is wrong or we din't have points before stop bar
                        # print 'err in'
                        return None

                elif check_answer == sc_detector.STOPBAR_POST:
                    if len(pre_stopbar) == 0 or len(on_stopbar) == 0:
                        # we didn't have pre-stop-bar or inside-stop-bar items yet
                        # car direction is wrong or we din't have points before stop bar
                        # print 'err post'
                        return None

                    post_stop_bar.append(tr_item)

            return pre_stopbar, on_stopbar, post_stop_bar

        if '*' in base.get_plate_text():
            eventlog.log_d('done: skip event due to asterisk in plate-text')
            self._stats.lr_invalid += 1
            return

        answer = split_track(self._sc_detector)
        if answer is None:
            # controlling direction of movement is not defined
            eventlog.log_e('fail: skip event due to lack of points')
            self._stats.zone_invalid += 1
            return

        elif 0 in list(map(len, answer)):
            # stopbar crossing is not detected
            eventlog.log_e('fail: skip event due to lack of points')
            self._stats.zone_invalid += 1
            return

        for rule in self._rules:
            reply = rule.check(track, base, *answer)
            if reply is None:
                continue

            evidence, ref_ts = reply
            if evidence:
                handler = functools.partial(self.handle_evidence, base, evidence)
                rule.add_video(ref_ts, handler)
                break

    def collect_stats(self, json_handler: Callable[[Dict], None]) -> None:
        total_fails = self._stats.lr_invalid + self._stats.zone_invalid
        jnode = {
            "type": "red-light",
            "name": "red-light.{}".format(self._identity.id),
            "stat": {
                "assert-track": {
                    "count": [self._stats.total, self._stats.total - total_fails, total_fails],
                    "zone-unmatch": self._stats.zone_invalid,
                    "lr-invalid": self._stats.lr_invalid
                },
                "emit-event": {
                    "count": [self._stats.emit_total,
                              self._stats.emit_total - self._stats.emit_fail,
                              self._stats.emit_fail]
                }
            }
        }

        json_handler(jnode)
        list(map(lambda x: x.collect_stats(json_handler), self._rules))


def create(*args, **kwargs) -> IDetector:
    return _Detector(*args, **kwargs)
