from typing import Dict, Optional, Self

from edge import eventlog
from edge.image.jnode.types import JsonImEngineArgs
from edge.types import JnodeError

from speedcontrol.detectors.com.jnode.types import JsonBaseRuleArgs, JsonVideoArgs
from speedcontrol.detectors.red_light.types import DetectorArgs, RuleArgs
from speedcontrol.detectors.types import IDetector
from speedcontrol.model.jnode.types import JsonSceneArgs
from speedcontrol.model.types import Identity


class JsonRuleArgs(RuleArgs):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def deserialize(cls, rule_jnode: Dict) -> Optional[Self]:
        """
        :raise JnodeError:
        """
        if type(rule_jnode) != dict:
            eventlog.log_e_param('#/')
            raise JnodeError

        # deserialize #/name
        # deserialize #/enabled
        # deserialize #/ticket
        rule = JsonBaseRuleArgs.deserialize(rule_jnode)
        if not rule.enabled:
            return None

        return cls(rule.name, rule.ticket)


class JsonDetectorArgs(DetectorArgs):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def deserialize(cls, jnode: Dict, index: int, def_footer_mapping: Dict) -> Self:
        """
        :raise JnodeError:
        """
        # deserialize #/name
        name = jnode.get('name')
        if not type(name) in [str] or len(name) > 32:
            eventlog.log_e_param('#/name')
            raise JnodeError

        # deserialize #/direction
        direction = jnode.get('direction', 'receding')
        if not type(name) in [str] and direction not in ['receding', 'approaching']:
            eventlog.log_e_param('#/direction')
            raise JnodeError

        # deserialize #/store
        store = jnode.get('store')
        if not (type(store) in [str] and 1 <= len(store) <= 2):
            eventlog.log_e_param('#/store')
            raise JnodeError

        # deserialize #/scene
        scene_jnode = jnode.get('scene')
        if type(scene_jnode) != dict:
            eventlog.log_e_param('#/scene')
            raise JnodeError
        scene = JsonSceneArgs.deserialize(scene_jnode)

        # deserialize #/rules
        rules_jnode = jnode.get('rules')
        if type(rules_jnode) != list:
            eventlog.log_e_param('#/rules')
            raise JnodeError

        rules = list(filter(bool, list(map(JsonRuleArgs.deserialize, rules_jnode))))
        identity = Identity(IDetector.TYPE_RED_LIGHT, index)

        # deserialize #/images
        images_jnode = jnode.get('images')
        if type(images_jnode) != dict:
            eventlog.log_e_param('#/images')
            raise JnodeError

        im_engine = JsonImEngineArgs.deserialize(images_jnode['violation'])

        # deserialize #/videos/measuments
        video_jnode = jnode.get('videos', {}).get('violation')
        if video_jnode is not None:
            if video_jnode.get('enabled', False):
                video_args = JsonVideoArgs.deserialize(video_jnode)
            else:
                video_args = None
        else:
            video_args = None

        return cls(identity, name, store, scene, rules, im_engine,
                   def_footer_mapping, video_args, direction)
