from typing import cast, Dict, List, Optional, Tuple

from edge.image.types import ImEngineArgs
from edge.model.types import Origin
from edge.service.tz import ITzService

from speedcontrol.detectors.com.types import VideoArgs
from speedcontrol.model.types import Identity, Evidence, SceneArgs, RedLightViolationWaypoint, Ticket


class RuleArgs:
    def __init__(self, name: str, ticket: Ticket):
        self.name = name
        self.ticket = ticket


class DetectorArgs:
    def __init__(self,
                 identity: Identity,
                 name: str,
                 store: str,
                 scene: SceneArgs,
                 rules: List[RuleArgs],
                 im_engine: ImEngineArgs,
                 def_footer_mapping: Dict,
                 video: Optional[VideoArgs],
                 direction: str):

        self.identity = identity
        self.name = name
        self.store = store
        self.scene = scene
        self.rules = rules
        self.im_engine = im_engine
        self.def_footer_mapping = def_footer_mapping
        self.video = video
        self.receding_direction = direction == 'receding'


class RedLightEvidence(Evidence):
    def __init__(self, origin: Origin, ts: int, args: RuleArgs, detector_name: str):
        super().__init__(Evidence.NAME_RED_LIGHT, origin, ts)
        self.detector_name = detector_name
        self.args = args

    def serialize(self, tz_service: ITzService) -> Dict:
        jnode = Evidence.serialize(self, tz_service)
        jnode['evidence'] = {

        }
        # waypoints_jnode = dict()
        # for waypoint in self.waypoints:
        #     waypoints_jnode[waypoint.name] = waypoint.serialize(tz_service)

        pre_waypoints = [waypoint for waypoint in self._waypoints if waypoint.name == waypoint.NAME_PRE_STOPBAR]
        post_waypoints = [waypoint for waypoint in self._waypoints if waypoint.name == waypoint.NAME_POST_STOPBAR]

        pre_waypoints_jnode = [x.serialize(tz_service) for x in pre_waypoints]
        post_waypoints_jnode = [x.serialize(tz_service) for x in post_waypoints]

        jnode['evidence'] = {
            'detector': {
                'name': self.detector_name,
                'rule': {
                    'name': self.args.name,
                    'ticket': {
                        'name': self.args.ticket.name
                    }
                }
            },
            'measurements': {
                'waypoints': {
                    'pre-stopbar': pre_waypoints_jnode,
                    'post-stopbar': post_waypoints_jnode
                }
            }
        }

        if len(self._videos):
            videos_jnode = [x.serialize(tz_service) for x in self._videos]
            jnode['evidence']['measurements']['videos'] = videos_jnode

        return jnode

    def get_ukr_json_ev(self) -> Tuple[int, int, int, int]:
        """
        :returns
            (green_amber_time, red_time, red_amber_time, halt-time)
        """
        post_waypoints = [wp for wp in self._waypoints if wp.name == wp.NAME_POST_STOPBAR]
        post_waypoints.sort(key=lambda item: item.ts, reverse=True)

        post_waypoint_last = cast(RedLightViolationWaypoint, post_waypoints[0])

        green_amder_t = post_waypoint_last.green_amber_time
        red_t = post_waypoint_last.red_time
        red_amber_t = post_waypoint_last.red_amber_time
        halt_t = post_waypoint_last.halt_time

        return green_amder_t, red_t, red_amber_t, halt_t
