from typing import Callable

from edge import eventlog
from edge.service.about import IAboutService
from edge.service.tz import ITzService

from speedcontrol.detectors.com.schedule import Schedule
from speedcontrol.detectors.transit.types import DetectorArgs, TransitEvidence
from speedcontrol.detectors.types import IDetector
from speedcontrol.model.models import Track
from speedcontrol.model.types import BaseEvent, Event, BestWaypoint
from speedcontrol.services.clock import IClockService


def create(*args, **kwargs) -> IDetector:
    return _Detector(*args, **kwargs)


class _Detector(IDetector):
    def __init__(self,
                 about_service: IAboutService,
                 tz_service: ITzService,
                 clock_service: IClockService,
                 args: DetectorArgs):

        self._about = about_service
        self._clock_service = clock_service
        self._scheduler = Schedule(args.schedule_args, tz_service)

        self._enabled = args.enabled
        self._require_plate = args.require_plate
        self._identity = args.identity
        self._handler = None
        self._args = args
        IDetector.__init__(self)

    def set_handler(self, event_handler: Callable[[Event], None]) -> None:
        if self._handler:
            raise NotImplementedError

        self._handler = event_handler

    def analyze(self, track: Track, base: BaseEvent) -> None:
        # print 'transit-analyze {} {} {}'.format(track.le_track.ref, track.le_track.min_ts, track.le_track.max_ts)
        if not self._enabled:
            return

        waypoints_list = base.measurements[0].waypoints
        best_points = [waypoint for waypoint in waypoints_list if waypoint.name == waypoint.NAME_BEST]
        if len(best_points) != 1:
            eventlog.log_d('fail: cant find best waypoint')
            return

        best_waypoint = best_points[0]  # type: BestWaypoint
        if self._scheduler.check(best_waypoint.ts // 1000000):
            if self._require_plate and best_waypoint.vehicle.plate_absent:
                eventlog.log_d('fail: plate absent')
                return

            if track.le_track and len(track.le_track):
                if track.le_track.ref != self._args.le_loop_ref:
                    return

            control_ts = self._clock_service.get_control_ts(best_waypoint.ts)
            evidence = TransitEvidence(self._about.get_origin(best_waypoint.ts), best_waypoint.ts, control_ts=control_ts)
            event = Event(self._identity, base, evidence)
            self._handler(event)
