from typing import Dict, List, Self, Tuple

from edge import eventlog
from edge.types import JnodeError
from edge_bbox.types import RectF

from speedcontrol.model.types import Ticket, Mask, SceneArgs


class JsonSceneArgs(SceneArgs):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def deserialize(cls, zone_jnode: Dict) -> Self:
        """
        :raise JnodeError:
        """
        area_jnode = zone_jnode.get('area')
        if type(area_jnode) != dict:
            eventlog.log_e_param('#/area')
            raise JnodeError

        # deserialize #/area/mask
        mask = JsonMask.deserialize(area_jnode.get('mask'))

        # deserialize #/area/scene-wh
        scene_wh_jnode = area_jnode.get('scene-wh')
        if not type(scene_wh_jnode) in [list, tuple] or len(scene_wh_jnode) != 2:
            eventlog.log_e_param('#/area/scene-wh')
            raise JnodeError

        w, h = scene_wh_jnode
        if not (int == type(w) and 640 <= w <= 65535):
            eventlog.log_e_param('#area/scene-wh/0')
            raise JnodeError

        if not (int == type(h) and 480 <= h <= 65535):
            eventlog.log_e_param('#area/scene-wh/1')
            raise JnodeError

        return cls(mask, (w, h))


class JsonTicket(Ticket):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def deserialize(cls, ticket_jnode: Dict) -> Self:
        """
        :raise JnodeError:
        """
        # deserialize #/name
        name = ticket_jnode.get('name')
        if not (type(name) in [str]) or len(name) > 128:
            eventlog.log_e_param('#/name')
            raise JnodeError

        return cls(name)


class JsonMask(Mask):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def deserialize(cls, mask_jnode: List[Tuple[int, int]]) -> Self:
        """
        :raise JnodeError:
        """
        if type(mask_jnode) != list:
            eventlog.log_e_param('#/')
            raise JnodeError

        index = 0
        while index < len(mask_jnode):
            point_jnode = mask_jnode[index]
            if not (type(point_jnode) in [list, tuple] or len(point_jnode) != 2):
                eventlog.log_e_param('#/%s' % index)
                raise JnodeError

            x, y = point_jnode
            if not (int == type(x) and 0 <= x <= 65535):
                eventlog.log_e_param('#/%s/0' % index)
                raise JnodeError

            if not (int == type(y) and 0 <= y <= 65535):
                eventlog.log_e_param('#/%s/1' % index)
                raise JnodeError

            index += 1

        return cls(mask_jnode)

    @classmethod
    def deserialize_list(cls, mask_jnode: List[int]) -> Self:
        """
        :raise JnodeError:
        """
        if type(mask_jnode) != list:
            eventlog.log_e_param('#/')
            raise JnodeError

        it = iter(mask_jnode)
        mask_jnode = list(zip(it, it))

        return cls.deserialize(mask_jnode)


class JsonRectF(RectF):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def deserialize(cls, jnode: List[float]) -> Self:
        if len(jnode) != 4:
            raise JnodeError

        return cls(*jnode)
