import bisect

from typing import Dict, List, Optional, Self
from uuid import uuid4

from edge.model.track import ITrack, LrTrackItem
from edge.model.types import Dir
from edge.types import ImageRef

from edge_bbox.types import RectF, QuadI


def mean(items: List[float | int]) -> float:
    return sum(items) / len(items)


class RaTrackItem:
    def __init__(self, ts: int, x: float, y: float, speed: int, raw_f: int, raw_l: int, direction: Dir):
        self.ts = ts
        self.x = x
        self.y = y
        self.speed = speed
        self.raw_f = raw_f
        self.raw_l = raw_l
        self.dir = direction

    @classmethod
    def deserialize(cls, ra_track_item_jnode: Dict) -> Self:
        ts = ra_track_item_jnode['Ts']
        x, y = ra_track_item_jnode['Xy']
        speed_mps = ra_track_item_jnode['S']
        direction = Dir(ra_track_item_jnode.get('D', Dir.UNKNOWN))
        raw_f = ra_track_item_jnode.get('F', 0)
        raw_l = ra_track_item_jnode.get('L', 0)

        ra_track_item = cls(ts, x, y, speed_mps, raw_f, raw_l, direction)
        return ra_track_item


class LeTrackItem:
    def __init__(self, min_ts: int, max_ts: int):
        self.min_ts = min_ts
        self.max_ts = max_ts

    @property
    def ts(self) -> int:
        return self.min_ts

    @classmethod
    def deserialize(cls, jnode: List[int]) -> Self:
        if len(jnode) != 2:
            raise ValueError

        return cls(*jnode)


class LeTrack(list):
    def __init__(self, trace: List[LeTrackItem], ref: str):
        super().__init__(trace)
        self.ref = ref

    @property
    def min_ts(self) -> int:
        return self[0].ts

    @property
    def max_ts(self) -> int:
        return self[-1].ts

    @classmethod
    def deserialize(cls, jnode: Dict) -> Optional[Self]:
        active_jnode = jnode.get('active', [])
        ref = jnode.get('ref')
        trace = list(map(LeTrackItem.deserialize, active_jnode))
        if len(trace):
            return cls(trace, ref)


class RaTrack(list):
    def __init__(self, length: int, trace: List[RaTrackItem]):
        super().__init__(trace)
        self.length = length

    def get_median_speed_kmph(self) -> Optional[int]:
        speed_list = [item.speed for item in self]

        speed_list_len = len(speed_list)
        if speed_list_len > 0:
            speed_list.sort()
            index = speed_list_len // 2
            return int(speed_list[index])

    @classmethod
    def deserialize(cls, jnode: Dict) -> Self:
        """
        :param jnode:
            TrTrack#/Ra
        """
        length = jnode['Length']
        trace = [RaTrackItem.deserialize(item_jnode) for item_jnode in jnode['Trace']]
        return cls(length, trace)

    def get_direction(self) -> Dir:
        first_item: RaTrackItem = self[0]
        last_item: RaTrackItem = self[-1]
        forward_delta = first_item.raw_f - last_item.raw_f

        if forward_delta > 0:
            return Dir(Dir.APPROACHING)
        elif forward_delta < 0:
            return Dir(Dir.RECEDING)
        else:
            return Dir(Dir.UNKNOWN)


class WaTrack:
    def __init__(self,
                 weight: int,
                 height: int,
                 width: int,
                 length: int,
                 lane: str,
                 event_id: int,
                 event_ts: int,
                 duration: int,
                 provider: str,
                 vrp_front: str,
                 vrp_back: str,
                 vrp_front_w: Optional[int],
                 vrp_back_w: Optional[int],
                 speed: int,
                 axle_count: int,
                 axle_layout: str,
                 vehicle_class: int,
                 road_temperature: int,
                 overview_ref: ImageRef):

        self.weight = weight
        self.height = height
        self.width = width
        self.length = length
        self.lane = lane
        self.event_id = event_id
        self.event_ts = event_ts
        self.duration = duration

        self.provider = provider
        self.vrp_front = vrp_front
        self.vrp_back = vrp_back

        self.vrp_front_w = vrp_front_w
        self.vrp_back_w = vrp_back_w

        self.speed = speed
        self.axle_count = axle_count
        self.axle_layout = axle_layout

        self.vehicle_class = vehicle_class
        self.road_temperature = road_temperature

        self.overview_ref = overview_ref

    @classmethod
    def deserialize(cls, jnode: Dict) -> Self:
        """
        :param jnode:
            TrTrack#/Wa
        """
        weight = jnode['weight']
        height = jnode['size-h']
        width = jnode['size-w']
        length = jnode['size-l']
        lane = jnode['lane']
        event_id = jnode['event-id']
        event_ts = jnode['event-ts']
        duration = jnode['event-dur']

        provider = jnode['vrp-provider']
        vrp_front = jnode['vrp-front']
        vrp_back = jnode['vrp-back']

        vrp_front_w = jnode.get('vrp-front-w')
        vrp_back_w = jnode.get('vrp-back-w')

        speed = jnode['speed']
        axle_count = jnode['axle-count']
        axle_layout = jnode['axle-layout']

        vehicle_class = jnode['class']
        road_temperature = jnode['road-temperature']

        overview_ref = ImageRef(codec='JPEG', url=jnode.get('im-overview'), ts=jnode.get('im-overview-ts'))
        return cls(weight, height, width, length, lane, event_id, event_ts, duration, provider, vrp_front, vrp_back,
                       vrp_front_w, vrp_back_w, speed, axle_count, axle_layout, vehicle_class,
                       road_temperature, overview_ref)


class Plate:
    def __init__(self, validity: int, text: str, template: int, quad: QuadI, country_a2: str):
        self.validity = validity
        self.text = text
        self.template = template
        self.quad = quad
        self.country_a2 = country_a2

    @classmethod
    def deserialize_lr(cls, plate_jnode: Dict) -> Self:
        validity = plate_jnode['V']
        text = plate_jnode['P']
        template = plate_jnode['T']
        quad = QuadI.unpack(plate_jnode['Q'])
        country_a2 = plate_jnode['A2']

        return cls(validity, text, template, quad, country_a2)

    @classmethod
    def deserialize_raas(cls, jnode: Dict) -> Self:
        """
        :param jnode:
            RaasScene#/hypos/:index
        """
        # validity = jnode['weight']
        # text = jnode['number']
        # country = int(jnode['country']) << 16
        # template = country | int(jnode['template'])
        # return Plate(validity, text, template)

        raise NotImplementedError


class RaasSceneItem:
    def __init__(self, variants: List[Plate]):
        self.variants = variants

    @classmethod
    def deserialize(cls, jnode: Dict) -> Self:
        """
        :param jnode:
            RaasScene#/hypos/:index
        """
        plate = Plate.deserialize_raas(jnode)
        return cls([plate])


class RaasScene:
    def __init__(self, items: List[RaasSceneItem]):
        self.items = items

    def try_get_first_scene_item_plate(self) -> Optional[Plate]:
        for item in self.items:
            max_validity_variant = max(item.variants, key=lambda entry: entry.validity)
            return max_validity_variant

    @classmethod
    def deserialize(cls, jnode: Dict) -> Self:
        """
        :param jnode:
            RaasScene#/

        :raise ValueError:
        """
        if 0 != jnode['result'] and 1 != jnode['version']:
            raise ValueError

        items = [RaasSceneItem.deserialize(raas_scene_item_jnode) for raas_scene_item_jnode in jnode['hypos']]

        return cls(items)


class JsonImRef(ImageRef):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def deserialize(cls, jnode: Dict) -> Self:
        blob_ts = jnode['Ts']
        blob_url = jnode.get('Url')
        if blob_url is None or len(blob_url) == 0:
            blob_url = None

        blob_format = jnode['Codec']
        blob_w, blob_h = jnode['Wh']

        return cls(blob_ts, blob_url, blob_format, int(blob_w), int(blob_h))


class LrTrack(ITrack, list):
    def __init__(self,
                 plate: Plate,
                 plate_trace: List[LrTrackItem],
                 image_dir: int,
                 image_ref: ImageRef,
                 lr_alarms_jnode: Optional[List],
                 stat: int,
                 vehicle_type_jnode: Optional[Dict],
                 vehicle_seats_jnode: Optional[Dict],
                 spot_speed: Optional[float]):
        """
        :param plate:
        :param plate_trace:
        :param image_dir:
        :param image_ref:
        :param lr_alarms_jnode:
        :param stat:
        :param vehicle_type_jnode:
            TrTrack#/Lr/Vehicle/type
        :param vehicle_seats_jnode:
            TrTrack#/Lr/Vehicle/seats
        :param spot_speed:
        """

        list.__init__(self, plate_trace)

        self._direction = image_dir
        self.plate = plate
        self.best_image = image_ref
        self.lr_alarms_jnode = lr_alarms_jnode

        # {
        #     "info": [
        #         {
        #             "class": "CAR"
        #         }
        #     ],
        #     "weight": 36,
        #     "make": "HONDA",
        #     "weight-make": 75,
        #     "weight-class": 100,
        #     "model": "ACCORD",
        #     "class": "CAR"
        # }
        if vehicle_type_jnode is None:
            self._vehile_type = None
        else:
            self._vehile_type = dict()
            if vehicle_type_jnode['weight-class'] >= 0.0:
                self._vehile_type['info'] = [{
                    'class':  vehicle_type_jnode['class']
                }]
                self._vehile_type['class'] = vehicle_type_jnode['class']
                self._vehile_type['weight-class'] = vehicle_type_jnode['weight-class']

            if vehicle_type_jnode['weight-make'] >= 0.0:
                self._vehile_type['make'] = vehicle_type_jnode['make']
                self._vehile_type['weight-make'] = vehicle_type_jnode['weight-make']

                if vehicle_type_jnode['weight'] >= 0.0:
                    self._vehile_type['model'] = vehicle_type_jnode['model']
                    self._vehile_type['weight'] = vehicle_type_jnode['weight']

        self._seats_jnode = vehicle_seats_jnode

        self._spot_speed = spot_speed

        # kPLATE_DETECTED (0x01).
        # kPLATE_TRACKING (0x02).
        # kPLATE_LOST_VALID (0x04).
        # kPLATE_LOST_INVALID (0x8).
        self.status = stat

        self._stat_lost = self._lost_bit

        detected_not_lost = self._detected_bit and not self._lost_bit
        lost_not_detected = self._lost_bit and not self._detected_bit
        # self.allow_transit_ = detected_not_lost or lost_not_detected
        self._stat_detected = detected_not_lost or lost_not_detected

    @property
    def _detected_bit(self) -> bool:
        return bool(self.status & 0x01)

    @property
    def _lost_bit(self) -> bool:
        return bool(self.status & 0x04)

    @property
    def stat_detected(self) -> bool:
        return bool(self._stat_detected)

    @property
    def stat_lost(self) -> bool:
        return bool(self._stat_lost)

    @property
    def direction(self) -> int:
        return self._direction

    @property
    def trace(self) -> List[int]:
        tr: List[int] = []
        list(map(lambda item: tr.extend(list(item.plate_quad.center)), self))
        return tr

    @property
    def min_ts(self) -> int:
        return self[0].ts

    @property
    def max_ts(self) -> int:
        return self[-1].ts

    @property
    def type_jnode(self) -> Optional[Dict]:
        return self._vehile_type

    @property
    def seats_jnode(self) -> Optional[Dict]:
        return self._seats_jnode

    @property
    def spot_speed(self) -> Optional[float]:
        return self._spot_speed

    @classmethod
    def deserialize(cls, jnode: Dict, lr_alarms_jnode: Optional[List]) -> Self:
        """
        :param jnode:
            TrTrack#/Lr
        :param lr_alarms_jnode:
        """
        plate = Plate.deserialize_lr(jnode['Plate'])
        plate_dir = jnode['Dir']
        plate_trace = [LrTrackItem.deserialize(item_jnode) for item_jnode in jnode['PlateTrace']]

        lr_track_status = jnode['Stat']

        # todo: find error in tracker/lpr and delete this workaround
        plate_trace.sort(key=lambda item: item.ts, reverse=False)

        image_ref = JsonImRef.deserialize(jnode['Image'])

        vehicle_type_jnode = jnode.get('Vehicle', {}).get('type') or None
        vehicle_seats_jnode = jnode.get('Vehicle', {}).get('seats') or None
        spot_speed = jnode.get('SpotSpeed') or None

        return cls(plate, plate_trace, plate_dir, image_ref, lr_alarms_jnode, lr_track_status, vehicle_type_jnode,
                   vehicle_seats_jnode, spot_speed)

    def get_direction(self) -> Dir:
        # return Dir(self.direction)
        track_half_size = len(self) // 2
        y_list = [x.plate_quad.center[1] for x in self]
        init_mean = mean(y_list[:track_half_size])
        fini_mean = mean(y_list[track_half_size:])

        if init_mean > fini_mean:
            return Dir(Dir.RECEDING)
        elif init_mean < fini_mean:
            return Dir(Dir.APPROACHING)
        else:
            return Dir(Dir.UNKNOWN)


class TrTrackItem:
    def __init__(self, ts: int,
                 lr_item: Optional[LrTrackItem] = None,
                 ra_item: Optional[RaTrackItem] = None,
                 le_item: Optional[LeTrackItem] = None):

        self.ts = ts
        self.lr_item = lr_item
        self.ra_item = ra_item
        self.le_item = le_item

    def __lt__(self, other: Self) -> bool:
        return self.ts < other.ts

    def get_plate_rect(self, lr_scene_w: Optional[int] = None, lr_scene_h: Optional[int] = None) -> RectF:
        # image_ts, fake_radar_plate_rect = best_ra_ts_rect
        if self.lr_item and lr_scene_w and lr_scene_h:
            return RectF(*self.lr_item.plate_quad.get_rel_bbox(lr_scene_w, lr_scene_h))

        elif self.ra_item:
            x = self.ra_item.x
            y = self.ra_item.y
            size = 0.05
            return RectF(x - size, y - size, x + size, y + size)

        elif self.le_item:
            size = 0.05
            return RectF(0.5 - size, 0.5 - size, 0.5 + size, 0.5 + size)

        else:
            raise NotImplementedError

    def get_plate_quad(self, lr_scene_w: Optional[int] = None, lr_scene_h: Optional[int] = None) -> QuadI:
        # image_ts, fake_radar_plate_rect = best_ra_ts_rect
        if self.lr_item:
            return self.lr_item.plate_quad

        elif self.ra_item and lr_scene_w and lr_scene_h:
            fx = self.ra_item.x * lr_scene_w
            fy = self.ra_item.y * lr_scene_h
            size = 0.05 * lr_scene_w

            fl = int(fx - size)
            ft = int(fy - size)
            fr = int(fx + size)
            fb = int(fy + size)
            return QuadI(fl, ft, fr, ft, fr, fb, fl, fb)

        elif self.le_item and lr_scene_w and lr_scene_h:
            fx = 0.5 * lr_scene_w
            fy = 0.5 * lr_scene_h
            size = 0.05 * lr_scene_w

            fl = int(fx - size)
            ft = int(fy - size)
            fr = int(fx + size)
            fb = int(fy + size)
            return QuadI(fl, ft, fr, ft, fr, fb, fl, fb)

        else:
            raise NotImplementedError


class TrTrack(list):
    def __init__(self, trace: List[TrTrackItem]):
        super().__init__(trace)

    @property
    def min_ts(self) -> int:
        return self[0].ts

    @property
    def max_ts(self) -> int:
        return self[-1].ts

    @classmethod
    def create(cls,
               lr_track: Optional[LrTrack],
               ra_track: Optional[RaTrack],
               ledar_track: Optional[LeTrack] = None) -> Self:

        if lr_track and ra_track:
            lr_index = 0
            ra_index = 0
            max_lr_index = len(lr_track)
            max_ra_index = len(ra_track)

            trace = []
            while lr_index != max_lr_index and ra_index != max_ra_index:
                if lr_track[lr_index].ts == ra_track[ra_index].ts:
                    trace.append(TrTrackItem(lr_track[lr_index].ts, lr_track[lr_index], ra_track[ra_index]))
                    lr_index += 1
                    ra_index += 1

                elif lr_track[lr_index].ts < ra_track[ra_index].ts:
                    trace.append(TrTrackItem(lr_track[lr_index].ts, lr_item=lr_track[lr_index]))
                    lr_index += 1

                elif lr_track[lr_index].ts > ra_track[ra_index].ts:
                    trace.append(TrTrackItem(ra_track[ra_index].ts, ra_item=ra_track[ra_index]))
                    ra_index += 1

            if lr_index == max_lr_index:
                while ra_index < max_ra_index:
                    trace.append(TrTrackItem(ra_track[ra_index].ts, ra_item=ra_track[ra_index]))
                    ra_index += 1

            elif ra_index == max_ra_index:
                while lr_index < max_lr_index:
                    trace.append(TrTrackItem(lr_track[lr_index].ts, lr_item=lr_track[lr_index]))
                    lr_index += 1

        elif lr_track:
            trace = [TrTrackItem(item.ts, lr_item=item) for item in lr_track]

        elif ra_track:
            trace = [TrTrackItem(item.ts, ra_item=item) for item in ra_track]

        elif ledar_track:
            trace = [TrTrackItem(item.ts, le_item=item) for item in ledar_track]

        else:
            raise NotImplementedError

        return cls(trace)

    def get_item_direction(self, item: TrTrackItem, max_lookup_depth: float = 2.0) -> Dir:
        """
        :param item:
        :param max_lookup_depth:
            in seconds
        """
        if item is None:
            return Dir(Dir.UNKNOWN)
        # find item index in self
        item_index = bisect.bisect_left(self, item)

        def _lr_direction() -> Dir:
            if item.lr_item is None:
                return Dir(Dir.UNKNOWN)

            prev_center = None

            # find first quad with center not in plate_quad but not far then max_lookup_depth seconds
            plate_quad = item.get_plate_quad()
            time_delta = int(max_lookup_depth * 1e6)
            for index in range(item_index, 0, -1):
                prev_item: TrTrackItem = self[index]

                if item.ts - prev_item.ts > time_delta:
                    break

                if prev_item.lr_item is None:
                    continue

                prev_center_ = prev_item.get_plate_quad().center
                if prev_center_ in plate_quad:
                    continue

                prev_center = prev_center_
                break

            # define direction
            plate_x, plate_y = plate_quad.center

            direction = Dir(Dir.UNKNOWN)
            if prev_center:
                candidate_x, candidate_y = prev_center

                if candidate_y < plate_y:
                    direction = Dir(Dir.APPROACHING)

                else:
                    direction = Dir(Dir.RECEDING)

            if direction == Dir(Dir.UNKNOWN):
                next_center = None
                for index in range(item_index, len(self), 1):
                    next_item = self[index]  # type: TrTrackItem

                    if next_item.ts - item.ts > time_delta:
                        break

                    if next_item.lr_item is None:
                        continue

                    next_center_ = next_item.get_plate_quad().center
                    if next_center_ in plate_quad:
                        continue

                    next_center = next_center_
                    break

                if next_center:
                    candidate_x, candidate_y = next_center

                    if candidate_y < plate_y:
                        direction = Dir(Dir.RECEDING)

                    else:
                        direction = Dir(Dir.APPROACHING)

            return direction

        def _ra_direction() -> Dir:
            if item.ra_item is None:
                return Dir(Dir.UNKNOWN)

            return item.ra_item.dir

        ra_dir = _ra_direction()
        if ra_dir.dir != Dir.UNKNOWN:
            return ra_dir

        return _lr_direction()


class Track:
    def __init__(self,
                 lr_track: Optional[LrTrack],
                 ra_track: Optional[RaTrack],
                 tr_track: TrTrack,
                 le_track: Optional[LeTrack] = None,
                 wa_track: Optional[WaTrack] = None):

        self.uuid = str(uuid4())
        self.lr_track = lr_track
        self.ra_track = ra_track
        self.tr_track = tr_track
        self.le_track = le_track
        self.wa_track = wa_track

    @classmethod
    def deserialize(cls, tr_track_jnode: Dict) -> List[Self]:
        lr_track_jnode = tr_track_jnode.get('Lr')
        lr_alarms_jnode = tr_track_jnode.get('Alarms')
        lr_track = LrTrack.deserialize(lr_track_jnode, lr_alarms_jnode) if lr_track_jnode else None

        wa_track_jnode = tr_track_jnode.get('Wa')
        wa_track = WaTrack.deserialize(wa_track_jnode) if wa_track_jnode else None

        ra_track_jnode = tr_track_jnode.get('Ra')
        ra_track = RaTrack.deserialize(ra_track_jnode) if ra_track_jnode else None

        le_tracks_jnode = tr_track_jnode.get('leddar')
        le_tracks = []
        if le_tracks_jnode:
            for le_track_jnode in le_tracks_jnode:
                le_track = LeTrack.deserialize(le_track_jnode)
                if le_track is not None:
                    le_tracks.append(le_track)
        else:
            le_tracks.append(None)

        # print le_tracks_jnode
        tracks = []
        for le_track in le_tracks:
            tr_track = TrTrack.create(lr_track, ra_track, le_track)
            track = cls(lr_track, ra_track, tr_track, le_track, wa_track)
            tracks.append(track)

        return tracks

    def get_direction(self) -> Dir:
        ra_direction = None
        if self.ra_track is not None:
            ra_direction = self.ra_track.get_direction()

        if ra_direction is not None and ra_direction != Dir(Dir.UNKNOWN):
            return ra_direction

        lr_direction = None
        if self.lr_track is not None:
            lr_direction = self.lr_track.get_direction()

        if lr_direction is not None and lr_direction != Dir(Dir.UNKNOWN):
            return lr_direction

        return Dir(Dir.UNKNOWN)
