from typing import cast, Dict, List, Optional, Self, Tuple
from uuid import uuid4

from edge.image.types import MemBlob, Blob
from edge.model.types import Dir, Origin, Location
from edge.service.tz import ITzService

from edge_bbox.types import RectI, QuadI

from speedcontrol.model.models import WaTrack
from speedcontrol.services.plate_info import IPlateInfoService


class Identity:
    def __init__(self, sender_type: int, sender_id: int):
        self.type = sender_type
        self.id = sender_id

        self._hash = sender_type << 14 + sender_id

    def __str__(self) -> str:
        return '%s.%s' % (self.type, self.id)

    def __hash__(self) -> int:
        return self._hash

    def __eq__(self, other) -> bool:
        return getattr(other, 'type', None) == self.type and \
               getattr(other, 'id', None) == self.id and \
               isinstance(other, self.__class__)


class LicensePlateType:
    def __init__(self, template_ref: int, country_a2: str, plate_info_service: IPlateInfoService):
        self._numeric = plate_info_service.get_country_code(template_ref)
        self._country_a2 = country_a2
        self._draft_jnode = {
            'country': {
                'numeric': self._numeric,
                'alpha-2': self._country_a2
            }
        }

        plate_attrs = plate_info_service.get_plate_attrs(template_ref)
        plate_templ = plate_info_service.get_plate_template(template_ref)
        info_jnode = {}
        if plate_attrs:
            info_jnode['attributes'] = plate_attrs

        if plate_templ:
            info_jnode['template'] = plate_templ

        if info_jnode:
            self._draft_jnode['info'] = [info_jnode]

        self._jnode = None

    @property
    def country_numeric(self) -> int:
        return self._numeric

    @property
    def country_a2(self) -> str:
        return self._country_a2

    def serialize(self) -> Dict:
        if self._jnode is None:
            self._jnode = self._draft_jnode

        return self._jnode


class LicensePlate:
    def __init__(self, validity: int, text: str, plate_type: LicensePlateType,
                 quad: QuadI, wa_track: Optional[WaTrack] = None):

        self.plate_text = text
        self._draft_jnode = {
            'weight': validity,
            'type': plate_type.serialize(),
            'bbox-2': quad.get_abs_bbox(),
            'bbox-4': quad.serialize(),
            'text': {
                'ucode': text
            }
        }

        if wa_track:
            waga_jnode = {
                'provider': wa_track.provider,
                'front-text': wa_track.vrp_front,
                'back-text': wa_track.vrp_back
            }

            if wa_track.vrp_front_w:
                waga_jnode['front-weight'] = wa_track.vrp_front_w

            if wa_track.vrp_back_w:
                waga_jnode['back-weight'] = wa_track.vrp_back_w

            self._draft_jnode['extra'] = [waga_jnode]

        self._plate_type = plate_type
        self._jnode = None

    @property
    def country_numeric(self) -> int:
        return self._plate_type.country_numeric

    @property
    def country_a2(self) -> str:
        return self._plate_type.country_a2

    def serialize(self) -> Dict:
        if self._jnode is None:
            self._jnode = self._draft_jnode

        return self._jnode


class Position:
    def __init__(self, forward: float, left: float):
        self.forward = forward
        self.left = left

    def serialize(self) -> Dict:
        jnode = {
            'forward': self.forward,
            'left': self.left
        }
        return jnode


class Vehicle:
    def __init__(self,
                 plates: List[LicensePlate],
                 length: Optional[float],
                 direction: Dir,
                 direction_alias: Optional[str],
                 position: Optional[Position],
                 type_jnode: Optional[Dict],
                 seats_jnode: Optional[Dict],
                 ra_speed: Optional[int],
                 racam_speed: Optional[int],
                 lpr_speed: Optional[int],
                 wa_track: Optional[WaTrack] = None):

        self.position = position
        self.direction = direction

        self._ra_speed = ra_speed
        self._racam_speed = racam_speed
        self._cam_speed = None
        self._wim_speed = None
        self._control_speed = None
        self._main_speed = None

        self._jnode = None
        self._draft_jnode = {
            'license-plates': [plate.serialize() for plate in plates]
        }

        if wa_track:
            self._draft_jnode['extra'] = [{
                'provider': wa_track.provider,
                'road-temperature': wa_track.road_temperature,
                'lane': wa_track.lane,
                'class': wa_track.vehicle_class,
                'event-id': wa_track.event_id,
                'event-ts': wa_track.event_ts,
                'event-dur': wa_track.duration
            }]

            self._draft_jnode['weight'] = {
                "measured": wa_track.weight,
                "axle-count": wa_track.axle_count,
                "axle-layout": wa_track.axle_layout
            }

        self._type_jnode = type_jnode
        if type_jnode is not None:
            self._draft_jnode['type'] = type_jnode

        if seats_jnode is not None:
            self._draft_jnode['seats'] = seats_jnode

        if direction.dir != Dir.UNKNOWN:
            self._draft_jnode['dir'] = str(direction)

        if direction_alias:
            self._draft_jnode['dir-alias'] = direction_alias

        if ra_speed:
            if ra_speed > 240:
                ra_speed_str = '>240'

            elif ra_speed < 20:
                ra_speed_str = '<20'

            else:
                ra_speed_str = str(ra_speed)

            self._draft_jnode['spot-speed'] = {
                'radar': {
                    'measured': ra_speed,
                    'measured2': ra_speed_str
                },
                'main': {
                    'source': 'radar',
                    'measured': ra_speed
                }
            }

            self._main_speed = ra_speed

        if self._racam_speed:
            from random import randint
            max_error_kmh = 3 if ra_speed <= 40 else int(ra_speed * 0.08)

            self._cam_speed = ra_speed + randint(0, max_error_kmh)
            if self._racam_speed < ra_speed:
                self._racam_speed = ra_speed + randint(0, max_error_kmh)

            self._draft_jnode['spot-speed']['camera'] = {
                'measured': self._cam_speed,
                'measured2': self._racam_speed
            }
            self._control_speed = self._cam_speed

            self._draft_jnode['spot-speed']['control'] = [{
                'source': 'camera',
                'measured': self._cam_speed
            }]

        if lpr_speed:
            spotspeed_jnode = self._draft_jnode.get('spot-speed') or dict()
            camera_jnode = spotspeed_jnode.get('camera') or dict()
            camera_jnode['measured3'] = int(lpr_speed)
            spotspeed_jnode['camera'] = camera_jnode

            self._draft_jnode['spot-speed'] = spotspeed_jnode

        if length:
            self._draft_jnode['size'] = {'l': length}

        if wa_track:
            spotspeed_jnode = self._draft_jnode.get('spot-speed') or dict()
            spotspeed_jnode['wim'] = {
                "measured": wa_track.speed
            }

            self._main_speed = wa_track.speed

            from random import randint
            max_error_kmh = 3 if wa_track.speed <= 40 else int(wa_track.speed * 0.08)

            self._cam_speed = wa_track.speed + randint(0, max_error_kmh)
            self._control_speed = self._cam_speed
            spotspeed_jnode['main'] = {
                'source': 'wim',
                'measured': wa_track.speed
            }

            spotspeed_jnode['camera'] = {
                'measured': self._cam_speed
            }

            spotspeed_jnode['control'] = [{
                'source': 'camera',
                'measured': self._cam_speed
            }]

            self._draft_jnode['spot-speed'] = spotspeed_jnode

            self._draft_jnode['size'] = {
                "w": wa_track.width,
                "h": wa_track.height,
                "l": wa_track.length
            }

        if position:
            self._draft_jnode['position'] = position.serialize()

        self.plate_absent = len(plates) == 0
        self._plates = plates

    @property
    def vehicle_class(self) -> Optional[str]:
        if self._type_jnode:
            return self._type_jnode.get('class')

        return 'UNKNOWN'

    @property
    def vehicle_make(self) -> Optional[str]:
        if self._type_jnode:
            return self._type_jnode.get('make')

    @property
    def vehicle_model(self) -> Optional[str]:
        if self._type_jnode:
            return self._type_jnode.get('model')

    @property
    def plate_text(self) -> Optional[str]:
        if self._plates:
            return self._plates[0].plate_text

    @property
    def plate_county_numeric(self) -> Optional[int]:
        if self._plates:
            return self._plates[0].country_numeric

    @property
    def plate_county_a2(self) -> Optional[str]:
        if self._plates:
            return self._plates[0].country_a2

    @property
    def spot_speed(self) -> Optional[int]:
        return self._ra_speed

    @property
    def camera_speed(self) -> Optional[int]:
        return self._cam_speed

    @property
    def main_speed(self) -> Optional[int]:
        return self._main_speed

    @property
    def control_speed(self) -> Optional[int]:
        return self._control_speed

    @property
    def wim_speed(self) -> Optional[int]:
        return self._wim_speed

    @property
    def ra_cam_speed(self) -> Optional[int]:
        return self._racam_speed

    def serialize(self) -> Dict:
        if self._jnode is None:
            self._jnode = self._draft_jnode

        return self._jnode


class Waypoint:
    NAME_CATCH = 'catch'
    NAME_BEST = 'best'
    NAME_LOST = 'lost'
    NAME_VIOLATION = 'violation'
    NAME_VIOLATION_BEST = 'violation-best'
    NAME_PRE_STOPBAR = 'pre-stopbar'
    NAME_POST_STOPBAR = 'post-stopbar'

    def __init__(self, name: str, ts: int,
                 control_ts: Optional[int] = None,
                 location: Optional[Location] = None,
                 draft_jnode: Optional[Dict] = None):

        self._name = name
        self._ts = ts
        self._control_ts = control_ts
        self._images = []

        self._draft_jnode = draft_jnode or {}
        self._location = location or Location()
        self._jnode: Optional[Dict] = None

    @property
    def name(self) -> str:
        return self._name

    @property
    def ts(self) -> int:
        return self._ts

    def add_images(self, blobs: List[MemBlob]) -> Self:
        self._images.extend(blobs)
        return self

    def get_blobs(self) -> List[MemBlob]:
        return self._images

    def serialize(self, tz_service: ITzService) -> Dict:
        if self._jnode is None:
            self._jnode = {'time': tz_service.get_datetime_isostr(self._ts / 1e6)}
            if self._control_ts is not None:
                self._jnode['control-time'] = tz_service.get_datetime_isostr(self._control_ts / 1e6)

            self._location.serialize(self._draft_jnode)
            self._jnode.update(self._draft_jnode)

        images_jnode = dict()
        for blob in self._images:  # type: MemBlob
            images_jnode[blob.profile] = blob.serialize(tz_service=tz_service)

        if len(images_jnode):
            self._jnode['images'] = images_jnode

        return self._jnode


class BestWaypoint(Waypoint):
    DIR_UNKNOWN = 'unknown'
    DIR_APPROACHING = 'approching'
    DIR_RECEDING = 'receding'

    def __init__(self, ts: int,
                 vehicle: Vehicle,
                 control_ts: Optional[int] = None,
                 direction: str = 'unknown',
                 location: Location = Location()):

        super().__init__(Waypoint.NAME_BEST, ts, control_ts=control_ts, location=location)

        self._images: List[MemBlob] = []
        # self._draft_jnode['vehicle'] = vehicle.serialize()
        self._draft_jnode = {
            'vehicle': vehicle.serialize()
        }

        if direction != 'unknown':
            self._draft_jnode['dir'] = direction

        self.vehicle = vehicle
        self.location = location

        self._jnode: Optional[Dict] = None

    def serialize(self, tz_service: ITzService) -> Dict:
        if self._jnode is None:
            self._jnode = Waypoint.serialize(self, tz_service)
            self._jnode.update(self._draft_jnode)

            self.location.serialize(self._jnode)

        images_jnode = dict()
        images_jnode['license-plates'] = []
        for blob in self._images:
            if blob.profile == 'plate':
                images_jnode['license-plates'].append(blob.serialize(tz_service=tz_service))
            else:
                images_jnode[blob.profile] = blob.serialize(tz_service=tz_service)

        if len(images_jnode):
            self._jnode['images'] = images_jnode

        return self._jnode


class ViolationWaypoint(Waypoint):
    def __init__(self, ts: int,
                 location: Location,
                 control_ts: Optional[int] = None,
                 position: Optional[Position] = None,
                 direction: Optional[Dir] = None,
                 direction_alias: Optional[str] = None,
                 plate_rects: Optional[List[RectI]] = None,
                 plate_quads: Optional[List[QuadI]] = None,
                 w_name: str = Waypoint.NAME_VIOLATION):

        super().__init__(w_name, ts, control_ts=control_ts, location=location)
        plate_rects = plate_rects or list()
        plate_quads = plate_quads or list()

        if position or len(plate_rects) or len(plate_quads):
            self._draft_jnode['vehicle'] = dict()

            if direction and direction.dir != Dir.UNKNOWN:
                self._draft_jnode['vehicle']['dir'] = str(direction)

            if direction_alias:
                self._draft_jnode['vehicle']['dir-alias'] = direction_alias

            if position:
                self._draft_jnode['vehicle']['position'] = position.serialize()

            if len(plate_quads):
                self._draft_jnode['vehicle']['licence-plates'] = [
                    {'bbox4': quad.serialize(), 'bbox2': quad.get_abs_bbox()}
                    for quad in plate_quads
                ]
            elif len(plate_rects):
                self._draft_jnode['vehicle']['licence-plates'] = [{'bbox2': rect.serialize()} for rect in plate_rects]

    def serialize(self, tz_service: ITzService) -> Dict:
        if self._jnode is None:
            self._jnode = Waypoint.serialize(self, tz_service)
            self._jnode.update(self._draft_jnode)

        return self._jnode


class BestViolationWaypoint(ViolationWaypoint):
    def __init__(self, ts,
                 location: Location,
                 control_ts: Optional[int] = None,
                 position: Optional[Position] = None,
                 direction: Optional[Dir] = None,
                 direction_alias: Optional[str] = None,
                 plate_rects: Optional[List[RectI]] = None,
                 plate_quads: Optional[List[QuadI]] = None):

        super().__init__(ts, location, control_ts=control_ts, position=position, direction=direction,
                                   direction_alias=direction_alias, plate_rects=plate_rects,
                                   plate_quads=plate_quads, w_name=Waypoint.NAME_VIOLATION_BEST)


class RedLightViolationWaypoint(Waypoint):
    def __init__(self, name: str, ts: int,
                 location: Location,
                 plate_quads: List[QuadI],
                 green_amber_time: Optional[int],
                 red_time: Optional[int],
                 red_amber_time: Optional[int],
                 halt_time: Optional[int]):
        """
        :param name:
        :param ts:
        :param location:
        :param plate_quads:
        :param green_amber_time:
            value in usec
        :param red_time:
            value in usec
        :param red_amber_time:
            value in usec
        """
        super().__init__(name, ts, location=location)
        self._draft_jnode['vehicle'] = {
            'licence-plates': [{
                'bbox2': quad.get_abs_bbox(),
                'bbox4': quad.serialize()
            } for quad in plate_quads]
        }

        self._green_amber_time = self._red_time = self._red_amber_time = self._halt_time = 0
        if green_amber_time is not None:
            self._draft_jnode['green-amber-time'] = green_amber_time
            self._green_amber_time = green_amber_time

        if red_time is not None:
            self._draft_jnode['red-time'] = red_time
            self._red_time = red_time

        if red_amber_time is not None:
            self._draft_jnode['red-amber-time'] = red_amber_time
            self._red_amber_time = red_amber_time

        if halt_time is not None:
            self._draft_jnode['halt_time'] = halt_time
            self._halt_time = halt_time

    def serialize(self, tz_service: ITzService) -> Dict:
        if self._jnode is None:
            self._jnode = Waypoint.serialize(self, tz_service)
            self._jnode.update(self._draft_jnode)

        return self._jnode

    @property
    def red_time(self) -> Optional[int]:
        return self._red_time

    @property
    def green_amber_time(self) -> Optional[int]:
        return self._green_amber_time

    @property
    def red_amber_time(self) -> Optional[int]:
        return self._red_amber_time

    @property
    def halt_time(self):
        """
        :rtype: int
        """
        return self._halt_time


class Measurement:
    def __init__(self, uuid: str, origin: Optional[Origin] = None, direction: Optional[Dir] = None):
        self._waypoints: List[Waypoint] = []
        self._jnode = {
            'uuid': uuid
        }

        if origin:
            self._jnode['origin'] = origin.serialize()

        if direction:
            self._jnode['vehicle'] = {'dir': str(direction)}

    @property
    def best_waypoint(self) -> Optional[BestWaypoint]:
        bw = [point for point in self._waypoints if point.name == point.NAME_BEST]
        return bw[0] if len(bw) == 1 else None

    @property
    def waypoints(self) -> List[Waypoint]:
        return self._waypoints

    def add_waypoints(self, waypoints: List[Waypoint]) -> Self:
        self._waypoints.extend(waypoints)
        return self

    def get_blobs(self) -> List[MemBlob]:
        blobs: List[MemBlob] = []
        list(map(lambda waypoint: blobs.extend(waypoint.get_blobs()), self._waypoints))

        return blobs

    def serialize(self, tz_service: ITzService) -> Dict:
        waypoints_jnode = dict()
        for waypoint in self._waypoints:
            # from edge import eventlog
            # eventlog.log_i("{} - {}".format(waypoint.name, waypoint._draft_jnode))
            waypoints_jnode[waypoint.name] = waypoint.serialize(tz_service)

        self._jnode['waypoints'] = waypoints_jnode

        return self._jnode


class BaseEvent:
    def __init__(self, measurements: List[Measurement], alarms_jnode: Optional[List[Dict]]):
        self._measurements = list(measurements)
        self.alarms_jnode = alarms_jnode

    @property
    def measurements(self) -> List[Measurement]:
        return self._measurements

    @property
    def best_wp(self) -> Optional[BestWaypoint]:
        waypoints_list = self._measurements[0].waypoints
        best_points = [waypoint for waypoint in waypoints_list if waypoint.name == waypoint.NAME_BEST]
        if len(best_points) != 1:
            return None

        return cast(BestWaypoint, best_points[0])

    def get_blobs(self) -> List[MemBlob]:
        blobs: List[MemBlob] = []
        list(map(lambda measurement: blobs.extend(measurement.get_blobs()), self._measurements))

        return blobs

    def get_plate_text(self) -> Optional[str]:
        waypoints = self.measurements[0].waypoints
        best_waypoint_list = [waypoint for waypoint in waypoints if waypoint.name == waypoint.NAME_BEST]

        best_waypoint = cast(BestWaypoint, best_waypoint_list[0])
        return best_waypoint.vehicle.plate_text

    def serialize(self, tz_service: ITzService) -> Dict:
        if len(self._measurements) != 1:
            raise NotImplementedError

        jnode = {'measurements': [self._measurements[0].serialize(tz_service)]}
        if self.alarms_jnode:
            jnode['alarms'] = self.alarms_jnode

        return jnode


class Evidence:
    """
    :type _jnode: dict | None
    :type _waypoints: list[Waypoint]
    :type _videos: list[FileBlob]
    """

    NAME_TRANSIT = 'transit'
    NAME_PARKING = 'parking'
    NAME_RED_LIGHT = 'red-light'
    NAME_WRONG_WAY = 'wrong-way'
    NAME_BUS_LANE = 'bus-lane'
    NAME_SPOT_SPEED = 'spot-speed'
    NAME_WANTED = 'wanted'
    NAME_SECTION_SPEED = 'section-speed'

    def __init__(self, name: str, origin: Origin, ts: int, control_ts: Optional[int] = None):
        self._ts = ts
        self._control_ts = control_ts

        self._draft_jnode = {
            'type': name,
            'origin': origin.serialize()
        }
        self._origin = origin
        self._type = name

        self._waypoints = []
        self._jnode = None
        self._videos = []

    @property
    def origin(self) -> Origin:
        return self._origin

    @property
    def type(self) -> str:
        return self._type

    @property
    def ts(self) -> int:
        return self._ts

    def add_waypoints(self, waypoints: List[Waypoint]) -> None:
        self._waypoints.extend(waypoints)

    def add_videos(self, videos: List[Blob]) -> None:
        self._videos.extend(videos)

    @property
    def videos(self) -> List[Blob]:
        return self._videos

    def get_blobs(self) -> List[Blob]:
        blobs = []
        blobs.extend(self._videos)

        for waypoint in self._waypoints:
            blobs.extend(waypoint.get_blobs())
        return blobs

    @property
    def waypoints(self) -> List[Waypoint]:
        return self._waypoints

    def serialize(self, tz_service: ITzService) -> Dict:
        if self._jnode is None:
            self._draft_jnode['time'] = tz_service.get_datetime_isostr(self._ts / 1e6)
            if self._control_ts is not None:
                self._draft_jnode['control-time'] = tz_service.get_datetime_isostr(self._control_ts / 1e6)

            self._jnode = self._draft_jnode

        return self._jnode


class Event:
    TYPE_TRANSIT = Evidence.NAME_TRANSIT
    TYPE_PARKING = Evidence.NAME_PARKING
    TYPE_RED_LIGHT = Evidence.NAME_RED_LIGHT
    TYPE_WRONG_WAY = Evidence.NAME_WRONG_WAY
    TYPE_BUS_LANE = Evidence.NAME_BUS_LANE
    TYPE_SPOT_SPEED = Evidence.NAME_SPOT_SPEED
    TYPE_WANTED = Evidence.NAME_WANTED
    TYPE_SECTION_SPEED = Evidence.NAME_SECTION_SPEED

    def __init__(self, sender: Identity, base: BaseEvent, evidence: Evidence):
        self._uuid = str(uuid4())
        self._sender = sender
        self._base = base
        self._evidence = evidence

    @property
    def uuid(self) -> str:
        return self._uuid

    @property
    def base(self) -> BaseEvent:
        return self._base

    @property
    def evidence(self) -> Evidence:
        return self._evidence

    @property
    def type(self) -> str:
        return self._evidence.type

    @property
    def sender(self) -> Identity:
        return self._sender

    @property
    def ts_usec(self) -> int:
        return self._evidence.ts // 1000

    @property
    def ts(self) -> int:
        return self._evidence.ts

    def get_blobs(self) -> List[Blob]:
        blobs = self._base.get_blobs()
        blobs.extend(self._evidence.get_blobs())
        return blobs

    def serialize(self, tz_service: ITzService) -> Dict:
        jnode = {
            'uuid': self._uuid
        }

        jnode.update(self._base.serialize(tz_service))
        jnode.update(self._evidence.serialize(tz_service))

        return jnode


class Ticket:
    def __init__(self, name: str):
        self.name = name


class Mask(list):
    def __init__(self, points: List[Tuple[int, int]]):
        super().__init__(points)

    def normalize(self, image_w: int, image_h: int) -> None:
        raise NotImplementedError


class SceneArgs:
    def __init__(self, mask_args: Mask, mask_wh: Tuple[int, int]):
        self.mask = mask_args
        self.mask_wh = mask_wh
