from typing import cast, Union

from edge import eventlog
from edge.com import bs
from edge.image.types import Blob, FileBlob, MemBlob, StructBlobMeta

from speedcontrol.model.types import Identity
from speedcontrol.services.types import BsServiceArgs


class IBsService:
    def __init__(self):
        pass

    def get_size(self, sender: Identity, blob_id: str, profile: str) -> int:
        raise NotImplementedError

    def save(self, sender: Identity, blob: Blob) -> None:
        """
        :raise bs.BsError:
        """
        raise NotImplementedError


class _BsService(IBsService):
    def __init__(self, twisted_reactor, bs_conf: BsServiceArgs, counters: bs.IBsCounters):
        """
        :param twisted_reactor:
            twisted.internet.interfaces.IReactorUNIX
        :param bs_conf:
        :param counters:
        """
        IBsService.__init__(self)
        self._reactor = twisted_reactor
        self._session_args = 'speedcontrol.{}'.format(bs_conf.module_id), counters,\
                             65535, bs.LOG_LEVEL_NAME_TO_CODE_MAP.get(bs_conf.log_level, 'info')

        self._session = bs.Session(counters=counters, closed=True)

        self._counters = counters
        self._sender_to_chan_map = bs_conf.sender_to_chan_map

    def destroy(self) -> None:
        self._session.close()

    def get_size(self, sender: Identity, blob_id: str, profile: str) -> int:
        chan_or_profiles = self._sender_to_chan_map[sender]
        if dict == type(chan_or_profiles):
            chan = chan_or_profiles[profile]
        else:
            chan = chan_or_profiles

        reply = self._session.get_blob_size(chan, blob_id)
        return reply

    def save(self, sender: Identity, blob: Union[FileBlob, MemBlob]) -> None:
        chan_or_profiles = self._sender_to_chan_map[sender]
        if dict == type(chan_or_profiles):
            chan = chan_or_profiles[blob.profile]
        else:
            chan = chan_or_profiles

        self._counters.inc_blob(chan)
        self._save(chan, blob)

    @bs.reopen_session
    def _save(self, chan: str, blob: Union[FileBlob, MemBlob]) -> None:
        """
        :type chan: str
        :type blob: MemBlob | FileBlob

        :raise bs.BsError:
        :raise bs.BsSessionError:
        """
        try:
            blob_ts = blob.min_ts // 1000
            if blob.file_path is None:
                if isinstance(blob.meta, StructBlobMeta):
                    ident = self._session.save_blob(chan, blob_ts, blob.file_data, blob.meta.struct)
                else:
                    ident = self._session.save_blob(chan, blob_ts, blob.file_data, blob.file_meta)
            else:
                ident = self._session.save_blob2(chan, blob_ts, blob.file_path, blob.file_meta)

        except bs.BsError:
            eventlog.log_e('fail: save <chan={}, profile={}>'.format(str(chan), blob.profile))
            raise

        blob.name = '%s_%s.%s' % (chan, ident, blob.file_ext)


def init_service(twisted_reactor, args: BsServiceArgs, counters: bs.IBsCounters) -> IBsService:
    """
    :param twisted_reactor:
        twisted.internet.interfaces.IReactorUNIX
    :param args:
    :param counters:
    """
    return _BsService(twisted_reactor, args, counters)


def fini_service(service: IBsService):
    self = cast(_BsService, service)
    self.destroy()
