import datetime
import os
import time

from typing import Dict, List, Optional, Self

from edge.com.db import DbArgs
from edge import eventlog
from edge.types import JnodeError

from speedcontrol.detectors.bus_lane.types import DetectorArgs as BusLaneArgs
from speedcontrol.detectors.parking.types import DetectorArgs as ParkingArgs
from speedcontrol.detectors.red_light.types import DetectorArgs as RedLightArgs
from speedcontrol.detectors.red_light_leddar.types import DetectorArgs as RailroadCrossingArgs
from speedcontrol.detectors.section_speed.types import DetectorArgs as SectionSpeedArgs
from speedcontrol.detectors.spot_speed.types import DetectorArgs as SpotSpeedArgs
from speedcontrol.detectors.spot_speed_waga.types import DetectorArgs as SpotSpeedWagaArgs
from speedcontrol.detectors.transit.types import DetectorArgs as TransitArgs
from speedcontrol.detectors.wanted.types import DetectorArgs as WantedArgs
from speedcontrol.detectors.wrong_way.types import DetectorArgs as WrongWayArgs
from speedcontrol.model.jnode.types import JsonMask
from speedcontrol.model.types import Identity
from speedcontrol.services.types import (
    BsServiceArgs, RaasServiceArgs, ReactServiceArgs,
    EventServiceArgs, EventQueueArgs, TrainerServiceArgs,
    StatServiceArgs, LaneDetectorServiceArgs, TlServiceArgs,
    DbServiceArgs, SystestServiceArgs, SignServiceArgs,
    LeddarServiceArgs, PlateInfoServiceConf)


class JsonRaasServiceArgs(RaasServiceArgs):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def deserialize(cls, raas_jnode: Optional[Dict]) -> Optional[Self]:
        """
        :raise JnodeError:
        """
        if raas_jnode is None:
            return None

        # deserialize #/host
        raas_host = raas_jnode.get('host')
        if not (type(raas_host) in [str] and 1 <= len(raas_host) <= 16):
            eventlog.log_e_param('#/host')
            raise JnodeError

        # deserialize #/port
        raas_port = raas_jnode.get('port')
        if not (type(raas_port) in [str] and 1 <= len(raas_port) <= 16):
            eventlog.log_e_param('#/port')
            raise JnodeError

        # deserialize #/user
        raas_user = raas_jnode.get('user')
        if not (type(raas_user) in [str] and 1 <= len(raas_user) <= 16):
            eventlog.log_e_param('#/user')
            raise JnodeError

        # deserialize #/pass
        raas_pass = raas_jnode.get('pass')
        if not (type(raas_pass) in [str] and 1 <= len(raas_pass) <= 16):
            eventlog.log_e_param('#/pass')
            raise JnodeError

        return cls(raas_host, raas_port, raas_user, raas_pass)


class JsonEventServiceArgs(EventServiceArgs):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def deserialize(cls, fact_event_jnode: Dict) -> Self:
        max_file_count = fact_event_jnode.get('max-file-count', 15)
        if not (int == type(max_file_count) and 15 <= max_file_count <= 50):
            raise ValueError('main.conf#/event-store/max-file-count')

        json_queue_args_list = [EventQueueArgs(endpoint, max_file_count - 5, max_file_count)
                                for endpoint in fact_event_jnode['endpoints']]

        ua_json_required = fact_event_jnode.get('args-ua', False)
        require_plate = fact_event_jnode.get('require-plate', True)

        local_arch_base_dir = fact_event_jnode.get('local-arch-base', None)
        return cls(require_plate, json_queue_args_list, ua_json_required, local_arch_base_dir)


class JsonReactServiceArgs(ReactServiceArgs):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def deserialize(cls, main_conf_jnode: Dict) -> Self:
        """
        :raise JnodeError:
        """
        track_react_endpoint = main_conf_jnode['services']['react']
        return cls(track_react_endpoint)


class JsonBsServiceArgs(BsServiceArgs):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def deserialize(cls,
                    module_id: int,
                    log_level: str,
                    transit_args: TransitArgs,
                    wrong_way_args: List[WrongWayArgs],
                    spot_speed_args: List[SpotSpeedArgs],
                    spot_speed_waga_args: List[SpotSpeedWagaArgs],
                    red_light_args: List[RedLightArgs],
                    railroad_crossing_args: List[RailroadCrossingArgs],
                    parking_args: List[ParkingArgs],
                    wanted_args: List[WantedArgs],
                    section_speed_args: List[SectionSpeedArgs],
                    bus_lane_args: List[BusLaneArgs]) -> Self:

        from speedcontrol.detectors.types import IDetector

        sender_to_chan_map = {
            Identity(IDetector.TYPE_TRANSIT, 0): {
                'args': transit_args.blob_store_args.vehicle_store,
                'plate': transit_args.blob_store_args.vehicle_store,
                'vehicle': transit_args.blob_store_args.vehicle_store,
                'thumb': transit_args.blob_store_args.vehicle_store,
                'scene': transit_args.blob_store_args.scene_store,
                'overview': transit_args.blob_store_args.scene_store
            }
        }

        for args in wrong_way_args:
            sender_to_chan_map.update({args.identity: args.store})

        for args in spot_speed_args:
            sender_to_chan_map.update({args.identity: args.store})

        for args in spot_speed_waga_args:
            sender_to_chan_map.update({args.identity: args.store})

        for args in parking_args:
            sender_to_chan_map.update({args.identity: args.store})

        for args in red_light_args:
            sender_to_chan_map.update({args.identity: args.store})

        for args in railroad_crossing_args:
            sender_to_chan_map.update({args.identity: args.store})

        for args in wanted_args:
            sender_to_chan_map.update({args.identity: args.store})

        for args in section_speed_args:
            sender_to_chan_map.update({args.identity: args.store})

        for args in bus_lane_args:
            sender_to_chan_map.update({args.identity: args.store})

        return cls(module_id, sender_to_chan_map, log_level)


class JsonTrainerServiceArgs(TrainerServiceArgs):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def deserialize(cls, trainer: Dict) -> Self:
        """
        :raise JnodeError:
        """
        try:
            db_directory = trainer['directory']
            min_df = trainer.get('min-df', 0.1)
            time_min_max = trainer['limits'].get('time-min-max', [])
            lanes = trainer['limits'].get('lanes', [])

            min_time = datetime.time(hour=0, minute=0, second=0)
            max_time = datetime.time(hour=23, minute=59, second=59)
            if len(time_min_max) == 2:
                raw_min_time, raw_max_time = time_min_max
                min_struct_time = time.strptime(raw_min_time, "%H:%M")
                max_struct_time = time.strptime(raw_max_time, "%H:%M")

                min_time = datetime.time(min_struct_time.tm_hour, min_struct_time.tm_min)
                max_time = datetime.time(max_struct_time.tm_hour, max_struct_time.tm_min)

            lr = trainer['limits'].get('lr', JsonTrainerServiceArgs.LR_ANY)

            return cls(db_directory, lanes, min_time, max_time, min_df, lr)
        except KeyError:
            eventlog.log_d('fail: plate_absent_db jnode have invalid keys')

        raise JnodeError


class JsonStatServiceArgs(StatServiceArgs):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def deserialize(cls, monitoring_jnode: Dict) -> Self:
        """
        :raise JnodeError:
        """
        stat_filepath = monitoring_jnode.get('path')
        if not (type(stat_filepath) in [str] and len(stat_filepath) < 64 and os.path.isabs(stat_filepath)):
            eventlog.log_e_param('#/path')
            raise JnodeError

        stat_tout_sec = monitoring_jnode.get('tout')
        if not (int == type(stat_tout_sec) and 0 <= stat_tout_sec <= 255):
            eventlog.log_e_param('#/tout')
            raise JnodeError

        return cls(stat_filepath, stat_tout_sec)


class JsonLaneDetectorServiceArgs(LaneDetectorServiceArgs):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def deserialize(cls, raw_road_jnode: Dict) -> Self:
        """
        :raise JnodeError:
        """
        if not len(raw_road_jnode):
            eventlog.log_d('warning: no road.conf, making fake LaneDetector args')
            return JsonLaneDetectorServiceArgs([], (14, 14), 'approaching', 'receding', 'unknown')

        road_jnode = raw_road_jnode.get('road')
        if not type(road_jnode) in [dict]:
            eventlog.log_e_param('#/road')
            raise JnodeError

        # deserialize #/mask/resolution
        resolution_jnode = raw_road_jnode.get('scene-wh')
        if not resolution_jnode or len(resolution_jnode) != 2:
            eventlog.log_e_param('#/scene-wh')
            raise JnodeError

        im_width, im_height = resolution_jnode
        if not (type(im_width) in [int] and 0 <= im_width <= 65535):
            eventlog.log_e_param('#/mask/scene-wh/0')
            raise JnodeError

        if not (type(im_height) in [int] and 0 <= im_height <= 65535):
            eventlog.log_e_param('#/mask/scene-wh/1')
            raise JnodeError

        # deserialize #/mask/lanes
        lanes_jnode = road_jnode.get('lanes', [])
        if list != type(lanes_jnode):
            eventlog.log_e_param('#/mask/lanes')
            raise JnodeError

        id_to_lane_list = [(str(lane_jnode['name']), JsonMask.deserialize(lane_jnode['mask']))
                           for lane_jnode in lanes_jnode]

        road_jnode = raw_road_jnode.get('road') or dict()
        dir_jnode = road_jnode.get('dir') or dict()
        approaching = dir_jnode.get('approaching') or 'approaching'
        receding = dir_jnode.get('receding') or 'receding'
        unknown = dir_jnode.get('unknown') or 'unknown'

        return cls(id_to_lane_list, im_width, im_height, approaching, receding, unknown)


class JsonTlServiceArgs(TlServiceArgs):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def deserialize(cls, jnode: Optional[Dict]) -> Self:
        """
        :raise JnodeError:
        """
        if jnode is None:
            return None

        if len(jnode) == 3:
            red_only, endpoint, tl_ident = jnode
            return cls(endpoint=endpoint, tl_ident=tl_ident, red_only=red_only)


class JsonLeddarServiceArgs(LeddarServiceArgs):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def deserialize(cls, jnode: Optional[str]) -> Optional[Self]:
        """
        :raise JnodeError:
        """
        if jnode is None:
            return None

        return cls(endpoint=jnode, ref='loop-rl')


class JsonDbServiceArgs(DbServiceArgs):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def deserialize(cls, origin_id: int, jnode: Dict) -> Self:
        """
        :raise JnodeError:
        """
        db_jnode = jnode.get('db')
        if not db_jnode:
            eventlog.log_e('fail: param #/db is absent')
            raise JnodeError

        db_args = DbArgs.deserialize(db_jnode)
        return cls(origin_id, db_args)


class JsonSystestServiceArgs(SystestServiceArgs):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def deserialize(cls, main_conf_jnode: Dict) -> Self:
        """
        :raise JnodeError:
        """
        systest_conf_filepath = main_conf_jnode['services'].get('systest')
        return cls(systest_conf_filepath)


class JsonSignServiceArgs(SignServiceArgs):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def deserialize(cls, sign_jnode: Dict, log_level: str) -> Self:
        """
        :param sign_jnode:
            main.conf#/services/sign
        :param log_level:
            none | error | warn | info | debug | trace

        :rtype: JsonSignServiceArgs
        :raises JnodeError:
        """
        provider_jnode = sign_jnode.get('provider')
        if provider_jnode is None:
            eventlog.log_e_param('#/provider')
            raise JnodeError

        iit_conf = provider_jnode.get('iit')
        if type(iit_conf) not in [str]:
            eventlog.log_e_param('#/iit')
            raise JnodeError

        return cls(iit_conf, log_level)


class JsonPlateInfoServiceConf(PlateInfoServiceConf):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def deserialize(cls, main_conf_jnode: Dict) -> Self:
        """
        :raise JnodeError:
        """
        plate_info_path = main_conf_jnode['services'].get('plate-info')
        return cls(plate_info_path)
