from typing import Dict, List, Optional, Self

from edge import eventlog
from edge.model.types import TimeSpan
from edge.types import JnodeError

from speedcontrol.services.jnode.types import JsonLeddarServiceArgs


class ILeddarService:
    def ref(self) -> Optional[str]:
        return None

    def check(self, min_ts: int, max_ts: int) -> List[TimeSpan]:
        raise NotImplementedError

    def destroy(self) -> None:
        pass


class _StubService(ILeddarService):
    def __init__(self):
        ILeddarService.__init__(self)

    def check(self, min_ts: int, max_ts: int) -> List[TimeSpan]:
        return []


class LeddarLog:
    def __init__(self, time: TimeSpan, active: List[TimeSpan]):
        self.active = active
        self.time = time

    @classmethod
    def deserialize(cls, jnode: Dict, leddar_ref: str) -> Self:
        """
        :raise JnodeError:
        """
        ts_jnode = jnode.get('ts')

        if type(ts_jnode) != list or len(ts_jnode) != 2:
            eventlog.log_e_param('#/ts')
            raise JnodeError

        time_interval = TimeSpan(*ts_jnode)

        leddar_list_jnode = jnode.get('leddar')
        if type(leddar_list_jnode) != list:
            eventlog.log_e_param('#/leddar')
            raise JnodeError

        for leddar_jnode in leddar_list_jnode:
            if leddar_jnode['ref'] != leddar_ref:
                continue

            active_jnode = leddar_jnode['active']
            active_intervals = list(map(TimeSpan.create, active_jnode))
            return cls(time_interval, active_intervals)

        raise JnodeError

    def check(self, time_span: TimeSpan) -> List[TimeSpan]:
        spans = [x for x in self.active if x.check_overlap(time_span)]
        return spans


class _LeddarService(ILeddarService):
    def __init__(self, args: JsonLeddarServiceArgs):
        ILeddarService.__init__(self)

        self._leddar_log: Optional[LeddarLog] = None
        self._leddar_ref = args.ref

    def ref(self) -> str:
        return self._leddar_ref

    def update(self, react_jnode: Dict) -> None:
        # noinspection PyBroadException
        try:
            self._leddar_log = LeddarLog.deserialize(react_jnode, self._leddar_ref)
        except Exception:
            eventlog.log_x()
            return

    def check(self, min_ts: int, max_ts: int) -> List[TimeSpan]:
        if self._leddar_log is None:
            return []

        time_interval = TimeSpan(min_ts, max_ts)
        if not self._leddar_log.time.check_overlap(time_interval):
            # there is no stats for [cts ,lts]
            return []

        violation_list = self._leddar_log.check(time_interval)
        return violation_list


def init_service(args: Optional[JsonLeddarServiceArgs]) -> ILeddarService:
    if args is None:
        rl = _StubService()
    else:
        rl = _LeddarService(args)
    return rl


def fini_service(service: ILeddarService) -> None:
    service.destroy()
