from typing import Optional

from edge import eventlog
from edge.com import fs

from speedcontrol.services.types import PlateInfoServiceConf


class PlateTemplateError(RuntimeError):
    pass


class IPlateInfoService:
    def get_country_code(self, template_ref: int) -> int:
        raise NotImplementedError

    def get_plate_attrs(self, template_ref: int) -> Optional[str]:
        raise NotImplementedError

    def get_plate_template(self, template_ref: int) -> Optional[str]:
        raise NotImplementedError


class _PlateInfoService(IPlateInfoService):
    def __init__(self, conf: PlateInfoServiceConf):
        IPlateInfoService.__init__(self)
        if conf.info_path:
            self._ref_to_info_map = fs.load_json3(conf.info_path)
        else:
            self._ref_to_info_map = dict()

    def get_country_code(self, template_ref: int) -> int:
        return (template_ref >> 20) & 0xFFF

    def get_plate_attrs(self, template_ref: int) -> Optional[str]:
        plate_info = self._ref_to_info_map.get(str(template_ref)) or dict()
        if not plate_info:
            eventlog.log_w('fail: unsupported (template-ref:{})'.format(template_ref))

        return plate_info.get('attrs')

    def get_plate_template(self, template_ref: int) -> Optional[str]:
        plate_info = self._ref_to_info_map.get(str(template_ref)) or dict()
        if not plate_info:
            eventlog.log_w('fail: unsupported (template-ref:{})'.format(template_ref))

        return plate_info.get('template')


def init_service(conf: PlateInfoServiceConf) -> IPlateInfoService:
    return _PlateInfoService(conf)


# noinspection PyUnusedLocal
def fini_service(service: IPlateInfoService):
    pass
