import ujson

from PIL import Image
from io import BytesIO
from typing import Optional, Tuple

from twisted.internet.defer import Deferred, succeed
from twisted.internet.protocol import Protocol, connectionDone
from twisted.python.failure import Failure
from twisted.web.client import Agent, Response, FileBodyProducer
from twisted.web.http_headers import Headers

from edge import eventlog
from edge.image.image import ImHelpers, ImRgba

from edge_bbox.types import RectF

from speedcontrol.services.types import RaasServiceArgs
from speedcontrol.model.models import JsonImRef, RaasScene


class RaasError(RuntimeError):
    pass


class IRaasService:
    def analyze_scene(self, scene_ref: JsonImRef, rect: RectF) -> Deferred:
        """
        :returns:
            callbacks to RaasScene; errbacks to RaasError
        :raise RaasError:
        """
        raise NotImplementedError

    def analyze_scene2(self, rect: RectF, blob: ImRgba) -> Deferred:
        """
        :return:
            callbacks to RaasScene; errbacks to RaasError
        :raise RaasError:
        """
        raise NotImplementedError


class _StubService(IRaasService):
    # noinspection PyUnusedLocal
    def __init__(self, *args, **kwargs):
        IRaasService.__init__(self)
        self._scene = RaasScene(items=[])

    def analyze_scene(self, *args, **kwargs) -> Deferred:
        """
        :returns:
            callbacks to RaasScene
        """
        return succeed(self._scene)

    def analyze_scene2(self, rect: RectF, blob: ImRgba) -> Deferred:
        """
        :returns:
            callbacks to RaasScene;
            errbacks to RaasError;
        :raise RaasError:
        """
        return succeed(self._scene)


class _ResponseBodyProto(Protocol):
    def __init__(self, handler: Deferred):
        """
        :param handler:
            Deferred - callbacks to str
        """
        self._handler = handler
        self._body_buffer = ''

    def dataReceived(self, body_data: str) -> None:
        self._body_buffer += body_data

    def connectionLost(self, reason: Failure = connectionDone) -> None:
        self._handler.callback(self._body_buffer[: -1])  # strip '\0'


class _Service(IRaasService):
    def __init__(self, twisted_reactor, args: RaasServiceArgs):
        IRaasService.__init__(self)

        self._agent = Agent(twisted_reactor)

        self._base_url = 'http://%s:%s/alpr/%s/' % (args.raas_host, args.raas_port, args.raas_user)
        self._password = args.raas_pass

    @staticmethod
    def _upscale_plate_rect(plate_rect: RectF, scene_w: int, scene_h: int) -> Tuple[int, int, int, int]:
        px_margin = int(scene_h * (plate_rect.fb - plate_rect.ft))

        px_l = -px_margin + int(scene_w * plate_rect.fl)
        px_t = -px_margin + int(scene_h * plate_rect.ft)
        px_r = +px_margin + int(scene_w * plate_rect.fr)
        px_b = +px_margin + int(scene_h * plate_rect.fb)
        return px_l, px_t, px_r, px_b

    @staticmethod
    def _mk_signature(plate_image_bytes_io: BytesIO, secret: bytes) -> str:
        from hashlib import md5

        md5_hash = md5(plate_image_bytes_io.getvalue())
        md5_hash.update(BytesIO(secret).getvalue())
        signature = md5_hash.hexdigest()

        return signature

    def analyze_scene(self, scene_ref: JsonImRef, rect: RectF) -> Deferred:
        """
        :returns:
            callbacks to RaasScene;
            errbacks to RaasError;
        :raise RaasError:
        """
        if scene_ref.codec != 'JPEG':
            raise NotImplementedError

        scene_image = Image.open(scene_ref.url)
        vehicle_ltrb = self._upscale_plate_rect(rect, *scene_image.size)
        vehicle_image = scene_image.crop(vehicle_ltrb)  # TODO: crop via stride

        mem_file = BytesIO()
        vehicle_image.save(mem_file, 'JPEG', quality=90)
        plate_image_signature = self._mk_signature(mem_file, self._password)

        mem_file.seek(0)
        qbody_producer = FileBodyProducer(mem_file)

        call_chain = self._agent.request(
            'POST', self._base_url + plate_image_signature,
            Headers({'User-Agent': ['EDGE'], 'Content-Type': ['image/jpeg']}),
            qbody_producer
        ).addCallbacks(
            callback=self._handle_reply_http_header_done,
            errback=self._handle_reply_http_header_fail
        ).addCallbacks(
            callback=self._handle_reply_http_body_done,
            errback=self._handle_reply_http_body_fail)

        return call_chain

    def analyze_scene2(self, rect: RectF, blob: ImRgba) -> Deferred:
        """
        :returns:
            callbacks to RaasScene;
            errbacks to RaasError;
        :raise RaasError:
        """
        vehicle_ltrb = self._upscale_plate_rect(rect, *blob.wh)
        vehicle_rectf = RectF(*vehicle_ltrb)
        vehicle_image = ImHelpers.mk_crop_resize_image(blob, vehicle_rectf, vehicle_rectf.fw, vehicle_rectf.fh)

        mem_file = BytesIO()
        vehicle_image.save(mem_file, 'JPEG', quality=90)
        plate_image_signature = self._mk_signature(mem_file, self._password)

        mem_file.seek(0)
        qbody_producer = FileBodyProducer(mem_file)

        call_chain = self._agent.request(
            'POST', self._base_url + plate_image_signature,
            Headers({'User-Agent': ['EDGE'], 'Content-Type': ['image/jpeg']}),
            qbody_producer
        ).addCallbacks(
            callback=self._handle_reply_http_header_done,
            errback=self._handle_reply_http_header_fail
        ).addCallbacks(
            callback=self._handle_reply_http_body_done,
            errback=self._handle_reply_http_body_fail)

        return call_chain

    @staticmethod
    def _handle_reply_http_header_done(response: Response) -> Deferred[str]:
        """
        :raise RaasError:
        """
        if response.code != 200:
            eventlog.log_d('fail: invalid-reply (http-code=%s)' % response.code)
            raise RaasError()

        call_chain = Deferred()
        response.deliverBody(_ResponseBodyProto(call_chain))
        return call_chain

    @staticmethod
    def _handle_reply_http_header_fail(failure: Failure) -> None:
        """
        :raise RaasError:
        """
        eventlog.log_d('fail: invalid-query (%s)' % str(failure))
        raise RaasError

    @staticmethod
    def _handle_reply_http_body_done(response_body: str) -> RaasScene:
        raas_scene_jnode = ujson.loads(response_body)
        raas_scene = RaasScene.deserialize(raas_scene_jnode)
        return raas_scene

    @staticmethod
    def _handle_reply_http_body_fail(failure: Failure) -> None:
        """
        :raise RaasError:
        """
        if not failure.check([RaasError]):
            eventlog.log_d('fail: load response-query (%s)' % str(failure))

        raise RaasError


def init_service(twisted_reactor, args: Optional[RaasServiceArgs]) -> IRaasService:
    if args is None:
        raas = _StubService()
    else:
        raas = _Service(twisted_reactor, args)

    return raas


# noinspection PyUnusedLocal
def fini_service(service: IRaasService):
    pass
