import datetime
import struct
import time
import ujson

from typing import Callable, Optional

from edge.com.fs import silent_mkdir_r
from edge.image.types import MemBlob
from edge.queue import ImageFileTree
from edge.service.tz import ITzService

from edge_bbox.types import RectF

from speedcontrol.model.models import Track
from speedcontrol.model.types import BaseEvent, BestWaypoint
from speedcontrol.services.types import TrainerServiceArgs


class ITrainerService:
    def collect(self, event_args: BaseEvent,
                get_blob_fn: Callable[[], MemBlob],
                ra_rect: RectF,
                track: Track
                ) -> None:

        raise NotImplementedError


class _StubTrainerService(ITrainerService):
    def __init__(self):
        ITrainerService.__init__(self)

    def collect(self, *args, **kwargs) -> None:
        pass


class _TrainerService(ITrainerService):
    def __init__(self, twisted_reactor, args: TrainerServiceArgs, tz_service: ITzService):
        ITrainerService.__init__(self)
        self._root_dirpath = args.root_dir
        silent_mkdir_r(self._root_dirpath)

        self._lr_limit = args.lr

        self._tz_service = tz_service
        self._lanes = args.lanes
        self._min_time = args.min_time
        self._max_time = args.max_time
        self._jpeg_queue = ImageFileTree(
            twisted_reactor, args.root_dir, args.min_df)

    def collect(self, event_args: BaseEvent,
                get_blob_fn: Callable[[], MemBlob],
                ra_rect: RectF,
                track: Track
                ) -> None:

        if track.lr_track and self._lr_limit in [TrainerServiceArgs.LR_ABSENT]:
            return
        elif track.lr_track is None and self._lr_limit in [TrainerServiceArgs.LR_PRESRNT]:
            return

        blob = get_blob_fn()
        # define best-waypoint
        best_waypoint = filter(
            lambda point: point.name == point.NAME_BEST,
            event_args.measurements[0].waypoints)[0]  # type: BestWaypoint

        #
        if not self._lanes or best_waypoint.location.lane in self._lanes:
            image_struct_time = time.gmtime(blob.min_ts / 1000000)
            image_time = datetime.time(image_struct_time.tm_hour, image_struct_time.tm_min)

            if self._min_time < image_time < self._max_time:
                # add event jnode to the end of jpeg
                event_jnode = event_args.serialize(self._tz_service)
                event_str = ujson.dumps(event_jnode)
                # add event jnode size after jnode
                event_size = struct.pack('I', len(event_str))

                self._jpeg_queue.add_blob_image(
                    blob_name=self._tz_service.get_datetime_isostr(blob.min_ts / 1e6),
                    blob_bytes=blob.file_data + event_str + event_size, blob_ts=blob.min_ts)


def init_service(twisted_reactor, args: Optional[TrainerServiceArgs], tz_service: ITzService) -> ITrainerService:
    return _TrainerService(twisted_reactor, args, tz_service) if args else _StubTrainerService()


# noinspection PyUnusedLocal
def fini_service(service: ITrainerService):
    pass
