from collections import defaultdict
from os.path import isabs
from typing import Dict, List, Optional, Self, Set

from edge import eventlog
from edge.landmark.jnode.types import LandmarkServiceArgs, JsonLandmarkServiceArgs


class SceneReactArgs:
    def __init__(self, mc_dir: str, mg_dir: str, lr_dir: str, ra_dir: str, tout_msec: str):
        self.mc_dir = mc_dir
        self.mg_dir = mg_dir
        self.lr_dir = lr_dir
        self.ra_dir = ra_dir
        self.tout_msec = tout_msec

    @classmethod
    def deserialize(cls, scene_react_jnode: Dict) -> Self:
        mc_dir = scene_react_jnode.get('McDir', '')
        mg_dir = scene_react_jnode.get('MgDir', '')
        lr_dir = scene_react_jnode.get('LrDir', '')
        ra_dir = scene_react_jnode.get('RaDir', '')
        tout = scene_react_jnode['Tout']

        if not mc_dir and not mg_dir:
            raise ValueError('At least one of mc_dir or mg_dir should be defined')

        if not lr_dir:
            raise ValueError('The lr_dir should be defined')

        args = cls(mc_dir, mg_dir, lr_dir, ra_dir, tout)
        return args


class TrackReactArgs:
    def __init__(self, react_dir: str):
        self.react_dir = react_dir

    @classmethod
    def deserialize(cls, track_react_jnode: Dict) -> Self:
        react_dir = track_react_jnode['Dir']
        args = TrackReactArgs(react_dir)
        return args


class SceneEventArgs:
    def __init__(self, event_dirs: Set[str], skip_rate: int):
        self.event_dirs = event_dirs
        self.skip_rate = skip_rate

    @classmethod
    def deserialize(cls, event_dirs: Set[str], skip_rate: int) -> Self:
        args = cls(event_dirs, skip_rate)
        return args


class ImageQueueArgs:
    def __init__(self, endpoint: str, min_file_count: int, max_file_count: int):
        self.endpoint = endpoint
        self.min_file_count = min_file_count
        self.max_file_count = max_file_count
        self.log_level = 'info'

    @classmethod
    def deserialize(cls, store_jnode: Dict) -> Self:
        endpoint = store_jnode.get('endpoint')
        if not (str == type(endpoint) and len(endpoint) < 64 and isabs(endpoint)):
            raise ValueError('main.conf#/store/endpoint')

        max_file_count = store_jnode.get('max-file-count', 15)
        if not (int == type(max_file_count) and 15 <= max_file_count <= 50):
            raise ValueError('main.conf#/store/max-file-count')

        return ImageQueueArgs(endpoint, max_file_count - 5, max_file_count)


class TrackEventArgs:
    def __init__(self, debounce_tout: int, event_dirs: List[str]):
        self.debounce_tout = debounce_tout
        self.event_dirs = event_dirs

    @classmethod
    def deserialize(cls, track_event_jnode: Dict) -> Self:
        debounce_tout = track_event_jnode.get('debounce-tout', 3)
        endpoints = track_event_jnode.get('endpoints', [])

        if not (0 <= debounce_tout <= 30):
            raise ValueError('main.conf#/fact-event/debounce-tout')
        args = cls(debounce_tout=debounce_tout, event_dirs=endpoints)
        return args


class LrUpscaleArgs:
    def __init__(self, left: float, top: float, right: float, bottom: float):
        self.left = left
        self.top = top
        self.right = right
        self.bottom = bottom

        self._jnode = [left, top, right, bottom]

    def serialize(self) -> List[float]:
        return self._jnode


class RaUpscaleArgs:
    def __init__(self, left: float, top: float, right: float, bottom: float):
        self.left = left
        self.top = top
        self.right = right
        self.bottom = bottom

        self._jnode = [left, top, right, bottom]

    @classmethod
    def deserialize(cls, jnode: List[float]) -> Self:
        if len(jnode) != 4:
            raise ValueError('main.conf#/ra-vehicle-ltrb')

        return cls(*jnode)

    def serialize(self) -> List[float]:
        return self._jnode


class StatCountersArgs:
    def __init__(self, stat_filepath: str, stat_tout_sec: int):
        self.stat_filepath = stat_filepath
        self.stat_tout_sec = stat_tout_sec

    @classmethod
    def deserialize(cls, monitoring_jnode: Dict) -> Self:
        stat_filepath = monitoring_jnode.get('StatFile')
        if not (str == type(stat_filepath) and len(stat_filepath) < 64 and isabs(stat_filepath)):
            raise ValueError('main.conf#/Monitoring/StatFile')

        raw_stat_tout_sec = monitoring_jnode.get('StatTout')
        if not (int == type(raw_stat_tout_sec) and 0 <= raw_stat_tout_sec <= 255):
            raise ValueError('main.conf#/Monitoring/StatTout')

        args = cls(stat_filepath, raw_stat_tout_sec)
        return args


class KernelArgs:
    DEF_LR_RA_OID_MATCH_STORE_TOUT_SEC = 5
    DEF_RA_TRACK_STORE_TOUT_SEC = 5
    DEF_MIN_RATRACK_LEN = 5

    def __init__(self, stat_counters_args: StatCountersArgs,
                 scene_react_args: SceneReactArgs,
                 track_react_args: TrackReactArgs,
                 scene_event_args_list: List[SceneEventArgs],
                 track_event_args: TrackEventArgs,
                 lr_ra_oid_match_store_tout: int,
                 ra_track_store_tout: int,
                 min_ra_track_len: int,
                 image_store_args: Optional[ImageQueueArgs],
                 systest: bool = False,
                 plate_detected_mode: bool = False,
                 ra_upscale_args: Optional[RaUpscaleArgs] = None,
                 landmark_args: Optional[LandmarkServiceArgs] = None):

        self.stat_counters_args = stat_counters_args

        self.scene_react_args = scene_react_args
        self.track_react_args = track_react_args

        self.scene_event_args_list = scene_event_args_list
        self.track_event_args = track_event_args

        self.lr_ra_oid_match_store_tout = lr_ra_oid_match_store_tout
        self.ra_track_store_tout = ra_track_store_tout
        self.min_ra_track_len = min_ra_track_len

        self.image_store_args = image_store_args
        self.systest = systest
        self.plate_detected_mode = plate_detected_mode
        self.ra_upscale_args = ra_upscale_args
        self.landmark_args = landmark_args

    @classmethod
    def deserialize(cls, main_conf_jnode: Dict, misc_conf_jnode: Dict) -> Self:
        # define stat-counters-args
        monitoring_jnode = main_conf_jnode.get('Monitoring')
        if not monitoring_jnode:
            raise ValueError('main.conf#/Monitoring')

        stat_counters_args = StatCountersArgs.deserialize(monitoring_jnode)

        # define others
        file_api_jnode = main_conf_jnode['FileApi']
        track_event_args = TrackEventArgs.deserialize(file_api_jnode['TrackEvent'])
        scene_react_args = SceneReactArgs.deserialize(file_api_jnode['AddSceneReact'])
        track_react_args = TrackReactArgs.deserialize(file_api_jnode['AddTrackReact'])
        image_store_jnode = main_conf_jnode.get('store')
        image_store_args = ImageQueueArgs.deserialize(image_store_jnode) if image_store_jnode else None

        skip_rate_to_dir_set_map = defaultdict(set)
        for entry in file_api_jnode['SceneEvent']:
            skip_rate_to_dir_set_map[entry["SkipRate"]].add(entry["Dir"])

        scene_event_args_list = [SceneEventArgs.deserialize(skip_rate_event_dirs[1], skip_rate_event_dirs[0])
                                 for skip_rate_event_dirs in list(skip_rate_to_dir_set_map.items())]

        lr_ra_oid_match_store_tout = misc_conf_jnode.get(
            'lpr_track_caching_time', KernelArgs.DEF_LR_RA_OID_MATCH_STORE_TOUT_SEC)

        ra_track_store_tout = misc_conf_jnode.get(
            'radar_track_caching_time', KernelArgs.DEF_RA_TRACK_STORE_TOUT_SEC)

        min_ra_track_len = misc_conf_jnode.get(
            'min_valid_radar_track_length', KernelArgs.DEF_MIN_RATRACK_LEN)

        systest = main_conf_jnode.get('systest', False)
        plate_detected_mode = main_conf_jnode.get('plate-detected-mode', False)

        ra_upscale_jnode = main_conf_jnode.get('ra-vehicle-ltrb')
        ra_upscale_args = None
        if ra_upscale_jnode:
            ra_upscale_args = RaUpscaleArgs.deserialize(ra_upscale_jnode)

        services_jnode = main_conf_jnode.get('services') or dict()
        landmark_args = None
        if ra_upscale_args is not None:
            cam_path = services_jnode.get('camera')
            if cam_path is None:
                eventlog.log_e_param('#/services/camera')
                raise ValueError()
            location_path = services_jnode.get('location')
            if location_path is None:
                eventlog.log_e_param('#/services/location')
                raise ValueError()

            # radar terminal
            landmark_args = JsonLandmarkServiceArgs.deserialize(cam_path, location_path)

        kernel_args = cls(
            stat_counters_args, scene_react_args, track_react_args, scene_event_args_list, track_event_args,
            lr_ra_oid_match_store_tout, ra_track_store_tout, min_ra_track_len, image_store_args,
            systest, plate_detected_mode, ra_upscale_args, landmark_args)

        return kernel_args
