import sys
import ujson

from typing import Dict, Self, Tuple, Optional, List

from edge import eventlog
from edge.types import ImageRef
from edge.model.track import LrTrackItem

from tracker._args import LrUpscaleArgs


DIR_UNKNOWN = 1
DIR_APPROACHING = 11
DIR_RECEDING = 12


class RaSceneItemImg:
    def __init__(self, x: float, y: float):
        self.x = x
        self.y = y

    @classmethod
    def deserialize(cls, jnode: Tuple[int, int]) -> Self:
        """
        :param jnode: RaScene#/scene/<index>/img-lt
        """
        x, y = jnode
        return cls(float(x), float(y))


class RaSceneItem:
    def __init__(self, ra_oid: int, ts_ts: int, ra_len: int, ra_speed: int, ra_dir: int,
                 img: RaSceneItemImg, ra_stat: int, ra_f: int, ra_l: int):

        self.oid = ra_oid
        self.ts = ts_ts
        self.len = ra_len
        self.speed = ra_speed
        self.direction = ra_dir
        self.img = img
        self.stat = ra_stat
        self.raw_f = ra_f
        self.raw_l = ra_l

    @classmethod
    def deserialize(cls, jnode: Dict, scene_ts: int) -> Optional[Self]:
        """
        :param jnode:
            RaScene#/scene/<index>
        :param scene_ts
        """
        img = RaSceneItemImg.deserialize(jnode['img-lt'])
        if not (0.0 < img.x < 1.0 and 0.0 < img.y < 1.0):
            return None

        ra_oid = jnode['oid']
        ra_stat = int(jnode['stat'])
        ra_len = int(jnode['len'])
        ra_speed = int(jnode['kmh'])
        ra_dir = int(jnode['dir'])
        ra_f = jnode.get('f', 0)  # cm
        ra_l = jnode.get('l', 0)  # cm

        return cls(ra_oid, scene_ts, ra_len, ra_speed, ra_dir, img, ra_stat, ra_f, ra_l)

    def serialize(self) -> Dict:
        """
        Serializes track-item
        """
        return {
            'Ts': self.ts,
            'Xy': [self.img.x, self.img.y],
            'S': self.speed,
            'D': self.direction,
            'F': self.raw_f,
            'L': self.raw_l
        }


class RaScene:
    _SCENE_VALID = 1
    _SCENE_INVALID = 2

    def __init__(self, ra_items: List[RaSceneItem], status: int):
        self.scene = ra_items
        self.status = status

    @classmethod
    def deserialize(cls, jnode: Dict) -> Self:
        """
        :param jnode: RaScene#/
        :raise ValueError:
        """
        if not jnode:
            raise ValueError

        strobe_ts = int(jnode['strobe-ts'])
        status = int(jnode['stat'])

        if status == RaScene._SCENE_VALID:
            all_scene_items = [RaSceneItem.deserialize(item_jnode, strobe_ts) for item_jnode in jnode['scene']]

            ra_items = [_f for _f in all_scene_items if _f]

        else:
            ra_items = []

        return cls(ra_items, status)


class RaTrack:
    """
    :type trace: list[RaSceneItem]
    """
    def __init__(self, ra_scene_item: RaSceneItem):
        self.catch_ts = sys.maxsize
        self.lost_ts = 0
        self.trace = []
        self.max_length = 0

        self.oid = ra_scene_item.oid
        self.append(ra_scene_item)

    @classmethod
    def create(cls, ra_scene_items: List[RaSceneItem]) -> Self:
        ra_track = RaTrack(ra_scene_items[0])
        for index in range(1, len(ra_scene_items), 1):
            ra_track.append(ra_scene_items[index])

        return ra_track

    def __len__(self):
        return len(self.trace)

    def append(self, ra_item: RaSceneItem) -> None:
        if ra_item.ts < self.catch_ts:
            self.catch_ts = ra_item.ts

        if ra_item.ts > self.lost_ts:
            self.lost_ts = ra_item.ts

        if ra_item.len > self.max_length:
            self.max_length = ra_item.len

        self.trace.append(ra_item)


class LrQuad:
    def __init__(self, x0: int, y0: int, x1: int, y1: int, x2: int, y2: int, x3: int, y3: int):
        self.x0 = x0
        self.y0 = y0

        self.x1 = x1
        self.y1 = y1

        self.x2 = x2
        self.y2 = y2

        self.x3 = x3
        self.y3 = y3

    def upscale_bbox_f(self, lr_upscale: LrUpscaleArgs, scene_w: int, scene_h: int):
        """
        :rtype: LrRect
        """
        # define self bbox
        bbox_il = min(self.x0, self.x3)
        bbox_ir = max(self.x1, self.x2)

        bbox_it = min(self.y0, self.y1)
        bbox_ib = max(self.y2, self.y3)

        bbox_iw = bbox_ir - bbox_il
        bbox_ih = bbox_ib - bbox_it

        # define self center
        center_x = (bbox_il + bbox_ir) / 2.0
        center_y = (bbox_it + bbox_ib) / 2.0

        # scale self bbox
        scaled_fl = float(center_x - (bbox_iw * lr_upscale.left)) / scene_w
        scaled_fr = float(center_x + (bbox_iw * lr_upscale.right)) / scene_w

        scaled_ft = float(center_y - (bbox_ih * lr_upscale.top)) / scene_h
        scaled_fb = float(center_y + (bbox_ih * lr_upscale.bottom)) / scene_h

        return LrRect(scaled_fl, scaled_ft, scaled_fr, scaled_fb)

    @property
    def center(self) -> Tuple[int, int]:
        return int((self.x0 + self.x1 + self.x2 + self.x3) / 4), int((self.y0 + self.y1 + self.y2 + self.y3) / 4)

    @classmethod
    def deserialize(cls, quad_jnode: List[float]) -> Self:
        return cls(*quad_jnode)

    @staticmethod
    def serialize(lr_quad: Self) -> List[int]:
        return [lr_quad.x0, lr_quad.y0, lr_quad.x1, lr_quad.y1,
                lr_quad.x2, lr_quad.y2, lr_quad.x3, lr_quad.y3]


class LrRect:
    LEFT_INDEX = 0
    TOP_INDEX = 1
    RIGHT_INDEX = 2
    BOTTOM_INDEX = 3

    def __init__(self, left: float, top: float, right: float, bottom: float):
        self.left = left
        self.top = top
        self.right = right
        self.bottom = bottom

    def upscale(self, lr_upscale: LrUpscaleArgs) -> Self:
        fw = self.right - self.left
        fh = self.bottom - self.top

        center_x = (self.left + self.right) / 2.0
        center_y = (self.top + self.bottom) / 2.0

        scaled_fl = center_x - (fw * lr_upscale.left)
        scaled_fr = center_x + (fw * lr_upscale.right)

        scaled_ft = center_y - (fh * lr_upscale.top)
        scaled_fb = center_y + (fh * lr_upscale.bottom)

        return LrRect(scaled_fl, scaled_ft, scaled_fr, scaled_fb)

    @classmethod
    def deserialize(cls, jnode: List[float]) -> Self:
        fl = float(jnode[LrRect.LEFT_INDEX])
        ft = float(jnode[LrRect.TOP_INDEX])
        fr = float(jnode[LrRect.RIGHT_INDEX])
        fb = float(jnode[LrRect.BOTTOM_INDEX])

        return cls(fl, ft, fr, fb)

    @staticmethod
    def serialize(lr_rect: Self) -> List[float]:
        return [lr_rect.left, lr_rect.top, lr_rect.right, lr_rect.bottom]


def check_point_in_rect(point_x: float, point_y: float, lr_rect: LrRect) -> bool:
    return lr_rect.left < point_x < lr_rect.right and lr_rect.top < point_y < lr_rect.bottom


def check_point_in_rect2(point_x: float, point_y: float, ltrb: Tuple[int, int, int, int]) -> bool:
    return ltrb[0] < point_x < ltrb[2] and ltrb[1] < point_y < ltrb[3]


class LrPlate:
    def __init__(self, plate_text: str, quad: LrQuad):
        self.plate_text = plate_text
        self.quad = quad

    @classmethod
    def deserialize(cls, lr_plate_jnode: Dict) -> Self:
        plate_text = lr_plate_jnode['P']
        quad = LrQuad.deserialize(lr_plate_jnode['Q'])

        lr_plate = cls(plate_text, quad)
        return lr_plate


class LrSceneItem:
    def __init__(self, oid: int, direction: int, best_variant: LrPlate, history: List, scene_ts: int):
        self.oid = oid
        self.direction = direction
        self.best_variant = best_variant
        self.history = history

        self.ts = scene_ts

    @classmethod
    def deserialize(cls, lr_scene_item_jnode: Dict, scene_ts: int) -> Self:
        oid = int(lr_scene_item_jnode['Id'])
        direction = lr_scene_item_jnode['Dir']
        plate = LrPlate.deserialize(lr_scene_item_jnode['Plate'])
        plate_trace = lr_scene_item_jnode['Trace']

        lr_scene_item = cls(oid, direction, plate, plate_trace, scene_ts)
        return lr_scene_item


class LrScene:
    def __init__(self, source_id: Tuple[int, int], scene: List[LrSceneItem], im_ref: ImageRef, stat: int):
        self.source_id = source_id
        self.scene = scene
        self.im_ref = im_ref
        self.stat = stat

    @classmethod
    def deserialize(cls, jnode: Dict) -> Self:
        """
        :param jnode: LrScene#/

        :raise ValueError:
        """
        stat = jnode['Stat']
        source_jnode = jnode['Source']

        im_ref_jnode = jnode.get('Image')
        if im_ref_jnode:
            im_ref = JsonImRef.deserialize_lr_scene(im_ref_jnode)
        else:
            eventlog.log_d("fail: image-jnode is absent")
            raise ValueError

        lr_items_jnode = jnode['Lr']
        lr_items = [LrSceneItem.deserialize(item_jnode, im_ref.ts) for item_jnode in lr_items_jnode]

        lr_scene = cls(source_jnode, lr_items, im_ref, stat)
        return lr_scene


class LrTrack:
    STAT_UNKNOWN = 0
    STAT_DETECTED = 1
    STAT_LOST = 2

    def __init__(self, source_id: Tuple[int, int], oids: List[int], src_jnode: Dict, waga_jnode: Dict,
                 plate_text: str, plate_trace: List[LrTrackItem], alarms_jnode: Optional[List], status: int):

        self.source_id = source_id
        self.oids = oids
        self.src_jnode = src_jnode
        self.waga_jnode = waga_jnode
        self.plate_text = plate_text
        self.plate_trace = plate_trace
        self.alarms_jnode = alarms_jnode
        self.status = status

    @classmethod
    def deserialize(cls, jnode: Dict) -> Self:
        """
        :param jnode: LrTrack#/
        """
        lr_jnode = jnode['Lr']
        alarms_jnode = jnode.get('Alarms')
        source = lr_jnode['Source']
        oids = lr_jnode['Id']
        plate_text = lr_jnode['Plate']['P']
        stat = lr_jnode.get('Stat', LrTrack.STAT_UNKNOWN)

        detected = stat == 1
        lost_bit_set = bool(stat & 0x04)

        if detected:
            status = LrTrack.STAT_DETECTED
        elif lost_bit_set:
            status = LrTrack.STAT_LOST
        else:
            status = LrTrack.STAT_UNKNOWN

        plate_trace = [LrTrackItem.deserialize(item_jnode) for item_jnode in lr_jnode['PlateTrace']]

        lr_track = cls(source, oids, lr_jnode, jnode.get('Wa'), plate_text, plate_trace, alarms_jnode, status)
        return lr_track


class JsonImRef(ImageRef):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def deserialize_mg_scene(cls, jnode: Dict) -> Self:
        """
        :param jnode: MgScene#/
        """
        blob_ts = int(jnode['Env']['ExpStartTs'])

        ref_jnode = jnode['Frame']
        blob_url = ref_jnode['Url']
        blob_format = ref_jnode['Codec']
        blob_w = ref_jnode['W']
        blob_h = ref_jnode['H']

        return cls(blob_ts, blob_url, blob_format, blob_w, blob_h)

    @classmethod
    def deserialize_mc_scene(cls, jnode: Dict) -> Self:
        """
        :param jnode: McScene#/
        :raise ValueError:
        """
        stat_done = 1
        # stat_fail = 2
        if jnode['Status'] != stat_done:
            raise ValueError('Failed to deserialize mc_scene')

        blob_ts = int(jnode['Ts'])
        blob_url = jnode['Url']
        blob_format = jnode['Codec']
        blob_w, blob_h = jnode['SizeWh']

        return cls(blob_ts, blob_url, blob_format, blob_w, blob_h)

    @classmethod
    def deserialize_lr_scene(cls, jnode: Dict) -> Self:
        """
        :param jnode: LrScene#/image
        """
        blob_ts = int(jnode['Ts'])
        blob_url = jnode['Url']
        blob_format = jnode['Codec']
        blob_w, blob_h = jnode['Wh']
        return cls(blob_ts, blob_url, blob_format, blob_w, blob_h)

    @classmethod
    def deserialize_lr_track(cls, jnode: Dict) -> Self:
        """
        :param jnode: LrTrack#/Image
        """
        blob_ts = jnode['Ts']
        blob_url = jnode.get('Url')
        if blob_url is None or len(blob_url) == 0:
            blob_url = None

        blob_format = jnode['Codec']
        blob_w, blob_h = jnode['Wh']

        return cls(blob_ts, blob_url, blob_format, int(blob_w), int(blob_h))


class TrSceneItem:
    def __init__(self, lr_item: Optional[LrSceneItem], ra_item: Optional[RaSceneItem],
                 match_rect: Optional[Tuple[int, int, int, int]]):
        if not lr_item and not ra_item:
            raise ValueError

        self.lr_item = lr_item
        self.ra_item = ra_item
        self.match_rect = match_rect


class TrScene:
    def __init__(self, tr_items: List[TrSceneItem], lr_items: List[TrSceneItem], ra_items: List[TrSceneItem]):
        self.tr_items = tr_items  # list of matched lr_scene_item and ra_scene_item
        self.lr_items = lr_items  # list of not matched lr_scene_items
        self.ra_items = ra_items  # list of not matched ra_scene_items

    @staticmethod
    def create(tr_items: List[TrSceneItem], lr_items: List[TrSceneItem], ra_items: List[TrSceneItem]):
        return TrScene(tr_items, lr_items, ra_items)

    @classmethod
    def create_lr(cls,  lr_items: List[TrSceneItem]) -> Self:
        tr_scene = cls([], lr_items, [])
        return tr_scene

    @classmethod
    def create_ra(cls, ra_items: List[TrSceneItem]) -> Self:
        tr_scene = cls([], [], ra_items)
        return tr_scene

    @classmethod
    def create_empty(cls) -> Self:
        tr_scene = cls([], [], [])
        return tr_scene

    @staticmethod
    def _serialize_lr_item(tr_item: TrSceneItem) -> Dict:
        lr_jnode = {
            'Id': tr_item.lr_item.oid,
            'Dir': tr_item.lr_item.direction,
            'Plate': {
                'P': tr_item.lr_item.best_variant.plate_text,
                'Q': tr_item.lr_item.best_variant.quad.serialize(tr_item.lr_item.best_variant.quad)
            },
            'Trace': tr_item.lr_item.history
        }

        return lr_jnode

    @staticmethod
    def _serialize_ra_item(tr_item: TrSceneItem) -> Dict:
        ra_item = tr_item.ra_item
        ra_jnode = {
            'oid': ra_item.oid,
            'len': ra_item.len,
            'speed': ra_item.speed,
            'dir': ra_item.direction,
            'xy': [ra_item.img.x, ra_item.img.y],
            'r': tr_item.match_rect
        }

        return ra_jnode

    def serialize(self, im_ref: ImageRef) -> str:
        tr_scene_jnode = {
            'Image': {
                'Ts': im_ref.ts,
                'Url': im_ref.url,
                'Codec': im_ref.codec,
                'Wh': [im_ref.width, im_ref.height]
            },
            'LrRa': [{
                    'Lr': TrScene._serialize_lr_item(tr_item),
                    'Ra': TrScene._serialize_ra_item(tr_item)
                } for tr_item in self.tr_items],
            'Lr': list(map(TrScene._serialize_lr_item, self.lr_items)),
            'Ra': list(map(TrScene._serialize_ra_item, self.ra_items))
        }

        #
        # https://pypi.python.org/pypi/ujson
        #
        # ensure_ascii limits output to ASCII and escapes all extended characters above 127. Default is true.
        # If your end format supports UTF-8 setting this option to false is highly recommended to save space.
        tr_scene_utf8 = ujson.dumps(tr_scene_jnode, ensure_ascii=False)
        return tr_scene_utf8


class TrTrack:
    def __init__(self, lr_track: Optional[LrTrack], ra_track: Optional[RaTrack]):
        if not lr_track and not ra_track:
            raise ValueError

        self.lr_track = lr_track
        self.ra_track = ra_track

    def serialize(self, im_ref: Optional[ImageRef]) -> str:
        tr_track_jnode = dict()

        lr_track_jnode = self.lr_track.src_jnode if self.lr_track else None
        if lr_track_jnode:
            if im_ref and im_ref.url:
                lr_track_jnode['Image']['Url'] = im_ref.url

            tr_track_jnode['Lr'] = lr_track_jnode

            if self.lr_track.alarms_jnode:
                tr_track_jnode['Alarms'] = self.lr_track.alarms_jnode

        waga_track_jnode = self.lr_track.waga_jnode if self.lr_track else None
        if waga_track_jnode:
            tr_track_jnode['Wa'] = waga_track_jnode

        ra_track = self.ra_track
        if ra_track:
            tr_track_jnode['Ra'] = {
                'CatchTs': ra_track.catch_ts,
                'LostTs': ra_track.lost_ts,
                'Length': ra_track.max_length,
                'Trace': [x.serialize() for x in ra_track.trace]
            }

        tr_track_utf8 = ujson.dumps(tr_track_jnode, ensure_ascii=False)
        return tr_track_utf8
