import os

from copy import deepcopy
from errno import ENOENT
from typing import Dict, Optional

from edge import eventlog
from edge.com import fs
from edge.queue import ImageFileQueue
from edge.types import ImageRef

from tracker._args import SceneEventArgs, TrackEventArgs, ImageQueueArgs
from tracker._models import JsonImRef, TrTrack, TrScene, LrTrack
from tracker.services.stats import TrackEventCounters, SceneEventCounters


class _ToutEntry:
    def __init__(self, track: TrTrack, im_ref: ImageRef, rm_delayed_call):
        """
        :param rm_delayed_call: twisted.internet.interfaces.IDelayedCall
        """
        self.tr_track = track
        self.image_ref = im_ref
        self.rm_delayed_call = rm_delayed_call


class ITrackEventConsumer:
    def consume_track(self, tr_track: TrTrack, plate_text: Optional[str] = None) -> None:
        raise NotImplementedError


class ISceneEventConsumer:
    def consume_scene(self, im_ref: ImageRef, tr_scene: TrScene) -> None:
        raise NotImplementedError


class TrackEventConsumer(ITrackEventConsumer):
    def __init__(self, twisted_reactor,
                 track_event_args: TrackEventArgs,
                 stat_counters: TrackEventCounters,
                 im_store_args: Optional[ImageQueueArgs]):

        ITrackEventConsumer.__init__(self)

        self._stat_counters = stat_counters

        self._event_dirs = track_event_args.event_dirs
        for event_dirpath in self._event_dirs:
            fs.silent_mkdir_r(event_dirpath)

        self._event_file_prefix = os.getpid()
        self._event_file_seq = 0
        self._plate_text_to_tout_entry_map: Dict[str, _ToutEntry] = {}
        self._debounce_tout: int = track_event_args.debounce_tout
        self._twisted_reactor = twisted_reactor

        if im_store_args:
            self._im_store = ImageFileQueue(
                twisted_reactor, None, im_store_args.endpoint,
                im_store_args.min_file_count,
                im_store_args.max_file_count,
                im_store_args.log_level)
        else:
            self._im_store = None

    def _handle_track_debounce_tout(self, plate_text: str) -> None:
        tout_entry: Optional[_ToutEntry] = self._plate_text_to_tout_entry_map.get(plate_text, None)
        if tout_entry is None:
            return

        try:
            self._consume_track(tout_entry.tr_track, tout_entry.image_ref)
        finally:
            self._plate_text_to_tout_entry_map.pop(plate_text, None)

    def consume_track(self, next_tr_track: TrTrack, plate_text: Optional[str] = None) -> None:
        # tr-track doesn't contain the lr-track
        if plate_text is None:
            self._consume_track(next_tr_track, None)
            return

        next_lr_track: LrTrack = next_tr_track.lr_track
        prev_entry: Optional[_ToutEntry] = self._plate_text_to_tout_entry_map.get(plate_text, None)

        # TODO: refactor debouncer
        if prev_entry is None:
            rm_delayed_call = self._twisted_reactor.callLater(
                self._debounce_tout, self._handle_track_debounce_tout, plate_text)

            im_jnode = next_lr_track.src_jnode['Image']
            im_ref = JsonImRef.deserialize_lr_track(im_jnode)
            if im_ref.url is not None and self._im_store:
                store_img_ref = self._im_store.add_image_ref(im_ref)
                im_ref.url = store_img_ref.url

            if plate_text is not None:
                self._plate_text_to_tout_entry_map[plate_text] = _ToutEntry(next_tr_track, im_ref, rm_delayed_call)

        else:
            self._stat_counters.inc_dublicate_count()

            prev_lr_track: LrTrack = prev_entry.tr_track.lr_track

            # longer track is better
            if len(next_lr_track.src_jnode['PlateTrace']) > len(prev_lr_track.src_jnode['PlateTrace']):
                best_tr_track: TrTrack = deepcopy(next_tr_track)
            else:
                best_tr_track: TrTrack = deepcopy(prev_entry.tr_track)

            # entry.tr_track always appears earlier than tr_track
            plate_trace = prev_lr_track.src_jnode['PlateTrace'] + next_lr_track.src_jnode['PlateTrace']
            best_tr_track.lr_track.src_jnode['PlateTrace'] = plate_trace

            lr_track_refs = set(prev_lr_track.src_jnode['Id'] + next_lr_track.src_jnode['Id'])
            best_tr_track.lr_track.src_jnode['Id'] = list(lr_track_refs)

            # add the new image to store
            im_jnode = best_tr_track.lr_track.src_jnode['Image']
            im_ref = JsonImRef.deserialize_lr_track(im_jnode)
            if im_ref.url is not None and self._im_store:
                store_img_ref = self._im_store.add_image_ref(im_ref)
                im_ref.url = store_img_ref.url

                # remove the old one
                self._im_store.remove_image(prev_entry.image_ref)

            prev_entry.image_ref = im_ref
            prev_entry.tr_track = best_tr_track
            # TODO : check ra_track's of entry.tr_track and tr_track (is it possible to have 2 r-track for 1 real car ?)
            prev_entry.rm_delayed_call.reset(secondsFromNow=self._debounce_tout)

    def _consume_track(self, tr_track: TrTrack, im_ref: Optional[ImageRef]) -> None:
        self._stat_counters.inc_track_count(tr_track)

        # get the event's filename.
        event_filename = f'{self._event_file_prefix}-{self._event_file_seq:d}'
        self._event_file_seq += 1

        # send the event to target directories.
        # noinspection PyBroadException
        try:
            event_args = tr_track.serialize(im_ref)

            for event_dir in self._event_dirs:
                try_count = 2
                track_dir_present = True
                event_abspath = os.path.join(event_dir, event_filename)
                while try_count > 0 and track_dir_present:
                    try:
                        with open(event_abspath, 'w') as event_file:
                            event_file.write(event_args)

                        break

                    except IOError as e:
                        if e.errno == ENOENT:
                            track_dir_present = False

                    if not track_dir_present:
                        fs.silent_mkdir_r(event_dir)
                        track_dir_present = True
                    try_count -= 1

        except Exception:
            eventlog.log_x()


class SceneEventConsumer(ISceneEventConsumer):
    def __init__(self, ident: int, scene_event_args: SceneEventArgs, stat_counters: SceneEventCounters):
        ISceneEventConsumer.__init__(self)
        eventlog.log_d(f'ident:{ident} dirs:{repr(scene_event_args.event_dirs)}')

        self._stat_counters = stat_counters

        self._ident = ident
        self._event_dirs = scene_event_args.event_dirs
        self._skip_rate = scene_event_args.skip_rate
        self._frame_counter = 0

        list(map(fs.silent_mkdir_r, self._event_dirs))

    def consume_scene(self, im_ref: ImageRef, tr_scene: TrScene) -> None:
        self._frame_counter += 1
        self._frame_counter %= self._skip_rate

        if self._frame_counter:
            return

        # get the event's filename.
        event_filename = str(im_ref.ts)

        # send the event to target directories.
        # noinspection PyBroadException
        try:
            event_args = tr_scene.serialize(im_ref)
            for event_dir in self._event_dirs:
                try_count = 2
                scene_dir_present = True
                event_abspath = os.path.join(event_dir, event_filename)
                while try_count > 0 and scene_dir_present:
                    try:
                        with open(event_abspath, 'w') as event_file:
                            event_file.write(event_args)

                        self._stat_counters.inc_scene_count(self._ident)
                        break

                    except IOError as x:
                        if x.errno == ENOENT:
                            scene_dir_present = False

                    if not scene_dir_present:
                        fs.silent_mkdir_r(event_dir)
                        scene_dir_present = True

                    try_count -= 1

        except Exception:
            eventlog.log_x()
